@@ -53,12 +53,13 @@ void branch_new(void){
     usage("new BRANCH-NAME BASIS ?OPTIONS?");
   }
   db_find_and_open_repository(0, 0);
-  noSign = db_get_int("omitsign", 0)|noSign;
+  noSign = db_get_boolean("omitsign", 0)|noSign;
+  if( db_get_boolean("clearsign", 0)==0 ){ noSign = 1; }
 
   /* fossil branch new name */
   zBranch = g.argv[3];
   if( zBranch==0 || zBranch[0]==0 ){
-    fossil_panic("branch name cannot be empty");
+    fossil_fatal("branch name cannot be empty");
   }
   if( db_exists(
         "SELECT 1 FROM tagxref"
@@ -75,7 +76,7 @@ void branch_new(void){
     fossil_fatal("unable to locate check-in off of which to branch");
   }
 
-  pParent = manifest_get(rootid, CFTYPE_MANIFEST);
+  pParent = manifest_get(rootid, CFTYPE_MANIFEST, 0);
   if( pParent==0 ){
     fossil_fatal("%s is not a valid check-in", g.argv[4]);
   }
@@ -135,14 +136,15 @@ void branch_new(void){
   }
   db_finalize(&q);
 
-  blob_appendf(&branch, "U %F\n", zUserOvrd ? zUserOvrd : g.zLogin);
+  blob_appendf(&branch, "U %F\n", zUserOvrd ? zUserOvrd : login_name());
   md5sum_blob(&branch, &mcksum);
   blob_appendf(&branch, "Z %b\n", &mcksum);
   if( !noSign && clearsign(&branch, &branch) ){
     Blob ans;
-    blob_zero(&ans);
+    char cReply;
     prompt_user("unable to sign manifest.  continue (y/N)? ", &ans);
-    if( blob_str(&ans)[0]!='y' ){
+    cReply = blob_str(&ans)[0];
+    if( cReply!='y' && cReply!='Y'){
       db_end_transaction(1);
       fossil_exit(1);
     }
@@ -150,11 +152,11 @@ void branch_new(void){
 
   brid = content_put_ex(&branch, 0, 0, 0, isPrivate);
   if( brid==0 ){
-    fossil_panic("trouble committing manifest: %s", g.zErrMsg);
+    fossil_fatal("trouble committing manifest: %s", g.zErrMsg);
   }
   db_multi_exec("INSERT OR IGNORE INTO unsent VALUES(%d)", brid);
-  if( manifest_crosslink(brid, &branch)==0 ){
-    fossil_panic("unable to install new manifest");
+  if( manifest_crosslink(brid, &branch, MC_PERMIT_HOOKS)==0 ){
+    fossil_fatal("%s", g.zErrMsg);
   }
   assert( blob_is_reset(&branch) );
   content_deltify(rootid, brid, 0);
@@ -176,9 +178,22 @@ void branch_new(void){
   db_end_transaction(0);
 
   /* Do an autosync push, if requested */
-  if( !isPrivate ) autosync(AUTOSYNC_PUSH);
+  if( !isPrivate ) autosync_loop(SYNC_PUSH, db_get_int("autosync-tries",1),0);
 }
 
+#if INTERFACE
+/*
+** Allows bits in the mBplqFlags parameter to branch_prepare_list_query().
+*/
+#define BRL_CLOSED_ONLY      0x001 /* Show only closed branches */
+#define BRL_OPEN_ONLY        0x002 /* Show only open branches */
+#define BRL_BOTH             0x003 /* Show both open and closed branches */
+#define BRL_OPEN_CLOSED_MASK 0x003
+#define BRL_MTIME            0x004 /* Include lastest check-in time */
+#define BRL_ORDERBY_MTIME    0x008 /* Sort by MTIME. (otherwise sort by name)*/
+
+#endif /* INTERFACE */
+
 /*
 ** Prepare a query that will list branches.
 **
@@ -186,36 +201,43 @@ void branch_new(void){
 ** (which>0) then the query pulls all (closed and opened)
 ** branches. Else the query pulls currently-opened branches.
 */
-void branch_prepare_list_query(Stmt *pQuery, int which ){
-  if( which < 0 ){
-    db_prepare(pQuery,
-      "SELECT value FROM tagxref"
-      " WHERE tagid=%d AND value NOT NULL "
-      "EXCEPT "
-      "SELECT value FROM tagxref"
-      " WHERE tagid=%d"
-      "   AND rid IN leaf"
-      "   AND NOT %z"
-      " ORDER BY value COLLATE nocase /*sort*/",
-      TAG_BRANCH, TAG_BRANCH, leaf_is_closed_sql("tagxref.rid")
-    );
-  }else if( which>0 ){
-    db_prepare(pQuery,
-      "SELECT DISTINCT value FROM tagxref"
-      " WHERE tagid=%d AND value NOT NULL"
-      "   AND rid IN leaf"
-      " ORDER BY value COLLATE nocase /*sort*/",
-      TAG_BRANCH
-    );
-  }else{
-    db_prepare(pQuery,
-      "SELECT DISTINCT value FROM tagxref"
-      " WHERE tagid=%d AND value NOT NULL"
-      "   AND rid IN leaf"
-      "   AND NOT %z"
-      " ORDER BY value COLLATE nocase /*sort*/",
-      TAG_BRANCH, leaf_is_closed_sql("tagxref.rid")
-    );
+void branch_prepare_list_query(Stmt *pQuery, int brFlags){
+  switch( brFlags & BRL_OPEN_CLOSED_MASK ){
+    case BRL_CLOSED_ONLY: {
+      db_prepare(pQuery,
+        "SELECT value FROM tagxref"
+        " WHERE tagid=%d AND value NOT NULL "
+        "EXCEPT "
+        "SELECT value FROM tagxref"
+        " WHERE tagid=%d"
+        "   AND rid IN leaf"
+        "   AND NOT %z"
+        " ORDER BY value COLLATE nocase /*sort*/",
+        TAG_BRANCH, TAG_BRANCH, leaf_is_closed_sql("tagxref.rid")
+      );
+      break;
+    }
+    case BRL_BOTH: {
+      db_prepare(pQuery,
+        "SELECT DISTINCT value FROM tagxref"
+        " WHERE tagid=%d AND value NOT NULL"
+        "   AND rid IN leaf"
+        " ORDER BY value COLLATE nocase /*sort*/",
+        TAG_BRANCH
+      );
+      break;
+    }
+    case BRL_OPEN_ONLY: {
+      db_prepare(pQuery,
+        "SELECT DISTINCT value FROM tagxref"
+        " WHERE tagid=%d AND value NOT NULL"
+        "   AND rid IN leaf"
+        "   AND NOT %z"
+        " ORDER BY value COLLATE nocase /*sort*/",
+        TAG_BRANCH, leaf_is_closed_sql("tagxref.rid")
+      );
+      break;
+    }
   }
 }
 
@@ -238,11 +260,18 @@ void branch_prepare_list_query(Stmt *pQuery, int which ){
 **        --date-override DATE  DATE to use instead of 'now'
 **        --user-override USER  USER to use instead of the current default
 **
-**    %fossil branch list ?--all | --closed?
-**    %fossil branch ls ?--all | --closed?
+**        DATE may be "now" or "YYYY-MM-DDTHH:MM:SS.SSS". If in
+**        year-month-day form, it may be truncated, the "T" may be
+**        replaced by a space, and it may also name a timezone offset
+**        from UTC as "-HH:MM" (westward) or "+HH:MM" (eastward).
+**        Either no timezone suffix or "Z" means UTC.
 **
-**        List all branches.  Use --all or --closed to list all branches
-**        or closed branches.  The default is to show only open branches.
+**    %fossil branch list ?-a|--all|-c|--closed?
+**    %fossil branch ls ?-a|--all|-c|--closed?
+**
+**        List all branches.  Use -a or --all to list all branches and
+**        -c or --closed to list all closed branches.  The default is to
+**        show only open branches.
 **
 ** Options:
 **    -R|--repository FILE       Run commands on repository FILE
@@ -251,9 +280,6 @@ void branch_cmd(void){
   int n;
   const char *zCmd = "list";
   db_find_and_open_repository(0, 0);
-  if( g.argc<2 ){
-    usage("new|list|ls ...");
-  }
   if( g.argc>=3 ) zCmd = g.argv[2];
   n = strlen(zCmd);
   if( strncmp(zCmd,"new",n)==0 ){
@@ -262,15 +288,16 @@ void branch_cmd(void){
     Stmt q;
     int vid;
     char *zCurrent = 0;
-    int showAll = find_option("all",0,0)!=0;
-    int showClosed = find_option("closed",0,0)!=0;
+    int brFlags = BRL_OPEN_ONLY;
+    if( find_option("all","a",0)!=0 ) brFlags = BRL_BOTH;
+    if( find_option("closed","c",0)!=0 ) brFlags = BRL_CLOSED_ONLY;
 
     if( g.localOpen ){
       vid = db_lget_int("checkout", 0);
       zCurrent = db_text(0, "SELECT value FROM tagxref"
                             " WHERE rid=%d AND tagid=%d", vid, TAG_BRANCH);
     }
-    branch_prepare_list_query(&q, showAll?1:(showClosed?-1:0));
+    branch_prepare_list_query(&q, brFlags);
     while( db_step(&q)==SQLITE_ROW ){
       const char *zBr = db_column_text(&q, 0);
       int isCur = zCurrent!=0 && fossil_strcmp(zCurrent,zBr)==0;
@@ -278,36 +305,134 @@ void branch_cmd(void){
     }
     db_finalize(&q);
   }else{
-    fossil_panic("branch subcommand should be one of: "
+    fossil_fatal("branch subcommand should be one of: "
                  "new list ls");
   }
 }
 
+static const char brlistQuery[] =
+@ SELECT
+@   tagxref.value,
+@   max(event.mtime),
+@   EXISTS(SELECT 1 FROM tagxref AS tx
+@           WHERE tx.rid=tagxref.rid
+@             AND tx.tagid=(SELECT tagid FROM tag WHERE tagname='closed')
+@             AND tx.tagtype>0),
+@   (SELECT tagxref.value
+@      FROM plink CROSS JOIN tagxref
+@    WHERE plink.pid=event.objid
+@       AND tagxref.rid=plink.cid
+@      AND tagxref.tagid=(SELECT tagid FROM tag WHERE tagname='branch')
+@      AND tagtype>0),
+@   count(*),
+@   (SELECT uuid FROM blob WHERE rid=tagxref.rid)
+@  FROM tagxref, tag, event
+@ WHERE tagxref.tagid=tag.tagid
+@   AND tagxref.tagtype>0
+@   AND tag.tagname='branch'
+@   AND event.objid=tagxref.rid
+@ GROUP BY 1
+@ ORDER BY 2 DESC;
+;
+
+/*
+** This is the new-style branch-list page that shows the branch names
+** together with their ages (time of last check-in) and whether or not
+** they are closed or merged to another branch.
+**
+** Control jumps to this routine from brlist_page() (the /brlist handler)
+** if there are no query parameters.
+*/
+static void new_brlist_page(void){
+  Stmt q;
+  double rNow;
+  login_check_credentials();
+  if( !g.perm.Read ){ login_needed(g.anon.Read); return; }
+  style_header("Branches");
+  style_adunit_config(ADUNIT_RIGHT_OK);
+  login_anonymous_available();
+
+  db_prepare(&q, brlistQuery/*works-like:""*/);
+  rNow = db_double(0.0, "SELECT julianday('now')");
+  @ <div class="brlist"><table id="branchlisttable">
+  @ <thead><tr>
+  @ <th>Branch Name</th>
+  @ <th>Age</th>
+  @ <th>Check-ins</th>
+  @ <th>Status</th>
+  @ <th>Resolution</th>
+  @ </tr></thead><tbody>
+  while( db_step(&q)==SQLITE_ROW ){
+    const char *zBranch = db_column_text(&q, 0);
+    double rMtime = db_column_double(&q, 1);
+    int isClosed = db_column_int(&q, 2);
+    const char *zMergeTo = db_column_text(&q, 3);
+    int nCkin = db_column_int(&q, 4);
+    const char *zLastCkin = db_column_text(&q, 5);
+    char *zAge = human_readable_age(rNow - rMtime);
+    sqlite3_int64 iMtime = (sqlite3_int64)(rMtime*86400.0);
+    if( zMergeTo && zMergeTo[0]==0 ) zMergeTo = 0;
+    @ <tr>
+    @ <td>%z(href("%R/timeline?n=100&r=%T",zBranch))%h(zBranch)</a></td>
+    @ <td data-sortkey="%016llx(-iMtime)">%s(zAge)</td>
+    @ <td>%d(nCkin)</td>
+    fossil_free(zAge);
+    @ <td>%s(isClosed?"closed":"")</td>
+    if( zMergeTo ){
+      @ <td>merged into
+      @ %z(href("%R/timeline?f=%!S",zLastCkin))%h(zMergeTo)</a></td>
+    }else{
+      @ <td></td>
+    }
+    @ </tr>
+  }
+  @ </tbody></table></div>
+  db_finalize(&q);
+  output_table_sorting_javascript("branchlisttable","tkNtt",2);
+  style_footer();
+}
+
 /*
 ** WEBPAGE: brlist
+** Show a list of branches.  With no query parameters, a sortable table
+** is used to show all branches.  If query parameters are present a
+** fixed bullet list is shown.
 **
-** Show a timeline of all branches
+** Query parameters:
+**
+**     all         Show all branches
+**     closed      Show only closed branches
+**     open        Show only open branches (default behavior)
+**     colortest   Show all branches with automatic color
 */
 void brlist_page(void){
   Stmt q;
   int cnt;
   int showClosed = P("closed")!=0;
   int showAll = P("all")!=0;
+  int showOpen = P("open")!=0;
   int colorTest = P("colortest")!=0;
+  int brFlags = BRL_OPEN_ONLY;
 
+  if( showClosed==0 && showAll==0 && showOpen==0 && colorTest==0 ){
+    new_brlist_page();
+    return;
+  }
   login_check_credentials();
-  if( !g.perm.Read ){ login_needed(); return; }
+  if( !g.perm.Read ){ login_needed(g.anon.Read); return; }
   if( colorTest ){
     showClosed = 0;
     showAll = 1;
   }
+  if( showAll ) brFlags = BRL_BOTH;
+  if( showClosed ) brFlags = BRL_CLOSED_ONLY;
 
-  style_header(showClosed ? "Closed Branches" :
-                  showAll ? "All Branches" : "Open Branches");
+  style_header("%s", showClosed ? "Closed Branches" :
+                        showAll ? "All Branches" : "Open Branches");
   style_submenu_element("Timeline", "Timeline", "brtimeline");
   if( showClosed ){
     style_submenu_element("All", "All", "brlist?all");
-    style_submenu_element("Open","Open","brlist");
+    style_submenu_element("Open","Open","brlist?open");
   }else if( showAll ){
     style_submenu_element("Closed", "Closed", "brlist?closed");
     style_submenu_element("Open","Open","brlist");
@@ -321,11 +446,12 @@ void brlist_page(void){
     style_submenu_element("All", "All", "brlist?all");
   }
   login_anonymous_available();
+#if 0
   style_sidebox_begin("Nomenclature:", "33%");
   @ <ol>
   @ <li> An <div class="sideboxDescribed">%z(href("brlist"))
-  @ open branch</a></div> is a branch that has one or
-  @ more %z(href("leaves"))open leaves.</a>
+  @ open branch</a></div> is a branch that has one or more
+  @ <div class="sideboxDescribed">%z(href("leaves"))open leaves.</a></div>
   @ The presence of open leaves presumably means
   @ that the branch is still being extended with new check-ins.</li>
   @ <li> A <div class="sideboxDescribed">%z(href("brlist?closed"))
@@ -333,21 +459,22 @@ void brlist_page(void){
   @ <div class="sideboxDescribed">%z(href("leaves?closed"))
   @ closed leaves</a></div>.
   @ Closed branches are fixed and do not change (unless they are first
-  @ reopened)</li>
+  @ reopened).</li>
   @ </ol>
   style_sidebox_end();
+#endif
 
-  branch_prepare_list_query(&q, showAll?1:(showClosed?-1:0));
+  branch_prepare_list_query(&q, brFlags);
   cnt = 0;
   while( db_step(&q)==SQLITE_ROW ){
     const char *zBr = db_column_text(&q, 0);
     if( cnt==0 ){
       if( colorTest ){
         @ <h2>Default background colors for all branches:</h2>
-      }else if( showAll ){
-        @ <h2>All Branches:</h2>
       }else if( showClosed ){
         @ <h2>Closed Branches:</h2>
+      }else if( showAll ){
+        @ <h2>All Branches:</h2>
       }else{
         @ <h2>Open Branches:</h2>
       }
@@ -359,19 +486,13 @@ void brlist_page(void){
       @ <li><span style="background-color: %s(zColor)">
       @ %h(zBr) &rarr; %s(zColor)</span></li>
     }else{
-      @ <li>%z(href("%R/timeline?r=%T",zBr))%h(zBr)</a></li>
+      @ <li>%z(href("%R/timeline?r=%T&n=200",zBr))%h(zBr)</a></li>
     }
   }
   if( cnt ){
     @ </ul>
   }
   db_finalize(&q);
-  @ <script  type="text/JavaScript">
-  @ function xin(id){
-  @ }
-  @ function xout(id){
-  @ }
-  @ </script>
   style_footer();
 }
 
@@ -393,7 +514,7 @@ static void brtimeline_extra(int rid){
   );
   while( db_step(&q)==SQLITE_ROW ){
     const char *zTagName = db_column_text(&q, 0);
-    @ %z(href("%R/timeline?r=%T",zTagName))[timeline]</a>
+    @ %z(href("%R/timeline?r=%T&n=200",zTagName))[timeline]</a>
   }
   db_finalize(&q);
 }
@@ -407,7 +528,7 @@ void brtimeline_page(void){
   Stmt q;
 
   login_check_credentials();
-  if( !g.perm.Read ){ login_needed(); return; }
+  if( !g.perm.Read ){ login_needed(g.anon.Read); return; }
 
   style_header("Branches");
   style_submenu_element("List", "List", "brlist");
@@ -419,13 +540,7 @@ void brtimeline_page(void){
     " ORDER BY event.mtime DESC",
     timeline_query_for_www(), TAG_BRANCH
   );
-  www_print_timeline(&q, 0, 0, 0, brtimeline_extra);
+  www_print_timeline(&q, 0, 0, 0, 0, brtimeline_extra);
   db_finalize(&q);
-  @ <script  type="text/JavaScript">
-  @ function xin(id){
-  @ }
-  @ function xout(id){
-  @ }
-  @ </script>
   style_footer();
 }
