@@ -28,14 +28,12 @@
 **
 **     0:   There is an existing checkout but it is unmodified
 **     1:   There is a modified checkout - there are unsaved changes
-**     2:   There is no existing checkout
 */
-int unsaved_changes(void){
+int unsaved_changes(unsigned int cksigFlags){
   int vid;
   db_must_be_within_tree();
   vid = db_lget_int("checkout",0);
-  if( vid==0 ) return 2;
-  vfile_check_signature(vid, CKSIG_ENOTFILE);
+  vfile_check_signature(vid, cksigFlags|CKSIG_ENOTFILE);
   return db_exists("SELECT 1 FROM vfile WHERE chnged"
                    " OR coalesce(origname!=pathname,0)");
 }
@@ -45,8 +43,9 @@ int unsaved_changes(void){
 ** Clear the VFILE table.
 */
 void uncheckout(int vid){
-  if( vid==0 ) return;
-  vfile_unlink(vid);
+  if( vid>0 ){
+    vfile_unlink(vid);
+  }
   db_multi_exec("DELETE FROM vfile WHERE vid=%d", vid);
 }
 
@@ -57,22 +56,24 @@ void uncheckout(int vid){
 **
 ** If anything goes wrong, panic.
 */
-int load_vfile(const char *zName){
+int load_vfile(const char *zName, int forceMissingFlag){
   Blob uuid;
   int vid;
 
   blob_init(&uuid, zName, -1);
   if( name_to_uuid(&uuid, 1, "ci") ){
-    fossil_panic(g.zErrMsg);
+    fossil_fatal("%s", g.zErrMsg);
   }
   vid = db_int(0, "SELECT rid FROM blob WHERE uuid=%B", &uuid);
   if( vid==0 ){
     fossil_fatal("no such check-in: %s", g.argv[2]);
   }
   if( !is_a_version(vid) ){
-    fossil_fatal("object [%.10s] is not a check-in", blob_str(&uuid));
+    fossil_fatal("object [%S] is not a check-in", blob_str(&uuid));
   }
-  load_vfile_from_rid(vid);
+  if( load_vfile_from_rid(vid) && !forceMissingFlag ){
+    fossil_fatal("missing content, unable to checkout");
+  };
   return vid;
 }
 
@@ -105,10 +106,10 @@ void checkout_set_all_exe(int vid){
 
   /* Check the EXE permission status of all files
   */
-  pManifest = manifest_get(vid, CFTYPE_MANIFEST);
+  pManifest = manifest_get(vid, CFTYPE_MANIFEST, 0);
   if( pManifest==0 ) return;
   blob_zero(&filename);
-  blob_appendf(&filename, "%s/", g.zLocalRoot);
+  blob_appendf(&filename, "%s", g.zLocalRoot);
   baseLen = blob_size(&filename);
   manifest_file_rewind(pManifest);
   while( (pFile = manifest_file_next(pManifest, 0))!=0 ){
@@ -128,40 +129,99 @@ void checkout_set_all_exe(int vid){
 ** If the "manifest" setting is true, then automatically generate
 ** files named "manifest" and "manifest.uuid" containing, respectively,
 ** the text of the manifest and the artifact ID of the manifest.
+** If the manifest setting is set, but is not a boolean value, then treat
+** each character as a flag to enable writing "manifest", "manifest.uuid" or
+** "manifest.tags".
 */
 void manifest_to_disk(int vid){
   char *zManFile;
   Blob manifest;
   Blob hash;
+  Blob taglist;
+  int flg;
 
-  if( db_get_boolean("manifest",0) ){
+  flg = db_get_manifest_setting();
+
+  if( flg & (MFESTFLG_RAW|MFESTFLG_UUID) ){
     blob_zero(&manifest);
     content_get(vid, &manifest);
+    blob_zero(&hash);
+    sha1sum_blob(&manifest, &hash);
+    sterilize_manifest(&manifest);
+  }
+  if( flg & MFESTFLG_RAW ){
     zManFile = mprintf("%smanifest", g.zLocalRoot);
     blob_write_to_file(&manifest, zManFile);
     free(zManFile);
-    blob_zero(&hash);
-    sha1sum_blob(&manifest, &hash);
-    zManFile = mprintf("%smanifest.uuid", g.zLocalRoot);
-    blob_append(&hash, "\n", 1);
-    blob_write_to_file(&hash, zManFile);
-    free(zManFile);
-    blob_reset(&hash);
   }else{
     if( !db_exists("SELECT 1 FROM vfile WHERE pathname='manifest'") ){
       zManFile = mprintf("%smanifest", g.zLocalRoot);
       file_delete(zManFile);
       free(zManFile);
     }
+  }
+  if( flg & MFESTFLG_UUID ){
+    zManFile = mprintf("%smanifest.uuid", g.zLocalRoot);
+    blob_append(&hash, "\n", 1);
+    blob_write_to_file(&hash, zManFile);
+    free(zManFile);
+    blob_reset(&hash);
+  }else{
     if( !db_exists("SELECT 1 FROM vfile WHERE pathname='manifest.uuid'") ){
       zManFile = mprintf("%smanifest.uuid", g.zLocalRoot);
       file_delete(zManFile);
       free(zManFile);
     }
   }
-    
+  if( flg & MFESTFLG_TAGS ){
+    blob_zero(&taglist);
+    zManFile = mprintf("%smanifest.tags", g.zLocalRoot);
+    get_checkin_taglist(vid, &taglist);
+    blob_write_to_file(&taglist, zManFile);
+    free(zManFile);
+    blob_reset(&taglist);
+  }else{
+    if( !db_exists("SELECT 1 FROM vfile WHERE pathname='manifest.tags'") ){
+      zManFile = mprintf("%smanifest.tags", g.zLocalRoot);
+      file_delete(zManFile);
+      free(zManFile);
+    }
+  }
 }
 
+/*
+** Find the branch name and all symbolic tags for a particular check-in
+** identified by "rid".
+**
+** The branch name is actually only extracted if this procedure is run
+** from within a local check-out.  And the branch name is not the branch
+** name for "rid" but rather the branch name for the current check-out.
+** It is unclear if the rid parameter is always the same as the current
+** check-out.
+*/
+void get_checkin_taglist(int rid, Blob *pOut){
+  Stmt stmt;
+  char *zCurrent;
+  blob_reset(pOut);
+  zCurrent = db_text(0, "SELECT value FROM tagxref"
+                        " WHERE rid=%d AND tagid=%d", rid, TAG_BRANCH);
+  blob_appendf(pOut, "branch %s\n", zCurrent);
+  db_prepare(&stmt, "SELECT substr(tagname, 5)"
+                    "  FROM tagxref, tag"
+                    " WHERE tagxref.rid=%d"
+                    "   AND tagxref.tagtype>0"
+                    "   AND tag.tagid=tagxref.tagid"
+                    "   AND tag.tagname GLOB 'sym-*'", rid);
+  while( db_step(&stmt)==SQLITE_ROW ){
+    const char *zName;
+    zName = db_column_text(&stmt, 0);
+    blob_appendf(pOut, "tag %s\n", zName);
+  }
+  db_reset(&stmt);
+  db_finalize(&stmt);
+}
+
+
 /*
 ** COMMAND: checkout*
 ** COMMAND: co*
@@ -177,32 +237,39 @@ void manifest_to_disk(int vid){
 **
 ** The --latest flag can be used in place of VERSION to checkout the
 ** latest version in the repository.
-** 
+**
 ** Options:
-**    --force   Ignore edited files in the current checkout
-**    --keep    Only update the manifest and manifest.uuid files
+**    --force           Ignore edited files in the current checkout
+**    --keep            Only update the manifest and manifest.uuid files
+**    --force-missing   Force checkout even if content is missing
 **
 ** See also: update
 */
 void checkout_cmd(void){
   int forceFlag;                 /* Force checkout even if edits exist */
+  int forceMissingFlag;          /* Force checkout even if missing content */
   int keepFlag;                  /* Do not change any files on disk */
   int latestFlag;                /* Checkout the latest version */
   char *zVers;                   /* Version to checkout */
   int promptFlag;                /* True to prompt before overwriting */
   int vid, prior;
   Blob cksum1, cksum1b, cksum2;
-  
+
   db_must_be_within_tree();
   db_begin_transaction();
   forceFlag = find_option("force","f",0)!=0;
+  forceMissingFlag = find_option("force-missing",0,0)!=0;
   keepFlag = find_option("keep",0,0)!=0;
   latestFlag = find_option("latest",0,0)!=0;
   promptFlag = find_option("prompt",0,0)!=0 || forceFlag==0;
+
+  /* We should be done with options.. */
+  verify_all_options();
+
   if( (latestFlag!=0 && g.argc!=2) || (latestFlag==0 && g.argc!=3) ){
      usage("VERSION|--latest ?--force? ?--keep?");
   }
-  if( !forceFlag && unsaved_changes()==1 ){
+  if( !forceFlag && unsaved_changes(0) ){
     fossil_fatal("there are unsaved changes in the current checkout");
   }
   if( forceFlag ){
@@ -222,12 +289,12 @@ void checkout_cmd(void){
                          " ORDER BY event.mtime DESC");
     }
     if( zVers==0 ){
-      fossil_fatal("cannot locate \"latest\" checkout");
+      return;
     }
   }else{
     zVers = g.argv[2];
   }
-  vid = load_vfile(zVers);
+  vid = load_vfile(zVers, forceMissingFlag);
   if( prior==vid ){
     return;
   }
@@ -236,7 +303,7 @@ void checkout_cmd(void){
   }
   db_multi_exec("DELETE FROM vfile WHERE vid!=%d", vid);
   if( !keepFlag ){
-    vfile_to_disk(vid, 0, 1, promptFlag);
+    vfile_to_disk(vid, 0, !g.fQuiet, promptFlag);
   }
   checkout_set_all_exe(vid);
   manifest_to_disk(vid);
@@ -263,7 +330,7 @@ void checkout_cmd(void){
 static void unlink_local_database(int manifestOnly){
   const char *zReserved;
   int i;
-  for(i=0; (zReserved = fossil_reserved_name(i))!=0; i++){
+  for(i=0; (zReserved = fossil_reserved_name(i, 1))!=0; i++){
     if( manifestOnly==0 || zReserved[0]=='m' ){
       char *z;
       z = mprintf("%s%s", g.zLocalRoot, zReserved);
@@ -279,8 +346,8 @@ static void unlink_local_database(int manifestOnly){
 ** Usage: %fossil close ?OPTIONS?
 **
 ** The opposite of "open".  Close the current database connection.
-** Require a -f or --force flag if there are unsaved changed in the
-** current check-out.
+** Require a -f or --force flag if there are unsaved changes in the
+** current check-out or if there is non-empty stash.
 **
 ** Options:
 **   --force|-f  necessary to close a check out with uncommitted changes
@@ -290,11 +357,23 @@ static void unlink_local_database(int manifestOnly){
 void close_cmd(void){
   int forceFlag = find_option("force","f",0)!=0;
   db_must_be_within_tree();
-  if( !forceFlag && unsaved_changes()==1 ){
+
+  /* We should be done with options.. */
+  verify_all_options();
+
+  if( !forceFlag && unsaved_changes(0) ){
     fossil_fatal("there are unsaved changes in the current checkout");
   }
+  if( !forceFlag
+   && db_table_exists("localdb","stash")
+   && db_exists("SELECT 1 FROM localdb.stash")
+  ){
+    fossil_fatal("closing the checkout will delete your stash");
+  }
   if( db_is_writeable("repository") ){
-    db_multi_exec("DELETE FROM config WHERE name='ckout:%q'", g.zLocalRoot);
+    char *zUnset = mprintf("ckout:%q", g.zLocalRoot);
+    db_unset(zUnset, 1);
+    fossil_free(zUnset);
   }
   unlink_local_database(1);
   db_close(1);
