@@ -17,7 +17,7 @@
 **
 ** This file contains code used to manage repository configurations.
 **
-** By "responsitory configure" we mean the local state of a repository
+** By "repository configure" we mean the local state of a repository
 ** distinct from the versioned files.
 */
 #include "config.h"
@@ -43,6 +43,13 @@
 #define CONFIGSET_OVERWRITE 0x100000     /* Causes overwrite instead of merge */
 #define CONFIGSET_OLDFORMAT 0x200000     /* Use the legacy format */
 
+/*
+** This mask is used for the common TH1 configuration settings (i.e. those
+** that are not specific to one particular subsystem, such as the transfer
+** subsystem).
+*/
+#define CONFIGSET_TH1       (CONFIGSET_SKIN|CONFIGSET_TKT|CONFIGSET_XFER)
+
 #endif /* INTERFACE */
 
 /*
@@ -56,7 +63,7 @@ static struct {
   { "/email",        CONFIGSET_ADDR,  "Concealed email addresses in tickets" },
   { "/project",      CONFIGSET_PROJ,  "Project name and description"         },
   { "/skin",         CONFIGSET_SKIN | CONFIGSET_CSS,
-                                      "Web interface apparance settings"     },
+                                      "Web interface appearance settings"    },
   { "/css",          CONFIGSET_CSS,   "Style sheet"                          },
   { "/shun",         CONFIGSET_SHUN,  "List of shunned artifacts"            },
   { "/ticket",       CONFIGSET_TKT,   "Ticket setup",                        },
@@ -81,30 +88,52 @@ static struct {
   { "css",                    CONFIGSET_CSS  },
   { "header",                 CONFIGSET_SKIN },
   { "footer",                 CONFIGSET_SKIN },
+  { "details",                CONFIGSET_SKIN },
   { "logo-mimetype",          CONFIGSET_SKIN },
   { "logo-image",             CONFIGSET_SKIN },
   { "background-mimetype",    CONFIGSET_SKIN },
   { "background-image",       CONFIGSET_SKIN },
-  { "index-page",             CONFIGSET_SKIN },
   { "timeline-block-markup",  CONFIGSET_SKIN },
   { "timeline-max-comment",   CONFIGSET_SKIN },
+  { "timeline-plaintext",     CONFIGSET_SKIN },
   { "adunit",                 CONFIGSET_SKIN },
   { "adunit-omit-if-admin",   CONFIGSET_SKIN },
   { "adunit-omit-if-user",    CONFIGSET_SKIN },
 
+#ifdef FOSSIL_ENABLE_TH1_DOCS
+  { "th1-docs",               CONFIGSET_TH1 },
+#endif
+#ifdef FOSSIL_ENABLE_TH1_HOOKS
+  { "th1-hooks",              CONFIGSET_TH1 },
+#endif
+  { "th1-setup",              CONFIGSET_TH1 },
+  { "th1-uri-regexp",         CONFIGSET_TH1 },
+
 #ifdef FOSSIL_ENABLE_TCL
-  { "tcl",                    CONFIGSET_SKIN|CONFIGSET_TKT|CONFIGSET_XFER },
-  { "tcl-setup",              CONFIGSET_SKIN|CONFIGSET_TKT|CONFIGSET_XFER },
+  { "tcl",                    CONFIGSET_TH1 },
+  { "tcl-setup",              CONFIGSET_TH1 },
 #endif
 
   { "project-name",           CONFIGSET_PROJ },
+  { "short-project-name",     CONFIGSET_PROJ },
   { "project-description",    CONFIGSET_PROJ },
+  { "index-page",             CONFIGSET_PROJ },
   { "manifest",               CONFIGSET_PROJ },
   { "binary-glob",            CONFIGSET_PROJ },
+  { "clean-glob",             CONFIGSET_PROJ },
   { "ignore-glob",            CONFIGSET_PROJ },
+  { "keep-glob",              CONFIGSET_PROJ },
   { "crnl-glob",              CONFIGSET_PROJ },
+  { "encoding-glob",          CONFIGSET_PROJ },
   { "empty-dirs",             CONFIGSET_PROJ },
   { "allow-symlinks",         CONFIGSET_PROJ },
+  { "dotfiles",               CONFIGSET_PROJ },
+  { "parent-project-code",    CONFIGSET_PROJ },
+  { "parent-project-name",    CONFIGSET_PROJ },
+
+#ifdef FOSSIL_ENABLE_LEGACY_MV_RM
+  { "mv-rm-files",            CONFIGSET_PROJ },
+#endif
 
   { "ticket-table",           CONFIGSET_TKT  },
   { "ticket-common",          CONFIGSET_TKT  },
@@ -127,6 +156,8 @@ static struct {
 
   { "xfer-common-script",     CONFIGSET_XFER },
   { "xfer-push-script",       CONFIGSET_XFER },
+  { "xfer-commit-script",     CONFIGSET_XFER },
+  { "xfer-ticket-script",     CONFIGSET_XFER },
 
 };
 static int iConfig = 0;
@@ -166,7 +197,7 @@ const char *configure_next_name(int iMask){
 /*
 ** Return a pointer to a string that contains the RHS of an IN operator
 ** that will select CONFIG table names that are part of the configuration
-** that matchines iMatch.
+** that matches iMatch.
 */
 const char *configure_inop_rhs(int iMask){
   Blob x;
@@ -174,15 +205,15 @@ const char *configure_inop_rhs(int iMask){
   const char *zSep = "";
 
   blob_zero(&x);
-  blob_append(&x, "(", 1);
+  blob_append_sql(&x, "(");
   for(i=0; i<count(aConfig); i++){
     if( (aConfig[i].groupMask & iMask)==0 ) continue;
     if( aConfig[i].zName[0]=='@' ) continue;
-    blob_appendf(&x, "%s'%s'", zSep, aConfig[i].zName);
+    blob_append_sql(&x, "%s'%q'", zSep/*safe-for-%s*/, aConfig[i].zName);
     zSep = ",";
   }
-  blob_append(&x, ")", 1);
-  return blob_str(&x);
+  blob_append_sql(&x, ")");
+  return blob_sql_text(&x);
 }
 
 /*
@@ -202,7 +233,7 @@ int configure_is_exportable(const char *zName){
     n -= 2;
   }
   for(i=0; i<count(aConfig); i++){
-    if( memcmp(zName, aConfig[i].zName, n)==0 && aConfig[i].zName[n]==0 ){
+    if( strncmp(zName, aConfig[i].zName, n)==0 && aConfig[i].zName[n]==0 ){
       int m = aConfig[i].groupMask;
       if( !g.perm.Admin ){
         m &= ~CONFIGSET_USER;
@@ -307,7 +338,7 @@ static void config_reset_function(
 
 /*
 ** Create the temporary _xfer_reportfmt and _xfer_user tables that are
-** necessary in order to evalute the SQL text generated by the
+** necessary in order to evaluate the SQL text generated by the
 ** configure_render_special_name() routine.
 **
 ** If replaceFlag is true, then the setup is such that the content in
@@ -341,7 +372,8 @@ void configure_prepare_to_receive(int replaceFlag){
     @ INSERT INTO _xfer_user
     @    SELECT uid,login,pw,cap,cookie,ipaddr,cexpire,info,photo FROM user;
   ;
-  db_multi_exec(zSQL1);
+  assert( strchr(zSQL1,'%')==0 );
+  db_multi_exec(zSQL1 /*works-like:""*/);
 
   /* When the replace flag is set, add triggers that run the first time
   ** that new data is seen.  The triggers run only once and delete all the
@@ -370,7 +402,8 @@ void configure_prepare_to_receive(int replaceFlag){
     sqlite3_create_function(g.db, "config_reset", 1, SQLITE_UTF8, 0,
          config_reset_function, 0, 0);
     configHasBeenReset = 0;
-    db_multi_exec(zSQL2);
+    assert( strchr(zSQL2,'%')==0 );
+    db_multi_exec(zSQL2 /*works-like:""*/);
   }
 }
 
@@ -387,7 +420,23 @@ void configure_finalize_receive(void){
     @ DROP TABLE _xfer_user;
     @ DROP TABLE _xfer_reportfmt;
   ;
-  db_multi_exec(zSQL);
+  assert( strchr(zSQL,'%')==0 );
+  db_multi_exec(zSQL /*works-like:""*/);
+}
+
+/*
+** Mask of modified configuration sets
+*/
+static int rebuildMask = 0;
+
+/*
+** Rebuild auxiliary tables as required by configuration changes.
+*/
+void configure_rebuild(void){
+  if( rebuildMask & CONFIGSET_TKT ){
+    ticket_rebuild();
+  }
+  rebuildMask = 0;
 }
 
 /*
@@ -462,7 +511,7 @@ static int safeInt(const char *z){
 **
 ** OLD FORMAT:
 **
-** The old format is retained for backwards compatiblity, but is deprecated.
+** The old format is retained for backwards compatibility, but is deprecated.
 ** The cutover from old format to new was on 2011-04-25.  After sufficient
 ** time has passed, support for the old format will be removed.
 **
@@ -528,33 +577,38 @@ void configure_receive(const char *zName, Blob *pContent, int groupMask){
     blob_zero(&sql);
     if( groupMask & CONFIGSET_OVERWRITE ){
       if( (thisMask & configHasBeenReset)==0 && aType[ii].zName[0]!='/' ){
-        db_multi_exec("DELETE FROM %s", &aType[ii].zName[1]);
+        db_multi_exec("DELETE FROM \"%w\"", &aType[ii].zName[1]);
         configHasBeenReset |= thisMask;
       }
-      blob_append(&sql, "REPLACE INTO ", -1);
+      blob_append_sql(&sql, "REPLACE INTO ");
     }else{
-      blob_append(&sql, "INSERT OR IGNORE INTO ", -1);
+      blob_append_sql(&sql, "INSERT OR IGNORE INTO ");
     }
-    blob_appendf(&sql, "%s(%s, mtime", &zName[1], aType[ii].zPrimKey);
+    blob_append_sql(&sql, "\"%w\"(\"%w\", mtime", &zName[1], aType[ii].zPrimKey);
     for(jj=2; jj<nToken; jj+=2){
-       blob_appendf(&sql, ",%s", azToken[jj]);
+       blob_append_sql(&sql, ",\"%w\"", azToken[jj]);
     }
-    blob_appendf(&sql,") VALUES(%s,%s", azToken[1], azToken[0]);
+    blob_append_sql(&sql,") VALUES(%s,%s",
+       azToken[1] /*safe-for-%s*/, azToken[0] /*safe-for-%s*/);
     for(jj=2; jj<nToken; jj+=2){
-       blob_appendf(&sql, ",%s", azToken[jj+1]);
+       blob_append_sql(&sql, ",%s", azToken[jj+1] /*safe-for-%s*/);
     }
-    db_multi_exec("%s)", blob_str(&sql));
+    db_multi_exec("%s)", blob_sql_text(&sql));
     if( db_changes()==0 ){
       blob_reset(&sql);
-      blob_appendf(&sql, "UPDATE %s SET mtime=%s", &zName[1], azToken[0]);
+      blob_append_sql(&sql, "UPDATE \"%w\" SET mtime=%s",
+                      &zName[1], azToken[0]/*safe-for-%s*/);
       for(jj=2; jj<nToken; jj+=2){
-        blob_appendf(&sql, ", %s=%s", azToken[jj], azToken[jj+1]);
+        blob_append_sql(&sql, ", \"%w\"=%s",
+                        azToken[jj], azToken[jj+1]/*safe-for-%s*/);
       }
-      blob_appendf(&sql, " WHERE %s=%s AND mtime<%s",
-                   aType[ii].zPrimKey, azToken[1], azToken[0]);
-      db_multi_exec("%s", blob_str(&sql));
+      blob_append_sql(&sql, " WHERE \"%w\"=%s AND mtime<%s",
+                   aType[ii].zPrimKey, azToken[1]/*safe-for-%s*/,
+                   azToken[0]/*safe-for-%s*/);
+      db_multi_exec("%s", blob_sql_text(&sql));
     }
     blob_reset(&sql);
+    rebuildMask |= thisMask;
   }else{
     /* Otherwise, the old format */
     if( (configure_is_exportable(zName) & groupMask)==0 ) return;
@@ -573,7 +627,7 @@ void configure_receive(const char *zName, Blob *pContent, int groupMask){
       ** as an administrator, so presumably we trust the client at this
       ** point.
       */
-      db_multi_exec("%s", blob_str(pContent));
+      db_multi_exec("%s", blob_str(pContent) /*safe-for-%s*/);
     }else{
       db_multi_exec(
          "REPLACE INTO config(name,value,mtime) VALUES(%Q,%Q,now())",
@@ -780,11 +834,11 @@ static void export_config(
 ** Usage: %fossil configuration METHOD ... ?OPTIONS?
 **
 ** Where METHOD is one of: export import merge pull push reset.  All methods
-** accept the -R or --repository option to specific a repository.
+** accept the -R or --repository option to specify a repository.
 **
 **    %fossil configuration export AREA FILENAME
 **
-**         Write to FILENAME exported configuraton information for AREA.
+**         Write to FILENAME exported configuration information for AREA.
 **         AREA can be one of:  all email project shun skin ticket user
 **
 **    %fossil configuration import FILENAME
@@ -813,7 +867,7 @@ static void export_config(
 **         by URL.  Admin privilege is required on the remote server for
 **         this to work.  When the same record exists both locally and on
 **         the remote end, the one that was most recently changed wins.
-**         Use the --legacy flag when talking to holder servers.
+**         Use the --legacy flag when talking to older servers.
 **
 **    %fossil configuration reset AREA
 **
@@ -832,11 +886,11 @@ static void export_config(
 void configuration_cmd(void){
   int n;
   const char *zMethod;
-  if( g.argc<3 ){
-    usage("export|import|merge|pull|reset ...");
-  }
   db_find_and_open_repository(0, 0);
-  db_open_config(0);
+  db_open_config(0, 0);
+  if( g.argc<3 ){
+    usage("SUBCOMMAND ...");
+  }
   zMethod = g.argv[2];
   n = strlen(zMethod);
   if( strncmp(zMethod, "export", n)==0 ){
@@ -877,42 +931,34 @@ void configuration_cmd(void){
    || strncmp(zMethod, "sync", n)==0
   ){
     int mask;
-    const char *zServer;
-    const char *zPw;
+    const char *zServer = 0;
     int legacyFlag = 0;
     int overwriteFlag = 0;
+
     if( zMethod[0]!='s' ) legacyFlag = find_option("legacy",0,0)!=0;
     if( strncmp(zMethod,"pull",n)==0 ){
       overwriteFlag = find_option("overwrite",0,0)!=0;
     }
     url_proxy_options();
     if( g.argc!=4 && g.argc!=5 ){
-      usage("pull AREA ?URL?");
+      usage(mprintf("%s AREA ?URL?", zMethod));
     }
     mask = configure_name_to_mask(g.argv[3], 1);
     if( g.argc==5 ){
       zServer = g.argv[4];
-      zPw = 0;
-      g.dontKeepUrl = 1;
-    }else{
-      zServer = db_get("last-sync-url", 0);
-      if( zServer==0 ){
-        fossil_fatal("no server specified");
-      }
-      zPw = unobscure(db_get("last-sync-pw", 0));
     }
-    url_parse(zServer);
-    if( g.urlPasswd==0 && zPw ) g.urlPasswd = mprintf("%s", zPw);
+    url_parse(zServer, URL_PROMPT_PW);
+    if( g.url.protocol==0 ) fossil_fatal("no server URL specified");
     user_select();
     url_enable_proxy("via proxy: ");
     if( legacyFlag ) mask |= CONFIGSET_OLDFORMAT;
     if( overwriteFlag ) mask |= CONFIGSET_OVERWRITE;
     if( strncmp(zMethod, "push", n)==0 ){
-      client_sync(0,0,0,0,0,mask);
+      client_sync(0,0,(unsigned)mask);
     }else if( strncmp(zMethod, "pull", n)==0 ){
-      client_sync(0,0,0,0,mask,0);
+      client_sync(0,(unsigned)mask,0);
     }else{
-      client_sync(0,0,0,0,mask,mask);
+      client_sync(0,(unsigned)mask,(unsigned)mask);
     }
   }else
   if( strncmp(zMethod, "reset", n)==0 ){
@@ -938,15 +984,210 @@ void configuration_cmd(void){
         db_multi_exec("DELETE FROM shun");
       }else if( fossil_strcmp(zName,"@reportfmt")==0 ){
         db_multi_exec("DELETE FROM reportfmt");
+        assert( strchr(zRepositorySchemaDefaultReports,'%')==0 );
+        db_multi_exec(zRepositorySchemaDefaultReports /*works-like:""*/);
       }
     }
     db_end_transaction(0);
     fossil_print("Configuration reset to factory defaults.\n");
     fossil_print("To recover, use:  %s %s import %s\n",
             g.argv[0], g.argv[1], zBackup);
+    rebuildMask |= mask;
   }else
   {
     fossil_fatal("METHOD should be one of:"
                  " export import merge pull push reset");
   }
+  configure_rebuild();
+}
+
+
+/*
+** COMMAND: test-var-list
+**
+** Usage: %fossil test-var-list ?PATTERN? ?--unset? ?--mtime?
+**
+** Show the content of the CONFIG table in a repository.  If PATTERN is
+** specified, then only show the entries that match that glob pattern.
+** Last modification time is shown if the --mtime option is present.
+**
+** If the --unset option is included, then entries are deleted rather than
+** being displayed.  WARNING! This cannot be undone.  Be sure you know what
+** you are doing!  The --unset option only works if there is a PATTERN.
+** Probably you should run the command once without --unset to make sure
+** you know exactly what is being deleted.
+**
+** If not in an open check-out, use the -R REPO option to specify a
+** a repository.
+*/
+void test_var_list_cmd(void){
+  Stmt q;
+  int i, j;
+  const char *zPattern = 0;
+  int doUnset;
+  int showMtime;
+  Blob sql;
+  Blob ans;
+  unsigned char zTrans[1000];
+
+  doUnset = find_option("unset",0,0)!=0;
+  showMtime = find_option("mtime",0,0)!=0;
+  db_find_and_open_repository(OPEN_ANY_SCHEMA, 0);
+  verify_all_options();
+  if( g.argc>=3 ){
+    zPattern = g.argv[2];
+  }
+  blob_init(&sql,0,0);
+  blob_appendf(&sql, "SELECT name, value, datetime(mtime,'unixepoch')"
+                     " FROM config");
+  if( zPattern ){
+    blob_appendf(&sql, " WHERE name GLOB %Q", zPattern);
+  }
+  if( showMtime ){
+    blob_appendf(&sql, " ORDER BY mtime, name");
+  }else{
+    blob_appendf(&sql, " ORDER BY name");
+  }
+  db_prepare(&q, "%s", blob_str(&sql)/*safe-for-%s*/);
+  blob_reset(&sql);
+#define MX_VAL 40
+#define MX_NM  28
+#define MX_LONGNM 60
+  while( db_step(&q)==SQLITE_ROW ){
+    const char *zName = db_column_text(&q,0);
+    int nName = db_column_bytes(&q,0);
+    const char *zValue = db_column_text(&q,1);
+    int szValue = db_column_bytes(&q,1);
+    const char *zMTime = db_column_text(&q,2);
+    for(i=j=0; j<MX_VAL && zValue[i]; i++){
+      unsigned char c = (unsigned char)zValue[i];
+      if( c>=' ' && c<='~' ){
+        zTrans[j++] = c;
+      }else{
+        zTrans[j++] = '\\';
+        if( c=='\n' ){
+          zTrans[j++] = 'n';
+        }else if( c=='\r' ){
+          zTrans[j++] = 'r';
+        }else if( c=='\t' ){
+          zTrans[j++] = 't';
+        }else{
+          zTrans[j++] = '0' + ((c>>6)&7);
+          zTrans[j++] = '0' + ((c>>3)&7);
+          zTrans[j++] = '0' + (c&7);
+        }
+      }
+    }
+    zTrans[j] = 0;
+    if( i<szValue ){
+      sqlite3_snprintf(sizeof(zTrans)-j, (char*)zTrans+j, "...+%d", szValue-i);
+      j += (int)strlen((char*)zTrans+j);
+    }
+    if( showMtime ){
+      fossil_print("%s:%*s%s\n", zName, 58-nName, "", zMTime);
+    }else if( nName<MX_NM-2 ){
+      fossil_print("%s:%*s%s\n", zName, MX_NM-1-nName, "", zTrans);
+    }else if( nName<MX_LONGNM-2 && j<10 ){
+      fossil_print("%s:%*s%s\n", zName, MX_LONGNM-1-nName, "", zTrans);
+    }else{
+      fossil_print("%s:\n%*s%s\n", zName, MX_NM, "", zTrans);
+    }
+  }
+  db_finalize(&q);
+  if( zPattern && doUnset ){
+    prompt_user("Delete all of the above? (y/N)? ", &ans);
+    if( blob_str(&ans)[0]=='y' || blob_str(&ans)[0]=='Y' ){
+      db_multi_exec("DELETE FROM config WHERE name GLOB %Q", zPattern);
+    }
+    blob_reset(&ans);
+  }
+}
+
+/*
+** COMMAND: test-var-get
+**
+** Usage: %fossil test-var-get VAR ?FILE?
+**
+** Write the text of the VAR variable into FILE.  If FILE is "-"
+** or is omitted then output goes to standard output.  VAR can be a
+** GLOB pattern.
+**
+** If not in an open check-out, use the -R REPO option to specify a
+** a repository.
+*/
+void test_var_get_cmd(void){
+  const char *zVar;
+  const char *zFile;
+  int n;
+  Blob x;
+  db_find_and_open_repository(OPEN_ANY_SCHEMA, 0);
+  verify_all_options();
+  if( g.argc<3 ){
+    usage("VAR ?FILE?");
+  }
+  zVar = g.argv[2];
+  zFile = g.argc>=4 ? g.argv[3] : "-";
+  n = db_int(0, "SELECT count(*) FROM config WHERE name GLOB %Q", zVar);
+  if( n==0 ){
+    fossil_fatal("no match for %Q", zVar);
+  }
+  if( n>1 ){
+    fossil_fatal("multiple matches: %s",
+      db_text(0, "SELECT group_concat(quote(name),', ') FROM ("
+                 " SELECT name FROM config WHERE name GLOB %Q ORDER BY 1)",
+                 zVar));
+  }
+  blob_init(&x,0,0);
+  db_blob(&x, "SELECT value FROM config WHERE name GLOB %Q", zVar);
+  blob_write_to_file(&x, zFile);
+}
+
+/*
+** COMMAND: test-var-set
+**
+** Usage: %fossil test-var-set VAR ?VALUE? ?--file FILE?
+**
+** Store VALUE or the content of FILE (exactly one of which must be
+** supplied) into variable VAR.  Use a FILE of "-" to read from
+** standard input.
+**
+** WARNING: changing the value of a variable can interfere with the
+** operation of Fossil.  Be sure you know what you are doing.
+**
+** Use "--blob FILE" instead of "--file FILE" to load a binary blob
+** such as a GIF.
+*/
+void test_var_set_cmd(void){
+  const char *zVar;
+  const char *zFile;
+  const char *zBlob;
+  Blob x;
+  Stmt ins;
+  zFile = find_option("file",0,1);
+  zBlob = find_option("blob",0,1);
+  db_find_and_open_repository(OPEN_ANY_SCHEMA, 0);
+  verify_all_options();
+  if( g.argc<3 || (zFile==0 && zBlob==0 && g.argc<4) ){
+    usage("VAR ?VALUE? ?--file FILE?");
+  }
+  zVar = g.argv[2];
+  if( zFile ){
+    if( zBlob ) fossil_fatal("cannot do both --file or --blob");
+    blob_read_from_file(&x, zFile);
+  }else if( zBlob ){
+    blob_read_from_file(&x, zBlob);
+  }else{
+    blob_init(&x,g.argv[3],-1);
+  }
+  db_prepare(&ins,
+     "REPLACE INTO config(name,value,mtime)"
+     "VALUES(%Q,:val,now())", zVar);
+  if( zBlob ){
+    db_bind_blob(&ins, ":val", &x);
+  }else{
+    db_bind_text(&ins, ":val", blob_str(&x));
+  }
+  db_step(&ins);
+  db_finalize(&ins);
+  blob_reset(&x);
 }
