@@ -30,6 +30,81 @@
 #  define NULL_DEVICE "/dev/null"
 #endif
 
+/*
+** Used when the name for the diff is unknown.
+*/
+#define DIFF_NO_NAME  "(unknown)"
+
+/*
+** Use the "exec-rel-paths" setting and the --exec-abs-paths and
+** --exec-rel-paths command line options to determine whether
+** certain external commands are executed using relative paths.
+*/
+static int determine_exec_relative_option(int force){
+  static int relativePaths = -1;
+  if( force || relativePaths==-1 ){
+    int relPathOption = find_option("exec-rel-paths", 0, 0)!=0;
+    int absPathOption = find_option("exec-abs-paths", 0, 0)!=0;
+#if defined(FOSSIL_ENABLE_EXEC_REL_PATHS)
+    relativePaths = db_get_boolean("exec-rel-paths", 1);
+#else
+    relativePaths = db_get_boolean("exec-rel-paths", 0);
+#endif
+    if( relPathOption ){ relativePaths = 1; }
+    if( absPathOption ){ relativePaths = 0; }
+  }
+  return relativePaths;
+}
+
+#if INTERFACE
+/*
+** An array of FileDirList objects describe the files and directories listed
+** on the command line of a "diff" command.  Only those objects listed are
+** actually diffed.
+*/
+struct FileDirList {
+  int nUsed;       /* Number of times each entry is used */
+  int nName;       /* Length of the entry */
+  char *zName;     /* Text of the entry */
+};
+#endif
+
+/*
+** Return true if zFile is a file named on the azInclude[] list or is
+** a file in a directory named on the azInclude[] list.
+**
+** if azInclude is NULL, then always include zFile.
+*/
+static int file_dir_match(FileDirList *p, const char *zFile){
+  if( p==0 || strcmp(p->zName,".")==0 ) return 1;
+  if( filenames_are_case_sensitive() ){
+    while( p->zName ){
+      if( strcmp(zFile, p->zName)==0
+       || (strncmp(zFile, p->zName, p->nName)==0
+           && zFile[p->nName]=='/')
+      ){
+        break;
+      }
+      p++;
+    }
+  }else{
+    while( p->zName ){
+      if( fossil_stricmp(zFile, p->zName)==0
+       || (fossil_strnicmp(zFile, p->zName, p->nName)==0
+           && zFile[p->nName]=='/')
+      ){
+        break;
+      }
+      p++;
+    }
+  }
+  if( p->zName ){
+    p->nUsed++;
+    return 1;
+  }
+  return 0;
+}
+
 /*
 ** Print the "Index:" message that patches wants to see at the top of a diff.
 */
@@ -51,11 +126,21 @@ void diff_print_filenames(const char *zLeft, const char *zRight, u64 diffFlags){
   }else if( diffFlags & DIFF_SIDEBYSIDE ){
     int w = diff_width(diffFlags);
     int n1 = strlen(zLeft);
+    int n2 = strlen(zRight);
     int x;
-    if( n1>w*2 ) n1 = w*2;
-    x = w*2+17 - (n1+2);
-    z = mprintf("%.*c %.*s %.*c\n",
-                x/2, '=', n1, zLeft, (x+1)/2, '=');
+    if( n1==n2 && fossil_strcmp(zLeft,zRight)==0 ){
+      if( n1>w*2 ) n1 = w*2;
+      x = w*2+17 - (n1+2);
+      z = mprintf("%.*c %.*s %.*c\n",
+                 x/2, '=', n1, zLeft, (x+1)/2, '=');
+    }else{
+      if( w<20 ) w = 20;
+      if( n1>w-10 ) n1 = w - 10;
+      if( n2>w-10 ) n2 = w - 10;
+      z = mprintf("%.*c %.*s %.*c versus %.*c %.*s %.*c\n",
+                  (w-n1+10)/2, '=', n1, zLeft, (w-n1+1)/2, '=',
+                  (w-n2)/2, '=', n2, zRight, (w-n2+1)/2, '=');
+    }
   }else{
     z = mprintf("--- %s\n+++ %s\n", zLeft, zRight);
   }
@@ -96,7 +181,7 @@ void diff_file(
     if( file_wd_size(zFile2)<0 ){
       zName2 = NULL_DEVICE;
     }else{
-      if( file_wd_islink(zFile2) ){
+      if( file_wd_islink(0) ){
         blob_read_link(&file2, zFile2);
       }else{
         blob_read_from_file(&file2, zFile2);
@@ -111,7 +196,7 @@ void diff_file(
       }
     }else{
       blob_zero(&out);
-      text_diff(pFile1, &file2, &out, diffFlags);
+      text_diff(pFile1, &file2, &out, 0, diffFlags);
       if( blob_size(&out) ){
         diff_print_filenames(zName, zName2, diffFlags);
         fossil_print("%s\n", blob_str(&out));
@@ -129,13 +214,13 @@ void diff_file(
     if( !fIncludeBinary ){
       Blob file2;
       if( isBin1 ){
-        fossil_print(DIFF_CANNOT_COMPUTE_BINARY);
+        fossil_print("%s",DIFF_CANNOT_COMPUTE_BINARY);
         return;
       }
       if( zBinGlob ){
         Glob *pBinary = glob_create(zBinGlob);
         if( glob_match(pBinary, zName) ){
-          fossil_print(DIFF_CANNOT_COMPUTE_BINARY);
+          fossil_print("%s",DIFF_CANNOT_COMPUTE_BINARY);
           glob_free(pBinary);
           return;
         }
@@ -143,14 +228,14 @@ void diff_file(
       }
       blob_zero(&file2);
       if( file_wd_size(zFile2)>=0 ){
-        if( file_wd_islink(zFile2) ){
+        if( file_wd_islink(0) ){
           blob_read_link(&file2, zFile2);
         }else{
           blob_read_from_file(&file2, zFile2);
         }
       }
       if( looks_like_binary(&file2) ){
-        fossil_print(DIFF_CANNOT_COMPUTE_BINARY);
+        fossil_print("%s",DIFF_CANNOT_COMPUTE_BINARY);
         blob_reset(&file2);
         return;
       }
@@ -163,7 +248,7 @@ void diff_file(
     do{
       blob_reset(&nameFile1);
       blob_appendf(&nameFile1, "%s~%d", zFile2, cnt++);
-    }while( file_access(blob_str(&nameFile1),0)==0 );
+    }while( file_access(blob_str(&nameFile1),F_OK)==0 );
     blob_write_to_file(pFile1, blob_str(&nameFile1));
 
     /* Construct the external diff command */
@@ -212,7 +297,7 @@ void diff_file_mem(
     Blob out;      /* Diff output text */
 
     blob_zero(&out);
-    text_diff(pFile1, pFile2, &out, diffFlags);
+    text_diff(pFile1, pFile2, &out, 0, diffFlags);
     diff_print_filenames(zName, zName, diffFlags);
     fossil_print("%s\n", blob_str(&out));
 
@@ -225,13 +310,13 @@ void diff_file_mem(
 
     if( !fIncludeBinary ){
       if( isBin1 || isBin2 ){
-        fossil_print(DIFF_CANNOT_COMPUTE_BINARY);
+        fossil_print("%s",DIFF_CANNOT_COMPUTE_BINARY);
         return;
       }
       if( zBinGlob ){
         Glob *pBinary = glob_create(zBinGlob);
         if( glob_match(pBinary, zName) ){
-          fossil_print(DIFF_CANNOT_COMPUTE_BINARY);
+          fossil_print("%s",DIFF_CANNOT_COMPUTE_BINARY);
           glob_free(pBinary);
           return;
         }
@@ -262,42 +347,6 @@ void diff_file_mem(
   }
 }
 
-/*
-** Do a diff against a single file named in zFileTreeName from version zFrom
-** against the same file on disk.
-**
-** Use the internal diff logic if zDiffCmd is NULL.  Otherwise call the
-** command zDiffCmd to do the diffing.
-**
-** When using an external diff program, zBinGlob contains the GLOB patterns
-** for file names to treat as binary.  If fIncludeBinary is zero, these files
-** will be skipped in addition to files that may contain binary content.
-*/
-static void diff_one_against_disk(
-  const char *zFrom,        /* Name of file */
-  const char *zDiffCmd,     /* Use this "diff" command */
-  const char *zBinGlob,     /* Treat file names matching this as binary */
-  int fIncludeBinary,       /* Include binary files for external diff */
-  u64 diffFlags,            /* Diff control flags */
-  const char *zFileTreeName
-){
-  Blob fname;
-  Blob content;
-  int isLink;
-  int isBin;
-  file_tree_name(zFileTreeName, &fname, 1);
-  historical_version_of_file(zFrom, blob_str(&fname), &content, &isLink, 0,
-                             fIncludeBinary ? 0 : &isBin, 0);
-  if( !isLink != !file_wd_islink(zFrom) ){
-    fossil_print(DIFF_CANNOT_COMPUTE_SYMLINK);
-  }else{
-    diff_file(&content, isBin, zFileTreeName, zFileTreeName,
-              zDiffCmd, zBinGlob, fIncludeBinary, diffFlags);
-  }
-  blob_reset(&content);
-  blob_reset(&fname);
-}
-
 /*
 ** Run a diff between the version zFrom and files on disk.  zFrom might
 ** be NULL which means to simply show the difference between the edited
@@ -310,19 +359,20 @@ static void diff_one_against_disk(
 ** for file names to treat as binary.  If fIncludeBinary is zero, these files
 ** will be skipped in addition to files that may contain binary content.
 */
-static void diff_all_against_disk(
+static void diff_against_disk(
   const char *zFrom,        /* Version to difference from */
   const char *zDiffCmd,     /* Use this diff command.  NULL for built-in */
   const char *zBinGlob,     /* Treat file names matching this as binary */
   int fIncludeBinary,       /* Treat file names matching this as binary */
-  u64 diffFlags             /* Flags controlling diff output */
+  u64 diffFlags,            /* Flags controlling diff output */
+  FileDirList *pFileDir     /* Which files to diff */
 ){
   int vid;
   Blob sql;
   Stmt q;
   int asNewFile;            /* Treat non-existant files as empty files */
 
-  asNewFile = (diffFlags & DIFF_NEWFILE)!=0;
+  asNewFile = (diffFlags & DIFF_VERBOSE)!=0;
   vid = db_lget_int("checkout", 0);
   vfile_check_signature(vid, CKSIG_ENOTFILE);
   blob_zero(&sql);
@@ -333,7 +383,7 @@ static void diff_all_against_disk(
       fossil_fatal("no such check-in: %s", zFrom);
     }
     load_vfile_from_rid(rid);
-    blob_appendf(&sql,
+    blob_append_sql(&sql,
       "SELECT v2.pathname, v2.deleted, v2.chnged, v2.rid==0, v1.rid, v1.islink"
       "  FROM vfile v1, vfile v2 "
       " WHERE v1.pathname=v2.pathname AND v1.vid=%d AND v2.vid=%d"
@@ -350,20 +400,20 @@ static void diff_all_against_disk(
       " WHERE v2.vid=%d"
       "   AND NOT EXISTS(SELECT 1 FROM vfile v1"
                         " WHERE v1.vid=%d AND v1.pathname=v2.pathname)"
-      " ORDER BY 1",
+      " ORDER BY 1 /*scan*/",
       rid, vid, rid, vid, vid, rid
     );
   }else{
-    blob_appendf(&sql,
+    blob_append_sql(&sql,
       "SELECT pathname, deleted, chnged , rid==0, rid, islink"
       "  FROM vfile"
       " WHERE vid=%d"
       "   AND (deleted OR chnged OR rid==0)"
-      " ORDER BY pathname",
+      " ORDER BY pathname /*scan*/",
       vid
     );
   }
-  db_prepare(&q, blob_str(&sql));
+  db_prepare(&q, "%s", blob_sql_text(&sql));
   while( db_step(&q)==SQLITE_ROW ){
     const char *zPathname = db_column_text(&q,0);
     int isDeleted = db_column_int(&q, 1);
@@ -371,13 +421,23 @@ static void diff_all_against_disk(
     int isNew = db_column_int(&q,3);
     int srcid = db_column_int(&q, 4);
     int isLink = db_column_int(&q, 5);
-    char *zFullName = mprintf("%s%s", g.zLocalRoot, zPathname);
-    char *zToFree = zFullName;
+    const char *zFullName;
     int showDiff = 1;
+    Blob fname;
+
+    if( !file_dir_match(pFileDir, zPathname) ) continue;
+    if( determine_exec_relative_option(0) ){
+      blob_zero(&fname);
+      file_relative_name(zPathname, &fname, 1);
+    }else{
+      blob_set(&fname, g.zLocalRoot);
+      blob_append(&fname, zPathname, -1);
+    }
+    zFullName = blob_str(&fname);
     if( isDeleted ){
       fossil_print("DELETED  %s\n", zPathname);
       if( !asNewFile ){ showDiff = 0; zFullName = NULL_DEVICE; }
-    }else if( file_access(zFullName, 0) ){
+    }else if( file_access(zFullName, F_OK) ){
       fossil_print("MISSING  %s\n", zPathname);
       if( !asNewFile ){ showDiff = 0; }
     }else if( isNew ){
@@ -388,6 +448,10 @@ static void diff_all_against_disk(
       fossil_print("ADDED_BY_MERGE %s\n", zPathname);
       srcid = 0;
       if( !asNewFile ){ showDiff = 0; }
+    }else if( isChnged==5 ){
+      fossil_print("ADDED_BY_INTEGRATE %s\n", zPathname);
+      srcid = 0;
+      if( !asNewFile ){ showDiff = 0; }
     }
     if( showDiff ){
       Blob content;
@@ -395,7 +459,7 @@ static void diff_all_against_disk(
       if( !isLink != !file_wd_islink(zFullName) ){
         diff_print_index(zPathname, diffFlags);
         diff_print_filenames(zPathname, zPathname, diffFlags);
-        fossil_print(DIFF_CANNOT_COMPUTE_SYMLINK);
+        fossil_print("%s",DIFF_CANNOT_COMPUTE_SYMLINK);
         continue;
       }
       if( srcid>0 ){
@@ -409,15 +473,14 @@ static void diff_all_against_disk(
                 zBinGlob, fIncludeBinary, diffFlags);
       blob_reset(&content);
     }
-    free(zToFree);
+    blob_reset(&fname);
   }
   db_finalize(&q);
   db_end_transaction(1);  /* ROLLBACK */
 }
 
 /*
-** Output the differences between two versions of a single file.
-** zFrom and zTo are the check-ins containing the two file versions.
+** Run a diff between the undo buffer and files on disk.
 **
 ** Use the internal diff logic if zDiffCmd is NULL.  Otherwise call the
 ** command zDiffCmd to do the diffing.
@@ -426,37 +489,29 @@ static void diff_all_against_disk(
 ** for file names to treat as binary.  If fIncludeBinary is zero, these files
 ** will be skipped in addition to files that may contain binary content.
 */
-static void diff_one_two_versions(
-  const char *zFrom,
-  const char *zTo,
-  const char *zDiffCmd,
-  const char *zBinGlob,
-  int fIncludeBinary,
-  u64 diffFlags,
-  const char *zFileTreeName
+static void diff_against_undo(
+  const char *zDiffCmd,     /* Use this diff command.  NULL for built-in */
+  const char *zBinGlob,     /* Treat file names matching this as binary */
+  int fIncludeBinary,       /* Treat file names matching this as binary */
+  u64 diffFlags,            /* Flags controlling diff output */
+  FileDirList *pFileDir     /* List of files and directories to diff */
 ){
-  char *zName;
-  Blob fname;
-  Blob v1, v2;
-  int isLink1, isLink2;
-  int isBin1, isBin2;
-  if( diffFlags & DIFF_BRIEF ) return;
-  file_tree_name(zFileTreeName, &fname, 1);
-  zName = blob_str(&fname);
-  historical_version_of_file(zFrom, zName, &v1, &isLink1, 0,
-                             fIncludeBinary ? 0 : &isBin1, 0);
-  historical_version_of_file(zTo, zName, &v2, &isLink2, 0,
-                             fIncludeBinary ? 0 : &isBin2, 0);
-  if( isLink1 != isLink2 ){
-    diff_print_filenames(zName, zName, diffFlags);
-    fossil_print(DIFF_CANNOT_COMPUTE_SYMLINK);
-  }else{
-    diff_file_mem(&v1, &v2, isBin1, isBin2, zName, zDiffCmd,
-                  zBinGlob, fIncludeBinary, diffFlags);
+  Stmt q;
+  Blob content;
+  db_prepare(&q, "SELECT pathname, content FROM undo");
+  blob_init(&content, 0, 0);
+  while( db_step(&q)==SQLITE_ROW ){
+    char *zFullName;
+    const char *zFile = (const char*)db_column_text(&q, 0);
+    if( !file_dir_match(pFileDir, zFile) ) continue;
+    zFullName = mprintf("%s%s", g.zLocalRoot, zFile);
+    db_column_blob(&q, 1, &content);
+    diff_file(&content, 0, zFullName, zFile,
+              zDiffCmd, zBinGlob, fIncludeBinary, diffFlags);
+    fossil_free(zFullName);
+    blob_reset(&content);
   }
-  blob_reset(&v1);
-  blob_reset(&v2);
-  blob_reset(&fname);
+  db_finalize(&q);
 }
 
 /*
@@ -481,7 +536,14 @@ static void diff_manifest_entry(
   Blob f1, f2;
   int isBin1, isBin2;
   int rid;
-  const char *zName =  pFrom ? pFrom->zName : pTo->zName;
+  const char *zName;
+  if( pFrom ){
+    zName = pFrom->zName;
+  }else if( pTo ){
+    zName = pTo->zName;
+  }else{
+    zName = DIFF_NO_NAME;
+  }
   if( diffFlags & DIFF_BRIEF ) return;
   diff_print_index(zName, diffFlags);
   if( pFrom ){
@@ -514,17 +576,18 @@ static void diff_manifest_entry(
 ** for file names to treat as binary.  If fIncludeBinary is zero, these files
 ** will be skipped in addition to files that may contain binary content.
 */
-static void diff_all_two_versions(
+static void diff_two_versions(
   const char *zFrom,
   const char *zTo,
   const char *zDiffCmd,
   const char *zBinGlob,
   int fIncludeBinary,
-  u64 diffFlags
+  u64 diffFlags,
+  FileDirList *pFileDir
 ){
   Manifest *pFrom, *pTo;
   ManifestFile *pFromFile, *pToFile;
-  int asNewFlag = (diffFlags & DIFF_NEWFILE)!=0 ? 1 : 0;
+  int asNewFlag = (diffFlags & DIFF_VERBOSE)!=0 ? 1 : 0;
 
   pFrom = manifest_get_by_name(zFrom, 0);
   manifest_file_rewind(pFrom);
@@ -543,29 +606,36 @@ static void diff_all_two_versions(
       cmp = fossil_strcmp(pFromFile->zName, pToFile->zName);
     }
     if( cmp<0 ){
-      fossil_print("DELETED %s\n", pFromFile->zName);
-      if( asNewFlag ){
-        diff_manifest_entry(pFromFile, 0, zDiffCmd, zBinGlob,
-                            fIncludeBinary, diffFlags);
+      if( file_dir_match(pFileDir, pFromFile->zName) ){
+        fossil_print("DELETED %s\n", pFromFile->zName);
+        if( asNewFlag ){
+          diff_manifest_entry(pFromFile, 0, zDiffCmd, zBinGlob,
+                              fIncludeBinary, diffFlags);
+        }
       }
       pFromFile = manifest_file_next(pFrom,0);
     }else if( cmp>0 ){
-      fossil_print("ADDED   %s\n", pToFile->zName);
-      if( asNewFlag ){
-        diff_manifest_entry(0, pToFile, zDiffCmd, zBinGlob,
-                            fIncludeBinary, diffFlags);
+      if( file_dir_match(pFileDir, pToFile->zName) ){
+        fossil_print("ADDED   %s\n", pToFile->zName);
+        if( asNewFlag ){
+          diff_manifest_entry(0, pToFile, zDiffCmd, zBinGlob,
+                              fIncludeBinary, diffFlags);
+        }
       }
       pToFile = manifest_file_next(pTo,0);
     }else if( fossil_strcmp(pFromFile->zUuid, pToFile->zUuid)==0 ){
       /* No changes */
+      (void)file_dir_match(pFileDir, pFromFile->zName); /* Record name usage */
       pFromFile = manifest_file_next(pFrom,0);
       pToFile = manifest_file_next(pTo,0);
     }else{
-      if( diffFlags & DIFF_BRIEF ){
-        fossil_print("CHANGED %s\n", pFromFile->zName);
-      }else{
-        diff_manifest_entry(pFromFile, pToFile, zDiffCmd, zBinGlob,
-                            fIncludeBinary, diffFlags);
+      if( file_dir_match(pFileDir, pToFile->zName) ){
+        if( diffFlags & DIFF_BRIEF ){
+          fossil_print("CHANGED %s\n", pFromFile->zName);
+        }else{
+          diff_manifest_entry(pFromFile, pToFile, zDiffCmd, zBinGlob,
+                              fIncludeBinary, diffFlags);
+        }
       }
       pFromFile = manifest_file_next(pFrom,0);
       pToFile = manifest_file_next(pTo,0);
@@ -580,7 +650,7 @@ static void diff_all_two_versions(
 ** no external diff command is defined.
 */
 const char *diff_command_external(int guiDiff){
-  char *zDefault;
+  const char *zDefault;
   const char *zName;
 
   if( guiDiff ){
@@ -597,79 +667,69 @@ const char *diff_command_external(int guiDiff){
   return db_get(zName, zDefault);
 }
 
-/* A Tcl/Tk script used to render diff output.
-*/
-static const char zDiffScript[] = 
-@ package require Tk
-@ wm withdraw .
-@ wm title . {Fossil Diff}
-@ wm iconname . {Fossil Diff}
-@ set body {}
-@ set mx 80          ;# Length of the longest line of text
-@ set nLine 0        ;# Number of lines of text
-@ text .t -width 180 -yscroll {.sb set}
-@ if {$tcl_platform(platform)=="windows"} {.t config -font {courier 9}}
-@ .t tag config ln -foreground gray
-@ .t tag config chng -background {#d0d0ff}
-@ .t tag config add -background {#c0ffc0}
-@ .t tag config rm -background {#ffc0c0}
-@ proc dehtml {x} {
-@   return [string map {&amp; & &lt; < &gt; > &#39; ' &quot; \"} $x]
-@ }
-@ # puts $cmd
-@ set in [open $cmd r]
-@ while {![eof $in]} {
-@   set line [gets $in]
-@   if {[regexp {^<a name="chunk.*"></a>} $line]} continue
-@   if {[regexp {^===} $line]} {
-@     set n [string length $line]
-@     if {$n>$mx} {set mx $n}
-@   }
-@   incr nLine
-@   while {[regexp {^(.*?)<span class="diff([a-z]+)">(.*?)</span>(.*)$} $line \
-@             all pre class mid tail]} {
-@     .t insert end [dehtml $pre] {} [dehtml $mid] $class
-@     set line $tail
-@   }
-@   .t insert end [dehtml $line]\n {}
-@ }
-@ close $in
-@ if {$mx>250} {set mx 250}      ;# Limit window width to 200 characters
-@ if {$nLine>55} {set nLine 55}  ;# Limit window height to 55 lines
-@ .t config -height $nLine -width $mx
-@ pack .t -side left -fill both -expand 1
-@ scrollbar .sb -command {.t yview} -orient vertical
-@ pack .sb -side left -fill y
-@ wm deiconify .
-;
-
 /*
 ** Show diff output in a Tcl/Tk window, in response to the --tk option
 ** to the diff command.
-** 
-** Steps:
+**
+** If fossil has direct access to a Tcl interpreter (either loaded
+** dynamically through stubs or linked in statically), we can use it
+** directly. Otherwise:
 ** (1) Write the Tcl/Tk script used for rendering into a temp file.
-** (2) Invoke "wish" on the temp file using fossil_system().
+** (2) Invoke "tclsh" on the temp file using fossil_system().
 ** (3) Delete the temp file.
 */
 void diff_tk(const char *zSubCmd, int firstArg){
   int i;
   Blob script;
-  char *zTempFile;
+  const char *zTempFile = 0;
   char *zCmd;
   blob_zero(&script);
-  blob_appendf(&script, "set cmd {| \"%/\" %s --html -y -i",
+  blob_appendf(&script, "set fossilcmd {| \"%/\" %s --html -y -i -v",
                g.nameOfExe, zSubCmd);
+  find_option("html",0,0);
+  find_option("side-by-side","y",0);
+  find_option("internal","i",0);
+  find_option("verbose","v",0);
+  /* The undocumented --script FILENAME option causes the Tk script to
+  ** be written into the FILENAME instead of being run.  This is used
+  ** for testing and debugging. */
+  zTempFile = find_option("script",0,1);
   for(i=firstArg; i<g.argc; i++){
-    blob_append(&script, " ", 1);
-    shell_escape(&script, g.argv[i]);
+    const char *z = g.argv[i];
+    if( sqlite3_strglob("*}*",z) ){
+      blob_appendf(&script, " {%/}", z);
+    }else{
+      int j;
+      blob_append(&script, " ", 1);
+      for(j=0; z[j]; j++) blob_appendf(&script, "\\%03o", (unsigned char)z[j]);
+    }
   }
-  blob_appendf(&script, "}\n%s", zDiffScript);
-  zTempFile = write_blob_to_temp_file(&script);
-  zCmd = mprintf("tclsh \"%s\"", zTempFile);
-  fossil_system(zCmd);
-  file_delete(zTempFile);
-  fossil_free(zCmd);
+  blob_appendf(&script, "}\n%s", builtin_file("diff.tcl", 0));
+  if( zTempFile ){
+    blob_write_to_file(&script, zTempFile);
+    fossil_print("To see diff, run: tclsh \"%s\"\n", zTempFile);
+  }else{
+#if defined(FOSSIL_ENABLE_TCL)
+    Th_FossilInit(TH_INIT_DEFAULT);
+    if( evaluateTclWithEvents(g.interp, &g.tcl, blob_str(&script),
+                              blob_size(&script), 1, 1, 0)==TCL_OK ){
+      blob_reset(&script);
+      return;
+    }
+    /*
+     * If evaluation of the Tcl script fails, the reason may be that Tk
+     * could not be found by the loaded Tcl, or that Tcl cannot be loaded
+     * dynamically (e.g. x64 Tcl with x86 Fossil).  Therefore, fallback
+     * to using the external "tclsh", if available.
+     */
+#endif
+    zTempFile = write_blob_to_temp_file(&script);
+    zCmd = mprintf("tclsh \"%s\"", zTempFile);
+    fossil_system(zCmd);
+    file_delete(zTempFile);
+    fossil_free(zCmd);
+  }
+  blob_reset(&script);
 }
 
 /*
@@ -704,11 +764,11 @@ const char *diff_get_binary_glob(void){
 **
 ** Show the difference between the current version of each of the FILEs
 ** specified (as they exist on disk) and that same file as it was checked
-** out.  Or if the FILE arguments are omitted, show the unsaved changed
+** out.  Or if the FILE arguments are omitted, show the unsaved changes
 ** currently in the working check-out.
 **
 ** If the "--from VERSION" or "-r VERSION" option is used it specifies
-** the source check-in for the diff operation.  If not specified, the 
+** the source check-in for the diff operation.  If not specified, the
 ** source check-in is the base check-in for the current check-out.
 **
 ** If the "--to VERSION" option appears, it specifies the check-in from
@@ -716,6 +776,9 @@ const char *diff_get_binary_glob(void){
 ** no "--to" option then the (possibly edited) files in the current check-out
 ** are used.
 **
+** The "--checkin VERSION" option shows the changes made by
+** check-in VERSION relative to its primary parent.
+**
 ** The "-i" command-line option forces the use of the internal diff logic
 ** rather than any external diff program that might be configured using
 ** the "setting" command.  If no external diff program is configured, then
@@ -732,32 +795,41 @@ const char *diff_get_binary_glob(void){
 ** This option overrides the "binary-glob" setting.
 **
 ** Options:
-**   --branch BRANCH     Show diff of all changes on BRANCH
-**   --brief             Show filenames only
-**   --context|-c N      Use N lines of context 
-**   --from|-r VERSION   select VERSION as source for the diff
-**   -i                  use internal diff logic
-**   --new-file|-N       output complete text of added or deleted files
-**   --tk                Launch a Tcl/Tk GUI for display
-**   --to VERSION        select VERSION as target for the diff
-**   --side-by-side|-y   side-by-side diff
-**   --unified           unified diff
-**   --width|-W N        Width of lines in side-by-side diff 
-**   --diff-binary BOOL  Include binary files when using external commands
-**   --binary PATTERN    Treat files that match the glob PATTERN as binary
+**   --binary PATTERN           Treat files that match the glob PATTERN as binary
+**   --branch BRANCH            Show diff of all changes on BRANCH
+**   --brief                    Show filenames only
+**   --checkin VERSION          Show diff of all changes in VERSION
+**   --context|-c N             Use N lines of context
+**   --diff-binary BOOL         Include binary files when using external commands
+**   --exec-abs-paths           Force absolute path names with external commands.
+**   --exec-rel-paths           Force relative path names with external commands.
+**   --from|-r VERSION          Select VERSION as source for the diff
+**   --internal|-i              Use internal diff logic
+**   --side-by-side|-y          Side-by-side diff
+**   --strip-trailing-cr        Strip trailing CR
+**   --tk                       Launch a Tcl/Tk GUI for display
+**   --to VERSION               Select VERSION as target for the diff
+**   --undo                     Diff against the "undo" buffer
+**   --unified                  Unified diff
+**   -v|--verbose               Output complete text of added or deleted files
+**   -w|--ignore-all-space      Ignore white space when comparing lines
+**   -W|--width <num>           Width of lines in side-by-side diff
+**   -Z|--ignore-trailing-space Ignore changes to end-of-line whitespace
 */
 void diff_cmd(void){
   int isGDiff;               /* True for gdiff.  False for normal diff */
   int isInternDiff;          /* True for internal diff */
-  int hasNFlag;              /* True if -N or --new-file flag is used */
+  int verboseFlag;           /* True if -v or --verbose flag is used */
   const char *zFrom;         /* Source version number */
   const char *zTo;           /* Target version number */
+  const char *zCheckin;      /* Check-in version number */
   const char *zBranch;       /* Branch to diff */
   const char *zDiffCmd = 0;  /* External diff command. NULL for internal diff */
   const char *zBinGlob = 0;  /* Treat file names matching this as binary */
   int fIncludeBinary = 0;    /* Include binary files for external diff */
+  int againstUndo = 0;       /* Diff against files in the undo buffer */
   u64 diffFlags = 0;         /* Flags to control the DIFF */
-  int f;
+  FileDirList *pFileDir = 0; /* Restrict the diff to these files */
 
   if( find_option("tk",0,0)!=0 ){
     diff_tk("diff", 2);
@@ -767,68 +839,114 @@ void diff_cmd(void){
   isInternDiff = find_option("internal","i",0)!=0;
   zFrom = find_option("from", "r", 1);
   zTo = find_option("to", 0, 1);
+  zCheckin = find_option("checkin", 0, 1);
   zBranch = find_option("branch", 0, 1);
+  againstUndo = find_option("undo",0,0)!=0;
   diffFlags = diff_options();
-  hasNFlag = find_option("new-file","N",0)!=0;
-  if( hasNFlag ) diffFlags |= DIFF_NEWFILE;
-
+  verboseFlag = find_option("verbose","v",0)!=0;
+  if( !verboseFlag ){
+    verboseFlag = find_option("new-file","N",0)!=0; /* deprecated */
+  }
+  if( verboseFlag ) diffFlags |= DIFF_VERBOSE;
+  if( againstUndo && ( zFrom!=0 || zTo!=0 || zCheckin!=0 || zBranch!=0) ){
+    fossil_fatal("cannot use --undo together with --from, --to, --checkin,"
+                 " or --branch");
+  }
   if( zBranch ){
-    if( zTo || zFrom ){
-      fossil_fatal("cannot use --from or --to with --branch");
+    if( zTo || zFrom || zCheckin ){
+      fossil_fatal("cannot use --from, --to, or --checkin with --branch");
     }
     zTo = zBranch;
     zFrom = mprintf("root:%s", zBranch);
   }
-  if( zTo==0 ){
+  if( zCheckin!=0 && ( zFrom!=0 || zTo!=0 ) ){
+    fossil_fatal("cannot use --checkin together with --from or --to");
+  }
+  if( zTo==0 || againstUndo ){
     db_must_be_within_tree();
-    if( !isInternDiff ){
-      zDiffCmd = diff_command_external(isGDiff);
-    }
-    zBinGlob = diff_get_binary_glob();
-    fIncludeBinary = diff_include_binary_files();
-    verify_all_options();
-    if( g.argc>=3 ){
-      for(f=2; f<g.argc; ++f){
-        diff_one_against_disk(zFrom, zDiffCmd, zBinGlob, fIncludeBinary,
-                              diffFlags, g.argv[f]);
-      }
-    }else{
-      diff_all_against_disk(zFrom, zDiffCmd, zBinGlob, fIncludeBinary,
-                            diffFlags);
-    }
   }else if( zFrom==0 ){
     fossil_fatal("must use --from if --to is present");
   }else{
     db_find_and_open_repository(0, 0);
-    if( !isInternDiff ){
-      zDiffCmd = diff_command_external(isGDiff);
-    }
-    zBinGlob = diff_get_binary_glob();
-    fIncludeBinary = diff_include_binary_files();
-    verify_all_options();
-    if( g.argc>=3 ){
-      for(f=2; f<g.argc; ++f){
-        diff_one_two_versions(zFrom, zTo, zDiffCmd, zBinGlob, fIncludeBinary,
-                              diffFlags, g.argv[f]);
+  }
+  if( !isInternDiff ){
+    zDiffCmd = diff_command_external(isGDiff);
+  }
+  zBinGlob = diff_get_binary_glob();
+  fIncludeBinary = diff_include_binary_files();
+  determine_exec_relative_option(1);
+  verify_all_options();
+  if( g.argc>=3 ){
+    int i;
+    Blob fname;
+    pFileDir = fossil_malloc( sizeof(*pFileDir) * (g.argc-1) );
+    memset(pFileDir, 0, sizeof(*pFileDir) * (g.argc-1));
+    for(i=2; i<g.argc; i++){
+      file_tree_name(g.argv[i], &fname, 0, 1);
+      pFileDir[i-2].zName = fossil_strdup(blob_str(&fname));
+      if( strcmp(pFileDir[i-2].zName,".")==0 ){
+        pFileDir[0].zName[0] = '.';
+        pFileDir[0].zName[1] = 0;
+        break;
       }
-    }else{
-      diff_all_two_versions(zFrom, zTo, zDiffCmd, zBinGlob, fIncludeBinary,
-                            diffFlags);
+      pFileDir[i-2].nName = blob_size(&fname);
+      pFileDir[i-2].nUsed = 0;
+      blob_reset(&fname);
     }
   }
+  if ( zCheckin!=0 ){
+    int ridTo = name_to_typed_rid(zCheckin, "ci");
+    zTo = zCheckin;
+    zFrom = db_text(0,
+      "SELECT uuid FROM blob, plink"
+      " WHERE plink.cid=%d AND plink.isprim AND plink.pid=blob.rid",
+      ridTo);
+    if( zFrom==0 ){
+      fossil_fatal("check-in %s has no parent", zTo);
+    }
+  }
+  if( againstUndo ){
+    if( db_lget_int("undo_available",0)==0 ){
+      fossil_print("No undo or redo is available\n");
+      return;
+    }
+    diff_against_undo(zDiffCmd, zBinGlob, fIncludeBinary,
+                      diffFlags, pFileDir);
+  }else if( zTo==0 ){
+    diff_against_disk(zFrom, zDiffCmd, zBinGlob, fIncludeBinary,
+                      diffFlags, pFileDir);
+  }else{
+    diff_two_versions(zFrom, zTo, zDiffCmd, zBinGlob, fIncludeBinary,
+                      diffFlags, pFileDir);
+  }
+  if( pFileDir ){
+    int i;
+    for(i=0; pFileDir[i].zName; i++){
+      if( pFileDir[i].nUsed==0
+       && strcmp(pFileDir[0].zName,".")!=0
+       && !file_wd_isdir(g.argv[i+2])
+      ){
+        fossil_fatal("not found: '%s'", g.argv[i+2]);
+      }
+      fossil_free(pFileDir[i].zName);
+    }
+    fossil_free(pFileDir);
+  }
 }
 
 /*
 ** WEBPAGE: vpatch
-** URL vpatch?from=UUID&to=UUID
+** URL: /vpatch?from=FROM&to=TO
+**
+** Show a patch that goes from check-in FROM to check-in TO.
 */
 void vpatch_page(void){
   const char *zFrom = P("from");
   const char *zTo = P("to");
   login_check_credentials();
-  if( !g.perm.Read ){ login_needed(); return; }
+  if( !g.perm.Read ){ login_needed(g.anon.Read); return; }
   if( zFrom==0 || zTo==0 ) fossil_redirect_home();
 
   cgi_set_content_type("text/plain");
-  diff_all_two_versions(zFrom, zTo, 0, 0, 0, DIFF_NEWFILE);
+  diff_two_versions(zFrom, zTo, 0, 0, 0, DIFF_VERBOSE, 0);
 }
