@@ -26,6 +26,7 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <unistd.h>
+#include <stdio.h>
 #include <string.h>
 #include <errno.h>
 #include "file.h"
@@ -37,18 +38,41 @@
 # include <direct.h>
 # include <windows.h>
 # include <sys/utime.h>
+#else
+# include <sys/time.h>
 #endif
 
-/*
-** The file status information from the most recent stat() call.
-**
-** Use _stati64 rather than stat on windows, in order to handle files
-** larger than 2GB.
-*/
+#if INTERFACE
+
+#include <dirent.h>
+#if defined(_WIN32)
+# define DIR _WDIR
+# define dirent _wdirent
+# define opendir _wopendir
+# define readdir _wreaddir
+# define closedir _wclosedir
+#endif /* _WIN32 */
+
 #if defined(_WIN32) && (defined(__MSVCRT__) || defined(_MSC_VER))
-# undef stat
-# define stat _stati64
+struct fossilStat {
+    i64 st_size;
+    i64 st_mtime;
+    int st_mode;
+};
 #endif
+
+#if defined(_WIN32) || defined(__CYGWIN__)
+# define fossil_isdirsep(a)    (((a) == '/') || ((a) == '\\'))
+#else
+# define fossil_isdirsep(a)    ((a) == '/')
+#endif
+
+#endif /* INTERFACE */
+
+#if !defined(_WIN32) || !(defined(__MSVCRT__) || defined(_MSC_VER))
+# define fossilStat stat
+#endif
+
 /*
 ** On Windows S_ISLNK always returns FALSE.
 */
@@ -56,27 +80,27 @@
 # define S_ISLNK(x) (0)
 #endif
 static int fileStatValid = 0;
-static struct stat fileStat;
+static struct fossilStat fileStat;
 
 /*
 ** Fill stat buf with information received from stat() or lstat().
 ** lstat() is called on Unix if isWd is TRUE and allow-symlinks setting is on.
 **
 */
-static int fossil_stat(const char *zFilename, struct stat *buf, int isWd){
+static int fossil_stat(const char *zFilename, struct fossilStat *buf, int isWd){
+  int rc;
+  void *zMbcs = fossil_utf8_to_path(zFilename, 0);
 #if !defined(_WIN32)
   if( isWd && g.allowSymlinks ){
-    return lstat(zFilename, buf);
+    rc = lstat(zMbcs, buf);
   }else{
-    return stat(zFilename, buf);
+    rc = stat(zMbcs, buf);
   }
 #else
-  int rc = 0;
-  wchar_t *zMbcs = fossil_utf8_to_unicode(zFilename);
-  rc = _wstati64(zMbcs, buf);
-  fossil_mbcs_free(zMbcs);
-  return rc;
+  rc = win32_stat(zMbcs, buf, isWd);
 #endif
+  fossil_path_free(zMbcs);
+  return rc;
 }
 
 /*
@@ -191,11 +215,10 @@ void symlink_create(const char *zTargetFile, const char *zLinkFile){
         zName[i] = '/';
       }
     }
-    if( zName!=zBuf ) free(zName);
-
     if( symlink(zTargetFile, zName)!=0 ){
       fossil_fatal_recursive("unable to create symlink \"%s\"", zName);
     }
+    if( zName!=zBuf ) free(zName);
   }else
 #endif
   {
@@ -218,29 +241,20 @@ void symlink_copy(const char *zFrom, const char *zTo){
 
 /*
 ** Return file permissions (normal, executable, or symlink):
-**   - PERM_EXE if file is executable;
+**   - PERM_EXE on Unix if file is executable;
 **   - PERM_LNK on Unix if file is symlink and allow-symlinks option is on;
 **   - PERM_REG for all other cases (regular file, directory, fifo, etc).
 */
 int file_wd_perm(const char *zFilename){
-  if( getStat(zFilename, 1) ) return PERM_REG;
-#if defined(_WIN32)
-#  ifndef S_IXUSR
-#    define S_IXUSR  _S_IEXEC
-#  endif
-  if( S_ISREG(fileStat.st_mode) && ((S_IXUSR)&fileStat.st_mode)!=0 )
-    return PERM_EXE;
-  else
-    return PERM_REG;
-#else
-  if( S_ISREG(fileStat.st_mode) &&
-      ((S_IXUSR|S_IXGRP|S_IXOTH)&fileStat.st_mode)!=0 )
-    return PERM_EXE;
-  else if( g.allowSymlinks && S_ISLNK(fileStat.st_mode) )
-    return PERM_LNK;
-  else
-    return PERM_REG;
+#if !defined(_WIN32)
+  if( !getStat(zFilename, 1) ){
+     if( S_ISREG(fileStat.st_mode) && ((S_IXUSR)&fileStat.st_mode)!=0 )
+      return PERM_EXE;
+    else if( g.allowSymlinks && S_ISLNK(fileStat.st_mode) )
+      return PERM_LNK;
+  }
 #endif
+  return PERM_REG;
 }
 
 /*
@@ -281,20 +295,32 @@ int file_isdir(const char *zFilename){
 }
 
 /*
-** Same as file_isdir(), but takes into account symlinks.
+** Same as file_isdir(), but takes into account symlinks.  Return 1 if
+** zFilename is a directory -OR- a symlink that points to a directory.
+** Return 0 if zFilename does not exist.  Return 2 if zFilename exists
+** but is something other than a directory.
 */
 int file_wd_isdir(const char *zFilename){
   int rc;
+  char *zFN;
 
-  if( zFilename ){
-    char *zFN = mprintf("%s", zFilename);
-    file_simplify_name(zFN, -1, 0);
-    rc = getStat(zFN, 1);
-    free(zFN);
+  zFN = mprintf("%s", zFilename);
+  file_simplify_name(zFN, -1, 0);
+  rc = getStat(zFN, 1);
+  if( rc ){
+    rc = 0; /* It does not exist at all. */
+  }else if( S_ISDIR(fileStat.st_mode) ){
+    rc = 1; /* It exists and is a real directory. */
+  }else if( S_ISLNK(fileStat.st_mode) ){
+    Blob content;
+    blob_read_link(&content, zFN); /* It exists and is a link. */
+    rc = file_wd_isdir(blob_str(&content)); /* Points to directory? */
+    blob_reset(&content);
   }else{
-    rc = getStat(0, 1);
+    rc = 2; /* It exists and is something else. */
   }
-  return rc ? 0 : (S_ISDIR(fileStat.st_mode) ? 1 : 2);
+  free(zFN);
+  return rc;
 }
 
 
@@ -302,13 +328,35 @@ int file_wd_isdir(const char *zFilename){
 ** Wrapper around the access() system call.
 */
 int file_access(const char *zFilename, int flags){
+  int rc;
+  void *zMbcs = fossil_utf8_to_path(zFilename, 0);
 #ifdef _WIN32
-  wchar_t *zMbcs = fossil_utf8_to_unicode(zFilename);
-  int rc = _waccess(zMbcs, flags);
-  fossil_mbcs_free(zMbcs);
+  rc = win32_access(zMbcs, flags);
 #else
-  int rc = access(zFilename, flags);
+  rc = access(zMbcs, flags);
 #endif
+  fossil_path_free(zMbcs);
+  return rc;
+}
+
+/*
+** Wrapper around the chdir() system call.
+** If bChroot=1, do a chroot to this dir as well
+** (UNIX only)
+*/
+int file_chdir(const char *zChDir, int bChroot){
+  int rc;
+  void *zPath = fossil_utf8_to_path(zChDir, 1);
+#ifdef _WIN32
+  rc = win32_chdir(zPath, bChroot);
+#else
+  rc = chdir(zPath);
+  if( !rc && bChroot ){
+    rc = chroot(zPath);
+    if( !rc ) rc = chdir("/");
+  }
+#endif
+  fossil_path_free(zPath);
   return rc;
 }
 
@@ -343,13 +391,54 @@ char *file_newname(const char *zBase, const char *zSuffix, int relFlag){
 */
 const char *file_tail(const char *z){
   const char *zTail = z;
+  if( !zTail ) return 0;
   while( z[0] ){
-    if( z[0]=='/' ) zTail = &z[1];
+    if( fossil_isdirsep(z[0]) ) zTail = &z[1];
     z++;
   }
   return zTail;
 }
 
+/*
+** Return the directory of a file path name.  The directory is all components
+** except the last one.  For example, the directory of "/a/b/c.d" is "/a/b".
+** If there is no directory, NULL is returned; otherwise, the returned memory
+** should be freed via fossil_free().
+*/
+char *file_dirname(const char *z){
+  const char *zTail = file_tail(z);
+  if( zTail && zTail!=z ){
+    return mprintf("%.*s", (int)(zTail-z-1), z);
+  }else{
+    return 0;
+  }
+}
+
+/*
+** Rename a file or directory.
+** Returns zero upon success.
+*/
+int file_rename(
+  const char *zFrom,
+  const char *zTo,
+  int isFromDir,
+  int isToDir
+){
+  int rc;
+#if defined(_WIN32)
+  wchar_t *zMbcsFrom = fossil_utf8_to_path(zFrom, isFromDir);
+  wchar_t *zMbcsTo = fossil_utf8_to_path(zTo, isToDir);
+  rc = _wrename(zMbcsFrom, zMbcsTo);
+#else
+  char *zMbcsFrom = fossil_utf8_to_path(zFrom, isFromDir);
+  char *zMbcsTo = fossil_utf8_to_path(zTo, isToDir);
+  rc = rename(zMbcsFrom, zMbcsTo);
+#endif
+  fossil_path_free(zMbcsTo);
+  fossil_path_free(zMbcsFrom);
+  return rc;
+}
+
 /*
 ** Copy the content of a file from one place to another.
 */
@@ -359,6 +448,7 @@ void file_copy(const char *zFrom, const char *zTo){
   char zBuf[8192];
   in = fossil_fopen(zFrom, "rb");
   if( in==0 ) fossil_fatal("cannot open \"%s\" for reading", zFrom);
+  file_mkfolder(zTo, 0, 0);
   out = fossil_fopen(zTo, "wb");
   if( out==0 ) fossil_fatal("cannot open \"%s\" for writing", zTo);
   while( (got=fread(zBuf, 1, sizeof(zBuf), in))>0 ){
@@ -368,6 +458,22 @@ void file_copy(const char *zFrom, const char *zTo){
   fclose(out);
 }
 
+/*
+** COMMAND: test-file-copy
+**
+** Usage: %fossil test-file-copy SOURCE DESTINATION
+**
+** Make a copy of the file at SOURCE into a new name DESTINATION.  Any
+** directories in the path leading up to DESTINATION that do not already
+** exist are created automatically.
+*/
+void test_file_copy(void){
+  if( g.argc!=4 ){
+    fossil_fatal("Usage: %s test-file-copy SOURCE DESTINATION", g.argv[0]);
+  }
+  file_copy(g.argv[2], g.argv[3]);
+}
+
 /*
 ** Set or clear the execute bit on a file.  Return true if a change
 ** occurred and false if this routine is a no-op.
@@ -379,12 +485,12 @@ int file_wd_setexe(const char *zFilename, int onoff){
   if( fossil_stat(zFilename, &buf, 1)!=0 || S_ISLNK(buf.st_mode) ) return 0;
   if( onoff ){
     int targetMode = (buf.st_mode & 0444)>>2;
-    if( (buf.st_mode & 0111)!=targetMode ){
+    if( (buf.st_mode & 0100)==0 ){
       chmod(zFilename, buf.st_mode | targetMode);
       rc = 1;
     }
   }else{
-    if( (buf.st_mode & 0111)!=0 ){
+    if( (buf.st_mode & 0100)!=0 ){
       chmod(zFilename, buf.st_mode & ~0111);
       rc = 1;
     }
@@ -398,19 +504,21 @@ int file_wd_setexe(const char *zFilename, int onoff){
 */
 void file_set_mtime(const char *zFilename, i64 newMTime){
 #if !defined(_WIN32)
+  char *zMbcs;
   struct timeval tv[2];
   memset(tv, 0, sizeof(tv[0])*2);
   tv[0].tv_sec = newMTime;
   tv[1].tv_sec = newMTime;
-  utimes(zFilename, tv);
+  zMbcs = fossil_utf8_to_path(zFilename, 0);
+  utimes(zMbcs, tv);
 #else
   struct _utimbuf tb;
-  wchar_t *zMbcs = fossil_utf8_to_unicode(zFilename);
+  wchar_t *zMbcs = fossil_utf8_to_path(zFilename, 0);
   tb.actime = newMTime;
   tb.modtime = newMTime;
   _wutime(zMbcs, &tb);
-  fossil_mbcs_free(zMbcs);
 #endif
+  fossil_path_free(zMbcs);
 }
 
 /*
@@ -425,7 +533,7 @@ void test_set_mtime(void){
   char *zDate;
   i64 iMTime;
   if( g.argc!=4 ){
-    usage("test-set-mtime FILENAME DATE/TIME");
+    usage("FILENAME DATE/TIME");
   }
   db_open_or_attach(":memory:", "mem");
   iMTime = db_int64(0, "SELECT strftime('%%s',%Q)", g.argv[3]);
@@ -438,15 +546,20 @@ void test_set_mtime(void){
 
 /*
 ** Delete a file.
+**
+** Returns zero upon success.
 */
-void file_delete(const char *zFilename){
+int file_delete(const char *zFilename){
+  int rc;
 #ifdef _WIN32
-  wchar_t *z = fossil_utf8_to_unicode(zFilename);
-  _wunlink(z);
-  fossil_mbcs_free(z);
+  wchar_t *z = fossil_utf8_to_path(zFilename, 0);
+  rc = _wunlink(z);
 #else
-  unlink(zFilename);
+  char *z = fossil_utf8_to_path(zFilename, 0);
+  rc = unlink(zFilename);
 #endif
+  fossil_path_free(z);
+  return rc;
 }
 
 /*
@@ -464,14 +577,79 @@ int file_mkdir(const char *zName, int forceFlag){
   }
   if( rc!=1 ){
 #if defined(_WIN32)
-    int rc;
-    wchar_t *zMbcs = fossil_utf8_to_unicode(zName);
+    wchar_t *zMbcs = fossil_utf8_to_path(zName, 1);
     rc = _wmkdir(zMbcs);
-    fossil_mbcs_free(zMbcs);
-    return rc;
 #else
-    return mkdir(zName, 0755);
+    char *zMbcs = fossil_utf8_to_path(zName, 1);
+    rc = mkdir(zName, 0755);
 #endif
+    fossil_path_free(zMbcs);
+    return rc;
+  }
+  return 0;
+}
+
+/*
+** Create the tree of directories in which zFilename belongs, if that sequence
+** of directories does not already exist.
+**
+** On success, return zero.  On error, return errorReturn if positive, otherwise
+** print an error message and abort.
+*/
+int file_mkfolder(const char *zFilename, int forceFlag, int errorReturn){
+  int i, nName, rc = 0;
+  char *zName;
+
+  nName = strlen(zFilename);
+  zName = mprintf("%s", zFilename);
+  nName = file_simplify_name(zName, nName, 0);
+  for(i=1; i<nName; i++){
+    if( zName[i]=='/' ){
+      zName[i] = 0;
+#if defined(_WIN32) || defined(__CYGWIN__)
+      /*
+      ** On Windows, local path looks like: C:/develop/project/file.txt
+      ** The if stops us from trying to create a directory of a drive letter
+      ** C: in this example.
+      */
+      if( !(i==2 && zName[1]==':') ){
+#endif
+        if( file_mkdir(zName, forceFlag) && file_wd_isdir(zName)!=1 ){
+          if (errorReturn <= 0) {
+            fossil_fatal_recursive("unable to create directory %s", zName);
+          }
+          rc = errorReturn;
+          break;
+        }
+#if defined(_WIN32) || defined(__CYGWIN__)
+      }
+#endif
+      zName[i] = '/';
+    }
+  }
+  free(zName);
+  return rc;
+}
+
+/*
+** Removes the directory named in the argument, if it exists.  The directory
+** must be empty and cannot be the current directory or the root directory.
+**
+** Returns zero upon success.
+*/
+int file_rmdir(const char *zName){
+  int rc = file_wd_isdir(zName);
+  if( rc==2 ) return 1; /* cannot remove normal file */
+  if( rc==1 ){
+#if defined(_WIN32)
+    wchar_t *zMbcs = fossil_utf8_to_path(zName, 1);
+    rc = _wrmdir(zMbcs);
+#else
+    char *zMbcs = fossil_utf8_to_path(zName, 1);
+    rc = rmdir(zName);
+#endif
+    fossil_path_free(zMbcs);
+    return rc;
   }
   return 0;
 }
@@ -483,21 +661,71 @@ int file_mkdir(const char *zName, int forceFlag){
 **
 **     *  Does not begin with "/"
 **     *  Does not contain any path element named "." or ".."
-**     *  Does not contain any of these characters in the path: "\*[]?"
+**     *  Does not contain any of these characters in the path: "\"
 **     *  Does not end with "/".
 **     *  Does not contain two or more "/" characters in a row.
 **     *  Contains at least one character
+**
+** Invalid UTF8 characters result in a false return if bStrictUtf8 is
+** true.  If bStrictUtf8 is false, invalid UTF8 characters are silently
+** ignored. See http://en.wikipedia.org/wiki/UTF-8#Invalid_byte_sequences
+** and http://en.wikipedia.org/wiki/Unicode (for the noncharacters)
+**
+** The bStrictUtf8 flag is true for new inputs, but is false when parsing
+** legacy manifests, for backwards compatibility.
 */
-int file_is_simple_pathname(const char *z){
+int file_is_simple_pathname(const char *z, int bStrictUtf8){
   int i;
-  char c = z[0];
+  unsigned char c = (unsigned char) z[0];
+  char maskNonAscii = bStrictUtf8 ? 0x80 : 0x00;
   if( c=='/' || c==0 ) return 0;
   if( c=='.' ){
     if( z[1]=='/' || z[1]==0 ) return 0;
     if( z[1]=='.' && (z[2]=='/' || z[2]==0) ) return 0;
   }
-  for(i=0; (c=z[i])!=0; i++){
-    if( c=='\\' || c=='*' || c=='[' || c==']' || c=='?' ){
+  for(i=0; (c=(unsigned char)z[i])!=0; i++){
+    if( c & maskNonAscii ){
+      if( (z[++i]&0xc0)!=0x80 ){
+        /* Invalid first continuation byte */
+        return 0;
+      }
+      if( c<0xc2 ){
+        /* Invalid 1-byte UTF-8 sequence, or 2-byte overlong form. */
+        return 0;
+      }else if( (c&0xe0)==0xe0 ){
+        /* 3-byte or more */
+        int unicode;
+        if( c&0x10 ){
+          /* Unicode characters > U+FFFF are not supported.
+           * Windows XP and earlier cannot handle them.
+           */
+          return 0;
+        }
+        /* This is a 3-byte UTF-8 character */
+        unicode = ((c&0x0f)<<12) + ((z[i]&0x3f)<<6) + (z[i+1]&0x3f);
+        if( unicode <= 0x07ff ){
+          /* overlong form */
+          return 0;
+        }else if( unicode>=0xe000 ){
+          /* U+E000..U+FFFF */
+          if( (unicode<=0xf8ff) || (unicode>=0xfffe) ){
+            /* U+E000..U+F8FF are for private use.
+             * U+FFFE..U+FFFF are noncharacters. */
+            return 0;
+          } else if( (unicode>=0xfdd0) && (unicode<=0xfdef) ){
+            /* U+FDD0..U+FDEF are noncharacters. */
+            return 0;
+          }
+        }else if( (unicode>=0xd800) && (unicode<=0xdfff) ){
+          /* U+D800..U+DFFF are for surrogate pairs. */
+          return 0;
+        }
+        if( (z[++i]&0xc0)!=0x80 ){
+          /* Invalid second continuation byte */
+          return 0;
+        }
+      }
+    }else if( bStrictUtf8 && (c=='\\') ){
       return 0;
     }
     if( c=='/' ){
@@ -530,7 +758,8 @@ static int backup_dir(const char *z, int *pJ){
 /*
 ** Simplify a filename by
 **
-**  * Convert all \ into / on windows
+**  * Remove extended path prefix on windows and cygwin
+**  * Convert all \ into / on windows and cygwin
 **  * removing any trailing and duplicate /
 **  * removing /./
 **  * removing /A/../
@@ -540,24 +769,34 @@ static int backup_dir(const char *z, int *pJ){
 ** is retained.
 */
 int file_simplify_name(char *z, int n, int slash){
-  int i, j;
+  int i = 1, j;
   if( n<0 ) n = strlen(z);
 
-  /* On windows convert all \ characters to / */
-#if defined(_WIN32)
-  for(i=0; i<n; i++){
-    if( z[i]=='\\' ) z[i] = '/';
+  /* On windows and cygwin convert all \ characters to /
+   * and remove extended path prefix if present */
+#if defined(_WIN32) || defined(__CYGWIN__)
+  for(j=0; j<n; j++){
+    if( z[j]=='\\' ) z[j] = '/';
+  }
+  if( n>3 && !memcmp(z, "//?/", 4) ){
+    if( fossil_strnicmp(z+4,"UNC", 3) ){
+      i += 4;
+      z[0] = z[4];
+    }else{
+      i += 6;
+      z[0] = '/';
+    }
   }
 #endif
 
   /* Removing trailing "/" characters */
-  if ( !slash ){
+  if( !slash ){
     while( n>1 && z[n-1]=='/' ){ n--; }
   }
 
   /* Remove duplicate '/' characters.  Except, two // at the beginning
   ** of a pathname is allowed since this is important on windows. */
-  for(i=j=1; i<n; i++){
+  for(j=1; i<n; i++){
     z[j++] = z[i];
     while( z[i]=='/' && i<n-1 && z[i+1]=='/' ) i++;
   }
@@ -589,7 +828,7 @@ int file_simplify_name(char *z, int n, int slash){
     if( j>=0 ) z[j] = z[i];
     j++;
   }
-  if( j==0 ) z[j++] = '.';
+  if( j==0 ) z[j++] = '/';
   z[j] = 0;
   return j;
 }
@@ -597,7 +836,7 @@ int file_simplify_name(char *z, int n, int slash){
 /*
 ** COMMAND: test-simplify-name
 **
-** %fossil test-simplify-name FILENAME...
+** Usage: %fossil test-simplify-name FILENAME...
 **
 ** Print the simplified versions of each FILENAME.
 */
@@ -622,25 +861,11 @@ void cmd_test_simplify_name(void){
 */
 void file_getcwd(char *zBuf, int nBuf){
 #ifdef _WIN32
-  char *zPwdUtf8;
-  int nPwd;
-  int i;
-  wchar_t zPwd[2000];
-  if( _wgetcwd(zPwd, sizeof(zPwd)/sizeof(zPwd[0])-1)==0 ){
-    fossil_fatal("cannot find the current working directory.");
-  }
-  zPwdUtf8 = fossil_unicode_to_utf8(zPwd);
-  nPwd = strlen(zPwdUtf8);
-  if( nPwd > nBuf-1 ){
-    fossil_fatal("pwd too big: max %d\n", nBuf-1);
-  }
-  for(i=0; zPwdUtf8[i]; i++) if( zPwdUtf8[i]=='\\' ) zPwdUtf8[i] = '/';
-  memcpy(zBuf, zPwdUtf8, nPwd+1);
-  fossil_mbcs_free(zPwdUtf8);
+  win32_getcwd(zBuf, nBuf);
 #else
   if( getcwd(zBuf, nBuf-1)==0 ){
     if( errno==ERANGE ){
-      fossil_fatal("pwd too big: max %d\n", nBuf-1);
+      fossil_fatal("pwd too big: max %d", nBuf-1);
     }else{
       fossil_fatal("cannot find current working directory; %s",
                    strerror(errno));
@@ -654,11 +879,10 @@ void file_getcwd(char *zBuf, int nBuf){
 ** if it is relative.
 */
 int file_is_absolute_path(const char *zPath){
-  if( zPath[0]=='/'
-#if defined(_WIN32)
-      || zPath[0]=='\\'
-      || (strlen(zPath)>3 && zPath[1]==':'
-           && (zPath[2]=='\\' || zPath[2]=='/'))
+  if( fossil_isdirsep(zPath[0])
+#if defined(_WIN32) || defined(__CYGWIN__)
+      || (fossil_isalpha(zPath[0]) && zPath[1]==':'
+           && (fossil_isdirsep(zPath[2]) || zPath[2]=='\0'))
 #endif
   ){
     return 1;
@@ -677,41 +901,43 @@ int file_is_absolute_path(const char *zPath){
 ** is retained.
 */
 void file_canonical_name(const char *zOrigName, Blob *pOut, int slash){
+  blob_zero(pOut);
   if( file_is_absolute_path(zOrigName) ){
-#if defined(_WIN32)
-    char *zOut;
-#endif
-    blob_set(pOut, zOrigName);
-    blob_materialize(pOut);
-#if defined(_WIN32)
-    /*
-    ** On Windows, normalize the drive letter to upper case.
-    */
-    zOut = blob_str(pOut);
-    if( fossil_isalpha(zOut[0]) && zOut[1]==':' ){
-      zOut[0] = fossil_toupper(zOut[0]);
-    }
-#endif
+    blob_appendf(pOut, "%/", zOrigName);
   }else{
     char zPwd[2000];
     file_getcwd(zPwd, sizeof(zPwd)-strlen(zOrigName));
-#if defined(_WIN32)
-    /*
-    ** On Windows, normalize the drive letter to upper case.
-    */
-    if( fossil_isalpha(zPwd[0]) && zPwd[1]==':' ){
-      zPwd[0] = fossil_toupper(zPwd[0]);
+    if( zPwd[0]=='/' && strlen(zPwd)==1 ){
+      /* when on '/', don't add an extra '/' */
+      if( zOrigName[0]=='.' && strlen(zOrigName)==1 ){
+        /* '.' when on '/' mean '/' */
+        blob_appendf(pOut, "%/", zPwd);
+      }else{
+        blob_appendf(pOut, "%/%/", zPwd, zOrigName);
+      }
+    }else{
+      blob_appendf(pOut, "%//%/", zPwd, zOrigName);
     }
-#endif
-    blob_zero(pOut);
-    blob_appendf(pOut, "%//%/", zPwd, zOrigName);
   }
+#if defined(_WIN32) || defined(__CYGWIN__)
+  {
+    char *zOut;
+    /*
+    ** On Windows/cygwin, normalize the drive letter to upper case.
+    */
+    zOut = blob_str(pOut);
+    if( fossil_islower(zOut[0]) && zOut[1]==':' && zOut[2]=='/' ){
+      zOut[0] = fossil_toupper(zOut[0]);
+    }
+  }
+#endif
   blob_resize(pOut, file_simplify_name(blob_buffer(pOut),
                                        blob_size(pOut), slash));
 }
 
 /*
-** COMMAND:  test-canonical-name
+** COMMAND: test-canonical-name
+**
 ** Usage: %fossil test-canonical-name FILENAME...
 **
 ** Test the operation of the canonical name generator.
@@ -720,11 +946,12 @@ void file_canonical_name(const char *zOrigName, Blob *pOut, int slash){
 void cmd_test_canonical_name(void){
   int i;
   Blob x;
+  int slashFlag = find_option("slash",0,0)!=0;
   blob_zero(&x);
   for(i=2; i<g.argc; i++){
     char zBuf[100];
     const char *zName = g.argv[i];
-    file_canonical_name(zName, &x, 0);
+    file_canonical_name(zName, &x, slashFlag);
     fossil_print("[%s] -> [%s]\n", zName, blob_buffer(&x));
     blob_reset(&x);
     sqlite3_snprintf(sizeof(zBuf), zBuf, "%lld", file_wd_size(zName));
@@ -748,8 +975,8 @@ void cmd_test_canonical_name(void){
 int file_is_canonical(const char *z){
   int i;
   if( z[0]!='/'
-#if defined(_WIN32)
-    && (z[0]==0 || z[1]!=':' || z[2]!='/')
+#if defined(_WIN32) || defined(__CYGWIN__)
+    && (!fossil_isupper(z[0]) || z[1]!=':' || z[2]!='/')
 #endif
   ) return 0;
 
@@ -796,23 +1023,29 @@ void file_relative_name(const char *zOrigName, Blob *pOut, int slash){
     file_getcwd(zBuf, sizeof(zBuf)-20);
     zPwd = file_without_drive_letter(zBuf);
     i = 1;
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__CYGWIN__)
     while( zPath[i] && fossil_tolower(zPwd[i])==fossil_tolower(zPath[i]) ) i++;
 #else
     while( zPath[i] && zPwd[i]==zPath[i] ) i++;
 #endif
     if( zPath[i]==0 ){
-      blob_reset(pOut);
+      memcpy(&tmp, pOut, sizeof(tmp));
       if( zPwd[i]==0 ){
-        blob_append(pOut, ".", 1);
+        blob_set(pOut, ".");
       }else{
-        blob_append(pOut, "..", 2);
+        blob_set(pOut, "..");
         for(j=i+1; zPwd[j]; j++){
-          if( zPwd[j]=='/' ) {
+          if( zPwd[j]=='/' ){
             blob_append(pOut, "/..", 3);
           }
         }
+        while( i>0 && (zPwd[i]!='/')) --i;
+        blob_append(pOut, zPath+i, j-i);
       }
+      if( slash && i>0 && zPath[strlen(zPath)-1]=='/'){
+        blob_append(pOut, "/", 1);
+      }
+      blob_reset(&tmp);
       return;
     }
     if( zPwd[i]==0 && zPath[i]=='/' ){
@@ -823,9 +1056,14 @@ void file_relative_name(const char *zOrigName, Blob *pOut, int slash){
       return;
     }
     while( zPath[i-1]!='/' ){ i--; }
-    blob_set(&tmp, "../");
+    if( zPwd[0]=='/' && strlen(zPwd)==1 ){
+      /* If on '/', don't go to higher level */
+      blob_zero(&tmp);
+    }else{
+      blob_set(&tmp, "../");
+    }
     for(j=i; zPwd[j]; j++){
-      if( zPwd[j]=='/' ) {
+      if( zPwd[j]=='/' ){
         blob_append(&tmp, "../", 3);
       }
     }
@@ -836,39 +1074,70 @@ void file_relative_name(const char *zOrigName, Blob *pOut, int slash){
 }
 
 /*
-** COMMAND:  test-relative-name
+** COMMAND: test-relative-name
 **
 ** Test the operation of the relative name generator.
 */
 void cmd_test_relative_name(void){
   int i;
   Blob x;
+  int slashFlag = find_option("slash",0,0)!=0;
   blob_zero(&x);
   for(i=2; i<g.argc; i++){
-    file_relative_name(g.argv[i], &x, 0);
+    file_relative_name(g.argv[i], &x, slashFlag);
     fossil_print("%s\n", blob_buffer(&x));
     blob_reset(&x);
   }
 }
 
 /*
-** Compute a pathname for a file relative to the root of the local
-** tree.  Return TRUE on success.  On failure, print and error
-** message and quit if the errFatal flag is true.  If errFatal is
-** false, then simply return 0.
-**
-** The root of the tree is defined by the g.zLocalRoot variable.
+** Compute a full path name for a file in the local tree.  If
+** the absolute flag is non-zero, the computed path will be
+** absolute, starting with the root path of the local tree;
+** otherwise, it will be relative to the root of the local
+** tree.  In both cases, the root of the local tree is defined
+** by the g.zLocalRoot variable.  Return TRUE on success.  On
+** failure, print and error message and quit if the errFatal
+** flag is true.  If errFatal is false, then simply return 0.
 */
-int file_tree_name(const char *zOrigName, Blob *pOut, int errFatal){
+int file_tree_name(
+  const char *zOrigName,
+  Blob *pOut,
+  int absolute,
+  int errFatal
+){
   Blob localRoot;
   int nLocalRoot;
   char *zLocalRoot;
   Blob full;
   int nFull;
   char *zFull;
+  int (*xCmp)(const char*,const char*,int);
 
   blob_zero(pOut);
-  db_must_be_within_tree();
+  if( !g.localOpen ){
+    if( absolute && !file_is_absolute_path(zOrigName) ){
+      if( errFatal ){
+        fossil_fatal("relative to absolute needs open checkout tree: %s",
+                     zOrigName);
+      }
+      return 0;
+    }else{
+      /*
+      ** The original path may be relative or absolute; however, without
+      ** an open checkout tree, the only things we can do at this point
+      ** is return it verbatim or generate a fatal error.  The caller is
+      ** probably expecting a tree-relative path name will be returned;
+      ** however, most places where this function is called already check
+      ** if the local checkout tree is open, either directly or indirectly,
+      ** which would make this situation impossible.  Alternatively, they
+      ** could check the returned path using the file_is_absolute_path()
+      ** function.
+      */
+      blob_appendf(pOut, "%s", zOrigName);
+      return 1;
+    }
+  }
   file_canonical_name(g.zLocalRoot, &localRoot, 1);
   nLocalRoot = blob_size(&localRoot);
   zLocalRoot = blob_buffer(&localRoot);
@@ -876,16 +1145,26 @@ int file_tree_name(const char *zOrigName, Blob *pOut, int errFatal){
   file_canonical_name(zOrigName, &full, 0);
   nFull = blob_size(&full);
   zFull = blob_buffer(&full);
+  if( filenames_are_case_sensitive() ){
+    xCmp = fossil_strncmp;
+  }else{
+    xCmp = fossil_strnicmp;
+  }
 
   /* Special case.  zOrigName refers to g.zLocalRoot directory. */
-  if( nFull==nLocalRoot-1 && memcmp(zLocalRoot, zFull, nFull)==0 ){
-    blob_append(pOut, ".", 1);
+  if( (nFull==nLocalRoot-1 && xCmp(zLocalRoot, zFull, nFull)==0)
+      || (nFull==1 && zFull[0]=='/' && nLocalRoot==1 && zLocalRoot[0]=='/') ){
+    if( absolute ){
+      blob_append(pOut, zLocalRoot, nLocalRoot);
+    }else{
+      blob_append(pOut, ".", 1);
+    }
     blob_reset(&localRoot);
     blob_reset(&full);
     return 1;
   }
 
-  if( nFull<=nLocalRoot || memcmp(zLocalRoot, zFull, nLocalRoot) ){
+  if( nFull<=nLocalRoot || xCmp(zLocalRoot, zFull, nLocalRoot) ){
     blob_reset(&localRoot);
     blob_reset(&full);
     if( errFatal ){
@@ -893,23 +1172,39 @@ int file_tree_name(const char *zOrigName, Blob *pOut, int errFatal){
     }
     return 0;
   }
-  blob_append(pOut, &zFull[nLocalRoot], nFull-nLocalRoot);
+  if( absolute ){
+    if( !file_is_absolute_path(zOrigName) ){
+      blob_append(pOut, zLocalRoot, nLocalRoot);
+    }
+    blob_append(pOut, zOrigName, -1);
+    blob_resize(pOut, file_simplify_name(blob_buffer(pOut),
+                                         blob_size(pOut), 0));
+  }else{
+    blob_append(pOut, &zFull[nLocalRoot], nFull-nLocalRoot);
+  }
   blob_reset(&localRoot);
   blob_reset(&full);
   return 1;
 }
 
 /*
-** COMMAND:  test-tree-name
+** COMMAND: test-tree-name
 **
 ** Test the operation of the tree name generator.
+**
+** Options:
+**   --absolute           Return an absolute path instead of a relative one.
+**   --case-sensitive B   Enable or disable case-sensitive filenames.  B is
+**                        a boolean: "yes", "no", "true", "false", etc.
 */
 void cmd_test_tree_name(void){
   int i;
   Blob x;
+  int absoluteFlag = find_option("absolute",0,0)!=0;
+  db_find_and_open_repository(0,0);
   blob_zero(&x);
   for(i=2; i<g.argc; i++){
-    if( file_tree_name(g.argv[i], &x, 1) ){
+    if( file_tree_name(g.argv[i], &x, absoluteFlag, 1) ){
       fossil_print("%s\n", blob_buffer(&x));
       blob_reset(&x);
     }
@@ -961,19 +1256,22 @@ void file_parse_uri(
 ** Construct a random temporary filename into zBuf[].
 */
 void file_tempname(int nBuf, char *zBuf){
-  static const char *azDirs[] = {
 #if defined(_WIN32)
+  const char *azDirs[] = {
      0, /* GetTempPath */
      0, /* TEMP */
      0, /* TMP */
+     ".",
+  };
 #else
+  static const char *const azDirs[] = {
      "/var/tmp",
      "/usr/tmp",
      "/tmp",
      "/temp",
-#endif
      ".",
   };
+#endif
   static const unsigned char zChars[] =
     "abcdefghijklmnopqrstuvwxyz"
     "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
@@ -986,7 +1284,7 @@ void file_tempname(int nBuf, char *zBuf){
   wchar_t zTmpPath[MAX_PATH];
 
   if( GetTempPathW(MAX_PATH, zTmpPath) ){
-    azDirs[0] = fossil_unicode_to_utf8(zTmpPath);
+    azDirs[0] = fossil_path_to_utf8(zTmpPath);
   }
 
   azDirs[1] = fossil_getenv("TEMP");
@@ -1020,8 +1318,9 @@ void file_tempname(int nBuf, char *zBuf){
   }while( file_size(zBuf)>=0 );
 
 #if defined(_WIN32)
-  fossil_mbcs_free((char *)azDirs[1]);
-  fossil_mbcs_free((char *)azDirs[2]);
+  fossil_path_free((char *)azDirs[0]);
+  fossil_path_free((char *)azDirs[1]);
+  fossil_path_free((char *)azDirs[2]);
 #endif
 }
 
@@ -1049,161 +1348,39 @@ int file_is_the_same(Blob *pContent, const char *zName){
   return rc==0;
 }
 
-/*
-** Portable unicode implementation of opendir()
-*/
-#if INTERFACE
-
-#include <dirent.h>
-#if defined(_WIN32)
-# define DIR _WDIR
-# define dirent _wdirent
-# define opendir _wopendir
-# define readdir _wreaddir
-# define closedir _wclosedir
-#endif /* _WIN32 */
-
-#endif /* INTERFACE */
-
-
-
-/**************************************************************************
-** The following routines translate between MBCS and UTF8 on windows.
-** Since everything is always UTF8 on unix, these routines are no-ops
-** there.
-*/
-
-/*
-** Translate MBCS to UTF8.  Return a pointer to the translated text.
-** Call fossil_mbcs_free() to deallocate any memory used to store the
-** returned pointer when done.
-*/
-char *fossil_mbcs_to_utf8(const char *zMbcs){
-#ifdef _WIN32
-  extern char *sqlite3_win32_mbcs_to_utf8(const char*);
-  return sqlite3_win32_mbcs_to_utf8(zMbcs);
-#else
-  return (char*)zMbcs;  /* No-op on unix */
-#endif
-}
-
-/*
-** Translate Unicode to UTF8.  Return a pointer to the translated text.
-** Call fossil_mbcs_free() to deallocate any memory used to store the
-** returned pointer when done.
-*/
-char *fossil_unicode_to_utf8(const void *zUnicode){
-#ifdef _WIN32
-  int nByte = WideCharToMultiByte(CP_UTF8, 0, zUnicode, -1, 0, 0, 0, 0);
-  char *zUtf = sqlite3_malloc( nByte );
-  if( zUtf==0 ){
-    return 0;
-  }
-  WideCharToMultiByte(CP_UTF8, 0, zUnicode, -1, zUtf, nByte, 0, 0);
-  return zUtf;
-#else
-  return (char *)zUnicode;  /* No-op on unix */
-#endif
-}
-
-/*
-** Translate UTF8 to MBCS for use in system calls.  Return a pointer to the
-** translated text..  Call fossil_mbcs_free() to deallocate any memory
-** used to store the returned pointer when done.
-*/
-char *fossil_utf8_to_mbcs(const char *zUtf8){
-#ifdef _WIN32
-  extern char *sqlite3_win32_utf8_to_mbcs(const char*);
-  return sqlite3_win32_utf8_to_mbcs(zUtf8);
-#else
-  return (char*)zUtf8;  /* No-op on unix */
-#endif
-}
-
-/*
-** Translate UTF8 to unicode for use in system calls.  Return a pointer to the
-** translated text..  Call fossil_mbcs_free() to deallocate any memory
-** used to store the returned pointer when done.
-*/
-void *fossil_utf8_to_unicode(const char *zUtf8){
-#ifdef _WIN32
-  int nByte = MultiByteToWideChar(CP_UTF8, 0, zUtf8, -1, 0, 0);
-  wchar_t *zUnicode = sqlite3_malloc( nByte * 2 );
-  if( zUnicode==0 ){
-    return 0;
-  }
-  MultiByteToWideChar(CP_UTF8, 0, zUtf8, -1, zUnicode, nByte);
-  return zUnicode;
-#else
-  return (void *)zUtf8;  /* No-op on unix */
-#endif
-}
-
 /*
 ** Return the value of an environment variable as UTF8.
+** Use fossil_path_free() to release resources.
 */
 char *fossil_getenv(const char *zName){
 #ifdef _WIN32
   wchar_t *uName = fossil_utf8_to_unicode(zName);
   void *zValue = _wgetenv(uName);
-  fossil_mbcs_free(uName);
-  if( zValue ) zValue = fossil_unicode_to_utf8(zValue);
+  fossil_unicode_free(uName);
 #else
   char *zValue = getenv(zName);
 #endif
+  if( zValue ) zValue = fossil_path_to_utf8(zValue);
   return zValue;
 }
 
 /*
-** Display UTF8 on the console.  Return the number of
-** Characters written. If stdout or stderr is redirected
-** to a file, -1 is returned and nothing is written
-** to the console.
+** Sets the value of an environment variable as UTF8.
 */
-int fossil_utf8_to_console(const char *zUtf8, int nByte, int toStdErr){
+int fossil_setenv(const char *zName, const char *zValue){
+  int rc;
+  char *zString = mprintf("%s=%s", zName, zValue);
 #ifdef _WIN32
-  int nChar;
-  wchar_t *zUnicode; /* Unicode version of zUtf8 */
-  DWORD dummy;
-
-  static int istty[2] = { -1, -1 };
-  if( istty[toStdErr] == -1 ){
-    istty[toStdErr] = _isatty(toStdErr + 1) != 0;
-  }
-  if( !istty[toStdErr] ){
-    /* stdout/stderr is not a console. */
-    return -1;
-  }
-
-  nChar = MultiByteToWideChar(CP_UTF8, 0, zUtf8, nByte, NULL, 0);
-  zUnicode = malloc( (nChar + 1) *sizeof(zUnicode[0]) );
-  if( zUnicode==0 ){
-    return 0;
-  }
-  nChar = MultiByteToWideChar(CP_UTF8, 0, zUtf8, nByte, zUnicode, nChar);
-  if( nChar==0 ){
-    free(zUnicode);
-    return 0;
-  }
-  zUnicode[nChar] = '\0';
-  WriteConsoleW(GetStdHandle(STD_OUTPUT_HANDLE - toStdErr), zUnicode, nChar, &dummy, 0);
-  return nChar;
+  wchar_t *uString = fossil_utf8_to_unicode(zString);
+  rc = _wputenv(uString);
+  fossil_unicode_free(uString);
+  fossil_free(zString);
 #else
-  return -1;  /* No-op on unix */
-#endif
-}
-
-/*
-** Translate MBCS to UTF8.  Return a pointer.  Call fossil_mbcs_free()
-** to deallocate any memory used to store the returned pointer when done.
-*/
-void fossil_mbcs_free(void *zOld){
-#ifdef _WIN32
-  extern void sqlite3_free(void*);
-  sqlite3_free(zOld);
-#else
-  /* No-op on unix */
+  rc = putenv(zString);
+  /* NOTE: Cannot free the string on POSIX. */
+  /* fossil_free(zString); */
 #endif
+  return rc;
 }
 
 /*
@@ -1212,12 +1389,51 @@ void fossil_mbcs_free(void *zOld){
 FILE *fossil_fopen(const char *zName, const char *zMode){
 #ifdef _WIN32
   wchar_t *uMode = fossil_utf8_to_unicode(zMode);
-  wchar_t *uName = fossil_utf8_to_unicode(zName);
+  wchar_t *uName = fossil_utf8_to_path(zName, 0);
   FILE *f = _wfopen(uName, uMode);
-  fossil_mbcs_free(uName);
-  fossil_mbcs_free(uMode);
+  fossil_path_free(uName);
+  fossil_unicode_free(uMode);
 #else
   FILE *f = fopen(zName, zMode);
 #endif
   return f;
 }
+
+/*
+** Return non-NULL if zFilename contains pathname elements that
+** are reserved on Windows.  The returned string is the disallowed
+** path element.
+*/
+const char *file_is_win_reserved(const char *zPath){
+  static const char *azRes[] = { "CON", "PRN", "AUX", "NUL", "COM", "LPT" };
+  static char zReturn[5];
+  int i;
+  while( zPath[0] ){
+    for(i=0; i<ArraySize(azRes); i++){
+      if( sqlite3_strnicmp(zPath, azRes[i], 3)==0
+       && ((i>=4 && fossil_isdigit(zPath[3])
+                 && (zPath[4]=='/' || zPath[4]=='.' || zPath[4]==0))
+          || (i<4 && (zPath[3]=='/' || zPath[3]=='.' || zPath[3]==0)))
+      ){
+        sqlite3_snprintf(5,zReturn,"%.*s", i>=4 ? 4 : 3, zPath);
+        return zReturn;
+      }
+    }
+    while( zPath[0] && zPath[0]!='/' ) zPath++;
+    while( zPath[0]=='/' ) zPath++;
+  }
+  return 0;
+}
+
+/*
+** COMMAND: test-valid-for-windows
+** Usage:  fossil test-valid-for-windows FILENAME ....
+**
+** Show which filenames are not valid for Windows
+*/
+void file_test_valid_for_windows(void){
+  int i;
+  for(i=2; i<g.argc; i++){
+    fossil_print("%s %s\n", file_is_win_reserved(g.argv[i]), g.argv[i]);
+  }
+}
