@@ -14,7 +14,7 @@
 **   http://www.hwaci.com/drh/
 **
 *******************************************************************************
-**  
+**
 ** Code to generate the ticket listings
 */
 #include "config.h"
@@ -25,8 +25,14 @@
 /* Forward references to static routines */
 static void report_format_hints(void);
 
+#ifndef SQLITE_RECURSIVE
+#  define SQLITE_RECURSIVE            33
+#endif
+
 /*
-** WEBPAGE: /reportlist
+** WEBPAGE: reportlist
+**
+** Main menu for Tickets.
 */
 void view_list(void){
   const char *zScript;
@@ -36,12 +42,16 @@ void view_list(void){
   int cnt = 0;
 
   login_check_credentials();
-  if( !g.perm.RdTkt && !g.perm.NewTkt ){ login_needed(); return; }
+  if( !g.perm.RdTkt && !g.perm.NewTkt ){
+    login_needed(g.anon.RdTkt || g.anon.NewTkt);
+    return;
+  }
   style_header("Ticket Main Menu");
+  ticket_standard_submenu(T_ALL_BUT(T_REPLIST));
   if( g.thTrace ) Th_Trace("BEGIN_REPORTLIST<br />\n", -1);
   zScript = ticket_reportlist_code();
   if( g.thTrace ) Th_Trace("BEGIN_REPORTLIST_SCRIPT<br />\n", -1);
-  
+
   blob_zero(&ril);
   ticket_init();
 
@@ -62,16 +72,16 @@ void view_list(void){
     }
     blob_appendf(&ril, "&nbsp;&nbsp;&nbsp;");
     if( g.perm.Write && zOwner && zOwner[0] ){
-      blob_appendf(&ril, "(by <i>%h</i></i>) ", zOwner);
+      blob_appendf(&ril, "(by <i>%h</i>) ", zOwner);
     }
     if( g.perm.TktFmt ){
       blob_appendf(&ril, "[%zcopy</a>] ",
                    href("%R/rptedit?rn=%d&copy=1", rn));
     }
-    if( g.perm.Admin 
+    if( g.perm.Admin
      || (g.perm.WrTkt && zOwner && fossil_strcmp(g.zLogin,zOwner)==0)
     ){
-      blob_appendf(&ril, "[%zedit</a>]", 
+      blob_appendf(&ril, "[%zedit</a>]",
                          href("%R/rptedit?rn=%d", rn));
     }
     if( g.perm.TktFmt ){
@@ -80,11 +90,12 @@ void view_list(void){
     }
     blob_appendf(&ril, "</li>\n");
   }
+  db_finalize(&q);
 
   Th_Store("report_items", blob_str(&ril));
-  
+
   Th_Render(zScript);
-  
+
   blob_reset(&ril);
   if( g.thTrace ) Th_Trace("END_REPORTLIST<br />\n", -1);
 
@@ -120,7 +131,7 @@ char *extract_integer(const char *zOrig){
 
 /*
 ** Remove blank lines from the beginning of a string and
-** all whitespace from the end. Removes whitespace preceeding a NL,
+** all whitespace from the end. Removes whitespace preceding a NL,
 ** which also converts any CRNL sequence into a single NL.
 */
 char *remove_blank_lines(const char *zOrig){
@@ -164,16 +175,18 @@ int report_query_authorizer(
   int rc = SQLITE_OK;
   if( *(char**)pError ){
     /* We've already seen an error.  No need to continue. */
-    return SQLITE_OK;
+    return SQLITE_DENY;
   }
   switch( code ){
     case SQLITE_SELECT:
+    case SQLITE_RECURSIVE:
     case SQLITE_FUNCTION: {
       break;
     }
     case SQLITE_READ: {
       static const char *const azAllowed[] = {
          "ticket",
+         "ticketchng",
          "blob",
          "filename",
          "mlink",
@@ -181,8 +194,12 @@ int report_query_authorizer(
          "event",
          "tag",
          "tagxref",
+         "unversioned",
       };
       int i;
+      if( fossil_strncmp(zArg1, "fx_", 3)==0 ){
+        break;
+      }
       for(i=0; i<sizeof(azAllowed)/sizeof(azAllowed[0]); i++){
         if( fossil_stricmp(zArg1, azAllowed[i])==0 ) break;
       }
@@ -207,7 +224,6 @@ int report_query_authorizer(
 ** Activate the query authorizer
 */
 static void report_restrict_sql(char **pzErr){
-  (void)fossil_localtime(0);
   sqlite3_set_authorizer(g.db, report_query_authorizer, (void*)pzErr);
 }
 static void report_unrestrict_sql(void){
@@ -229,11 +245,13 @@ char *verify_sql_statement(char *zSql){
   int rc;
 
   /* First make sure the SQL is a single query command by verifying that
-  ** the first token is "SELECT" and that there are no unquoted semicolons.
+  ** the first token is "SELECT" or "WITH" and that there are no unquoted
+  ** semicolons.
   */
   for(i=0; fossil_isspace(zSql[i]); i++){}
-  if( fossil_strnicmp(&zSql[i],"select",6)!=0 ){
-    return mprintf("The SQL must be a SELECT statement");
+  if( fossil_strnicmp(&zSql[i], "select", 6)!=0
+      && fossil_strnicmp(&zSql[i], "with", 4)!=0 ){
+    return mprintf("The SQL must be a SELECT or WITH statement");
   }
   for(i=0; zSql[i]; i++){
     if( zSql[i]==';' ){
@@ -251,10 +269,10 @@ char *verify_sql_statement(char *zSql){
       }
     }
   }
-  
+
   /* Compile the statement and check for illegal accesses or syntax errors. */
   report_restrict_sql(&zErr);
-  rc = sqlite3_prepare(g.db, zSql, -1, &pStmt, &zTail);
+  rc = sqlite3_prepare_v2(g.db, zSql, -1, &pStmt, &zTail);
   if( rc!=SQLITE_OK ){
     zErr = mprintf("Syntax error: %s", sqlite3_errmsg(g.db));
   }
@@ -269,7 +287,11 @@ char *verify_sql_statement(char *zSql){
 }
 
 /*
-** WEBPAGE: /rptsql
+** WEBPAGE: rptsql
+** URL: /rptsql?rn=N
+**
+** Display the SQL query used to generate a ticket report.  The rn=N
+** query parameter identifies the specific report number to be displayed.
 */
 void view_see_sql(void){
   int rn;
@@ -281,7 +303,7 @@ void view_see_sql(void){
 
   login_check_credentials();
   if( !g.perm.TktFmt ){
-    login_needed();
+    login_needed(g.anon.TktFmt);
     return;
   }
   rn = atoi(PD("rn","0"));
@@ -291,6 +313,7 @@ void view_see_sql(void){
   if( db_step(&q)!=SQLITE_ROW ){
     @ <p>Unknown report number: %d(rn)</p>
     style_footer();
+    db_finalize(&q);
     return;
   }
   zTitle = db_column_text(&q, 0);
@@ -312,11 +335,21 @@ void view_see_sql(void){
   @ </tr></table>
   report_format_hints();
   style_footer();
+  db_finalize(&q);
 }
 
 /*
-** WEBPAGE: /rptnew
-** WEBPAGE: /rptedit
+** WEBPAGE: rptnew
+** WEBPAGE: rptedit
+**
+** Create (/rptnew) or edit (/rptedit) a ticket report format.
+** Query parameters:
+**
+**     rn=N           Ticket report number. (required)
+**     t=TITLE        Title of the report format
+**     w=USER         Owner of the report format
+**     s=SQL          SQL text used to implement the report
+**     k=KEY          Color key
 */
 void view_edit(void){
   int rn;
@@ -329,7 +362,7 @@ void view_edit(void){
 
   login_check_credentials();
   if( !g.perm.TktFmt ){
-    login_needed();
+    login_needed(g.anon.TktFmt);
     return;
   }
   /*view_add_functions(0);*/
@@ -372,7 +405,7 @@ void view_edit(void){
     if( zSQL[0]==0 ){
       zErr = "Please supply an SQL query statement";
     }else if( (zTitle = trim_string(zTitle))[0]==0 ){
-      zErr = "Please supply a title"; 
+      zErr = "Please supply a title";
     }else{
       zErr = verify_sql_statement(zSQL);
     }
@@ -423,7 +456,7 @@ void view_edit(void){
   if( rn>0 ){
     style_submenu_element("Delete", "Delete", "rptedit?rn=%d&del1=1", rn);
   }
-  style_header(rn>0 ? "Edit Report Format":"Create New Report Format");
+  style_header("%s", rn>0 ? "Edit Report Format":"Create New Report Format");
   if( zErr ){
     @ <blockquote class="reportError">%h(zErr)</blockquote>
   }
@@ -491,9 +524,16 @@ static void report_format_hints(void){
   @ <li><p>If a column of the result set is named "bgcolor" then the content
   @ of that column determines the background color of the row.</p></li>
   @
+  @ <li><p>The text of all columns prior to the first column whose name begins
+  @ with underscore ("_") is shown character-for-character as it appears in
+  @ the database.  In other words, it is assumed to have a mimetype of
+  @ text/plain.
+  @
   @ <li><p>The first column whose name begins with underscore ("_") and all
-  @ subsequent columns are shown on their own rows in the table.  This might
-  @ be useful for displaying the description of tickets.
+  @ subsequent columns are shown on their own rows in the table and with
+  @ wiki formatting.  In other words, such rows are shown with a mimetype
+  @ of text/x-fossil-wiki.  This is recommended for the "description" field
+  @ of tickets.
   @ </p></li>
   @
   @ <li><p>The query can join other tables in the database besides TICKET.
@@ -591,21 +631,8 @@ static void report_format_hints(void){
   @    severity AS 'Svr',
   @    priority AS 'Pri',
   @    title AS 'Title',
-  @    description AS '_Description',   -- When the column name begins with '_'
-  @    remarks AS '_Remarks'            -- the data is shown on a separate row.
-  @  FROM ticket
-  @ </pre></blockquote>
-  @
-  @ <p>Or, to see part of the description on the same row, use the
-  @ <b>wiki()</b> function with some string manipulation. Using the
-  @ <b>tkt()</b> function on the ticket number will also generate a linked
-  @ field, but without the extra <i>edit</i> column:
-  @ </p>
-  @ <blockquote><pre>
-  @  SELECT
-  @    tkt(tn) AS '',
-  @    title AS 'Title',
-  @    wiki(substr(description,0,80)) AS 'Description'
+  @    description AS '_Description',  -- When the column name begins with '_'
+  @    remarks AS '_Remarks'           -- content is rendered as wiki
   @  FROM ticket
   @ </pre></blockquote>
   @
@@ -621,6 +648,9 @@ struct GenerateHTML {
   int isMultirow;  /* True if multiple table rows per query result row */
   int iNewRow;     /* Index of first column that goes on separate row */
   int iBg;         /* Index of column that defines background color */
+  int wikiFlags;   /* Flags passed into wiki_convert() */
+  const char *zWikiStart;    /* HTML before display of multi-line wiki */
+  const char *zWikiEnd;      /* HTML after display of multi-line wiki */
 };
 
 /*
@@ -629,13 +659,13 @@ struct GenerateHTML {
 static int generate_html(
   void *pUser,     /* Pointer to output state */
   int nArg,        /* Number of columns in this result row */
-  char **azArg,    /* Text of data in all columns */
-  char **azName    /* Names of the columns */
+  const char **azArg, /* Text of data in all columns */
+  const char **azName /* Names of the columns */
 ){
   struct GenerateHTML *pState = (struct GenerateHTML*)pUser;
   int i;
   const char *zTid;  /* Ticket UUID.  (value of column named '#') */
-  char *zBg = 0;     /* Use this background color */
+  const char *zBg = 0; /* Use this background color */
 
   /* Do initialization
   */
@@ -665,6 +695,19 @@ static int generate_html(
         if( azName[i][0]=='_' ){
           pState->isMultirow = 1;
           pState->iNewRow = i;
+          pState->wikiFlags = WIKI_NOBADLINKS;
+          pState->zWikiStart = "";
+          pState->zWikiEnd = "";
+          if( P("plaintext") ){
+            pState->wikiFlags |= WIKI_LINKSONLY;
+            pState->zWikiStart = "<pre class='verbatim'>";
+            pState->zWikiEnd = "</pre>";
+            style_submenu_element("Formatted", "Formatted",
+                                  "%R/rptview?rn=%d", pState->rn);
+          }else{
+            style_submenu_element("Plaintext", "Plaintext",
+                                  "%R/rptview?rn=%d&plaintext", pState->rn);
+          }
         }else{
           pState->nCol++;
         }
@@ -673,10 +716,10 @@ static int generate_html(
 
     /* The first time this routine is called, output a table header
     */
-    @ <tr>
+    @ <thead><tr>
     zTid = 0;
     for(i=0; i<nArg; i++){
-      char *zName = azName[i];
+      const char *zName = azName[i];
       if( i==pState->iBg ) continue;
       if( pState->iNewRow>=0 && i>=pState->iNewRow ){
         if( g.perm.Write && zTid ){
@@ -695,7 +738,7 @@ static int generate_html(
     if( g.perm.Write && zTid ){
       @ <th>&nbsp;</th>
     }
-    @ </tr>
+    @ </tr></thead><tbody>
   }
   if( azArg==0 ){
     @ <tr><td colspan="%d(pState->nCol)">
@@ -719,7 +762,7 @@ static int generate_html(
   @ <tr style="background-color:%h(zBg)">
   zTid = 0;
   for(i=0; i<nArg; i++){
-    char *zData;
+    const char *zData;
     if( i==pState->iBg ) continue;
     zData = azArg[i];
     if( zData==0 ) zData = "";
@@ -730,10 +773,13 @@ static int generate_html(
       }
       if( zData[0] ){
         Blob content;
-        @ </tr><tr style="background-color:%h(zBg)"><td colspan=%d(pState->nCol)>
+        @ </tr>
+        @ <tr style="background-color:%h(zBg)"><td colspan=%d(pState->nCol)>
+        @ %s(pState->zWikiStart)
         blob_init(&content, zData, -1);
-        wiki_convert(&content, 0, 0);
+        wiki_convert(&content, 0, pState->wikiFlags);
         blob_reset(&content);
+        @ %s(pState->zWikiEnd)
       }
     }else if( azName[i][0]=='#' ){
       zTid = zData;
@@ -778,8 +824,8 @@ static void output_no_tabs(const char *z){
 static int output_tab_separated(
   void *pUser,     /* Pointer to row-count integer */
   int nArg,        /* Number of columns in this result row */
-  char **azArg,    /* Text of data in all columns */
-  char **azName    /* Names of the columns */
+  const char **azArg, /* Text of data in all columns */
+  const char **azName /* Names of the columns */
 ){
   int *pCount = (int*)pUser;
   int i;
@@ -803,14 +849,15 @@ static int output_tab_separated(
 */
 void output_color_key(const char *zClrKey, int horiz, char *zTabArgs){
   int i, j, k;
-  char *zSafeKey, *zToFree;
+  const char *zSafeKey;
+  char *zToFree;
   while( fossil_isspace(*zClrKey) ) zClrKey++;
   if( zClrKey[0]==0 ) return;
   @ <table %s(zTabArgs)>
   if( horiz ){
     @ <tr>
   }
-  zToFree = zSafeKey = mprintf("%h", zClrKey);
+  zSafeKey = zToFree = mprintf("%h", zClrKey);
   while( zSafeKey[0] ){
     while( fossil_isspace(*zSafeKey) ) zSafeKey++;
     for(i=0; zSafeKey[i] && !fossil_isspace(zSafeKey[i]); i++){}
@@ -836,20 +883,22 @@ void output_color_key(const char *zClrKey, int horiz, char *zTabArgs){
 ** Execute a single read-only SQL statement.  Invoke xCallback() on each
 ** row.
 */
-int sqlite3_exec_readonly(
+static int db_exec_readonly(
   sqlite3 *db,                /* The database on which the SQL executes */
   const char *zSql,           /* The SQL to be executed */
-  sqlite3_callback xCallback, /* Invoke this callback routine */
+  int (*xCallback)(void*,int,const char**, const char**),
+                              /* Invoke this callback routine */
   void *pArg,                 /* First argument to xCallback() */
   char **pzErrMsg             /* Write error messages here */
 ){
   int rc = SQLITE_OK;         /* Return code */
   const char *zLeftover;      /* Tail of unprocessed SQL */
   sqlite3_stmt *pStmt = 0;    /* The current SQL statement */
-  char **azCols = 0;          /* Names of result columns */
+  const char **azCols = 0;    /* Names of result columns */
   int nCol;                   /* Number of columns of output */
-  char **azVals = 0;          /* Text of all output columns */
+  const char **azVals = 0;    /* Text of all output columns */
   int i;                      /* Loop counter */
+  int nVar;                   /* Number of parameters */
 
   pStmt = 0;
   rc = sqlite3_prepare_v2(db, zSql, -1, &pStmt, &zLeftover);
@@ -866,33 +915,226 @@ int sqlite3_exec_readonly(
     return SQLITE_ERROR;
   }
 
-  i = sqlite3_bind_parameter_index(pStmt, "$login");
-  if( i ) sqlite3_bind_text(pStmt, i, g.zLogin, -1, SQLITE_TRANSIENT);
-
+  nVar = sqlite3_bind_parameter_count(pStmt);
+  for(i=1; i<=nVar; i++){
+    const char *zVar = sqlite3_bind_parameter_name(pStmt, i);
+    if( zVar==0 ) continue;
+    if( zVar[0]!='$' && zVar[0]!='$' && zVar[0]!=':' ) continue;
+    if( !fossil_islower(zVar[1]) ) continue;
+    if( strcmp(zVar, "$login")==0 ){
+      sqlite3_bind_text(pStmt, i, g.zLogin, -1, SQLITE_TRANSIENT);
+    }else{
+      sqlite3_bind_text(pStmt, i, P(zVar+1), -1, SQLITE_TRANSIENT);
+    }
+  }
   nCol = sqlite3_column_count(pStmt);
   azVals = fossil_malloc(2*nCol*sizeof(const char*) + 1);
   while( (rc = sqlite3_step(pStmt))==SQLITE_ROW ){
     if( azCols==0 ){
       azCols = &azVals[nCol];
       for(i=0; i<nCol; i++){
-        azCols[i] = (char *)sqlite3_column_name(pStmt, i);
+        azCols[i] = sqlite3_column_name(pStmt, i);
       }
     }
     for(i=0; i<nCol; i++){
-      azVals[i] = (char *)sqlite3_column_text(pStmt, i);
+      azVals[i] = (const char *)sqlite3_column_text(pStmt, i);
     }
     if( xCallback(pArg, nCol, azVals, azCols) ){
       break;
     }
   }
   rc = sqlite3_finalize(pStmt);
-  fossil_free(azVals);
+  fossil_free((void *)azVals);
   return rc;
 }
 
+/*
+** Output Javascript code that will enables sorting of the table with
+** the id zTableId by clicking.
+**
+** The javascript was originally derived from:
+**
+**     http://www.webtoolkit.info/sortable-html-table.html
+**
+** But there have been extensive modifications.
+**
+** This variation allows column types to be expressed using the second
+** argument.  Each character of the second argument represent a column.
+**
+**       t      Sort by text
+**       n      Sort numerically
+**       k      Sort by the data-sortkey property
+**       x      This column is not sortable
+**
+** Capital letters mean sort in reverse order.
+** If there are fewer characters in zColumnTypes[] than their are columns,
+** then all extra columns assume type "t" (text).
+**
+** The third parameter is the column that was initially sorted (using 1-based
+** column numbers, like SQL).  Make this value 0 if none of the columns are
+** initially sorted.  Make the value negative if the column is initially sorted
+** in reverse order.
+**
+** Clicking on the same column header twice in a row inverts the sort.
+*/
+void output_table_sorting_javascript(
+  const char *zTableId,      /* ID of table to sort */
+  const char *zColumnTypes,  /* String for column types */
+  int iInitSort              /* Initially sorted column. Leftmost is 1. 0 for NONE */
+){
+  @ <script>
+  @ function SortableTable(tableEl,columnTypes,initSort){
+  @   this.tbody = tableEl.getElementsByTagName('tbody');
+  @   this.columnTypes = columnTypes;
+  @   var ncols = tableEl.rows[0].cells.length;
+  @   for(var i = columnTypes.length; i<=ncols; i++){this.columnTypes += 't';}
+  @   this.sort = function (cell) {
+  @     var column = cell.cellIndex;
+  @     var sortFn;
+  @     switch( cell.sortType ){
+  if( strchr(zColumnTypes,'n') ){
+    @       case "n": sortFn = this.sortNumeric;  break;
+  }
+  if( strchr(zColumnTypes,'N') ){
+    @       case "N": sortFn = this.sortReverseNumeric;  break;
+  }
+  @       case "t": sortFn = this.sortText;  break;
+  if( strchr(zColumnTypes,'T') ){
+    @       case "T": sortFn = this.sortReverseText;  break;
+  }
+  if( strchr(zColumnTypes,'k') ){
+    @       case "k": sortFn = this.sortKey;  break;
+  }
+  if( strchr(zColumnTypes,'K') ){
+    @       case "K": sortFn = this.sortReverseKey;  break;
+  }
+  @       default:  return;
+  @     }
+  @     this.sortIndex = column;
+  @     var newRows = new Array();
+  @     for (j = 0; j < this.tbody[0].rows.length; j++) {
+  @        newRows[j] = this.tbody[0].rows[j];
+  @     }
+  @     if( this.sortIndex==Math.abs(this.prevColumn)-1 ){
+  @       newRows.reverse();
+  @       this.prevColumn = -this.prevColumn;
+  @     }else{
+  @       newRows.sort(sortFn);
+  @       this.prevColumn = this.sortIndex+1;
+  @     }
+  @     for (i=0;i<newRows.length;i++) {
+  @       this.tbody[0].appendChild(newRows[i]);
+  @     }
+  @     this.setHdrIcons();
+  @   }
+  @   this.setHdrIcons = function() {
+  @     for (var i=0; i<this.hdrRow.cells.length; i++) {
+  @       if( this.columnTypes[i]=='x' ) continue;
+  @       var sortType;
+  @       if( this.prevColumn==i+1 ){
+  @         sortType = 'asc';
+  @       }else if( this.prevColumn==(-1-i) ){
+  @         sortType = 'desc'
+  @       }else{
+  @         sortType = 'none';
+  @       }
+  @       var hdrCell = this.hdrRow.cells[i];
+  @       var clsName = hdrCell.className.replace(/\s*\bsort\s*\w+/, '');
+  @       clsName += ' sort ' + sortType;
+  @       hdrCell.className = clsName;
+  @     }
+  @   }
+  @   this.sortText = function(a,b) {
+  @     var i = thisObject.sortIndex;
+  @     aa = a.cells[i].textContent.replace(/^\W+/,'').toLowerCase();
+  @     bb = b.cells[i].textContent.replace(/^\W+/,'').toLowerCase();
+  @     if(aa<bb) return -1;
+  @     if(aa==bb) return a.rowIndex-b.rowIndex;
+  @     return 1;
+  @   }
+  if( strchr(zColumnTypes,'T') ){
+    @   this.sortReverseText = function(a,b) {
+    @     var i = thisObject.sortIndex;
+    @     aa = a.cells[i].textContent.replace(/^\W+/,'').toLowerCase();
+    @     bb = b.cells[i].textContent.replace(/^\W+/,'').toLowerCase();
+    @     if(aa<bb) return +1;
+    @     if(aa==bb) return a.rowIndex-b.rowIndex;
+    @     return -1;
+    @   }
+  }
+  if( strchr(zColumnTypes,'n') ){
+    @   this.sortNumeric = function(a,b) {
+    @     var i = thisObject.sortIndex;
+    @     aa = parseFloat(a.cells[i].textContent);
+    @     if (isNaN(aa)) aa = 0;
+    @     bb = parseFloat(b.cells[i].textContent);
+    @     if (isNaN(bb)) bb = 0;
+    @     if(aa==bb) return a.rowIndex-b.rowIndex;
+    @     return aa-bb;
+    @   }
+  }
+  if( strchr(zColumnTypes,'N') ){
+    @   this.sortReverseNumeric = function(a,b) {
+    @     var i = thisObject.sortIndex;
+    @     aa = parseFloat(a.cells[i].textContent);
+    @     if (isNaN(aa)) aa = 0;
+    @     bb = parseFloat(b.cells[i].textContent);
+    @     if (isNaN(bb)) bb = 0;
+    @     if(aa==bb) return a.rowIndex-b.rowIndex;
+    @     return bb-aa;
+    @   }
+  }
+  if( strchr(zColumnTypes,'k') ){
+    @   this.sortKey = function(a,b) {
+    @     var i = thisObject.sortIndex;
+    @     aa = a.cells[i].getAttribute("data-sortkey");
+    @     bb = b.cells[i].getAttribute("data-sortkey");
+    @     if(aa<bb) return -1;
+    @     if(aa==bb) return a.rowIndex-b.rowIndex;
+    @     return 1;
+    @   }
+  }
+  if( strchr(zColumnTypes,'K') ){
+    @   this.sortReverseKey = function(a,b) {
+    @     var i = thisObject.sortIndex;
+    @     aa = a.cells[i].getAttribute("data-sortkey");
+    @     bb = b.cells[i].getAttribute("data-sortkey");
+    @     if(aa<bb) return +1;
+    @     if(aa==bb) return a.rowIndex-b.rowIndex;
+    @     return -1;
+    @   }
+  }
+  @   var x = tableEl.getElementsByTagName('thead');
+  @   if(!(this.tbody && this.tbody[0].rows && this.tbody[0].rows.length>0)){
+  @     return;
+  @   }
+  @   if(x && x[0].rows && x[0].rows.length > 0) {
+  @     this.hdrRow = x[0].rows[0];
+  @   } else {
+  @     return;
+  @   }
+  @   var thisObject = this;
+  @   this.prevColumn = initSort;
+  @   for (var i=0; i<this.hdrRow.cells.length; i++) {
+  @     if( columnTypes[i]=='x' ) continue;
+  @     var hdrcell = this.hdrRow.cells[i];
+  @     hdrcell.sTable = this;
+  @     hdrcell.style.cursor = "pointer";
+  @     hdrcell.sortType = columnTypes[i] || 't';
+  @     hdrcell.onclick = function () {
+  @       this.sTable.sort(this);
+  @       return false;
+  @     }
+  @   }
+  @   this.setHdrIcons()
+  @ }
+  @ var t = new SortableTable(gebi("%s(zTableId)"),"%s(zColumnTypes)",%d(iInitSort));
+  @ </script>
+}
+
 
 /*
-** WEBPAGE: /rptview
+** WEBPAGE: rptview
 **
 ** Generate a report.  The rn query parameter is the report number
 ** corresponding to REPORTFMT.RN.  If the tablist query parameter exists,
@@ -901,7 +1143,7 @@ int sqlite3_exec_readonly(
 */
 void rptview_page(void){
   int count = 0;
-  int rn;
+  int rn, rc;
   char *zSql;
   char *zTitle;
   char *zOwner;
@@ -912,17 +1154,21 @@ void rptview_page(void){
   char *zErr2 = 0;
 
   login_check_credentials();
-  if( !g.perm.RdTkt ){ login_needed(); return; }
-  rn = atoi(PD("rn","0"));
-  if( rn==0 ){
-    cgi_redirect("reportlist");
-    return;
-  }
+  if( !g.perm.RdTkt ){ login_needed(g.anon.RdTkt); return; }
   tabs = P("tablist")!=0;
-  /* view_add_functions(tabs); */
   db_prepare(&q,
-    "SELECT title, sqlcode, owner, cols FROM reportfmt WHERE rn=%d", rn);
-  if( db_step(&q)!=SQLITE_ROW ){
+    "SELECT title, sqlcode, owner, cols, rn FROM reportfmt WHERE rn=%d",
+     atoi(PD("rn","0")));
+  rc = db_step(&q);
+  if( rc!=SQLITE_ROW ){
+    db_finalize(&q);
+    db_prepare(&q,
+      "SELECT title, sqlcode, owner, cols, rn FROM reportfmt WHERE title GLOB %Q",
+      P("title"));
+    rc = db_step(&q);
+  }
+  if( rc!=SQLITE_ROW ){
+    db_finalize(&q);
     cgi_redirect("reportlist");
     return;
   }
@@ -930,6 +1176,7 @@ void rptview_page(void){
   zSql = db_column_malloc(&q, 1);
   zOwner = db_column_malloc(&q, 2);
   zClrKey = db_column_malloc(&q, 3);
+  rn = db_column_int(&q,4);
   db_finalize(&q);
 
   if( P("order_by") ){
@@ -949,12 +1196,12 @@ void rptview_page(void){
 
   count = 0;
   if( !tabs ){
-    struct GenerateHTML sState;
+    struct GenerateHTML sState = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
 
     db_multi_exec("PRAGMA empty_result_callbacks=ON");
-    style_submenu_element("Raw", "Raw", 
+    style_submenu_element("Raw", "Raw",
       "rptview?tablist=1&%h", PD("QUERY_STRING",""));
-    if( g.perm.Admin 
+    if( g.perm.Admin
        || (g.perm.TktFmt && g.zLogin && fossil_strcmp(g.zLogin,zOwner)==0) ){
       style_submenu_element("Edit", "Edit", "rptedit?rn=%d", rn);
     }
@@ -965,25 +1212,27 @@ void rptview_page(void){
       style_submenu_element("New Ticket", "Create a new ticket",
         "%s/tktnew", g.zTop);
     }
-    style_header(zTitle);
-    output_color_key(zClrKey, 1, 
+    style_header("%s", zTitle);
+    output_color_key(zClrKey, 1,
         "border=\"0\" cellpadding=\"3\" cellspacing=\"0\" class=\"report\"");
-    @ <table border="1" cellpadding="2" cellspacing="0" class="report">
+    @ <table border="1" cellpadding="2" cellspacing="0" class="report"
+    @  id="reportTable">
     sState.rn = rn;
     sState.nCount = 0;
     report_restrict_sql(&zErr1);
-    sqlite3_exec_readonly(g.db, zSql, generate_html, &sState, &zErr2);
+    db_exec_readonly(g.db, zSql, generate_html, &sState, &zErr2);
     report_unrestrict_sql();
-    @ </table>
+    @ </tbody></table>
     if( zErr1 ){
       @ <p class="reportError">Error: %h(zErr1)</p>
     }else if( zErr2 ){
       @ <p class="reportError">Error: %h(zErr2)</p>
     }
+    output_table_sorting_javascript("reportTable","",0);
     style_footer();
   }else{
     report_restrict_sql(&zErr1);
-    sqlite3_exec_readonly(g.db, zSql, output_tab_separated, &count, &zErr2);
+    db_exec_readonly(g.db, zSql, output_tab_separated, &count, &zErr2);
     report_unrestrict_sql();
     cgi_set_content_type("text/plain");
   }
@@ -1005,17 +1254,15 @@ static const char zFullTicketRptTitle[] = "full ticket export";
 */
 void rpt_list_reports(void){
   Stmt q;
-  char const aRptOutFrmt[] = "%s\t%s\n";
-
   fossil_print("Available reports:\n");
-  fossil_print(aRptOutFrmt,"report number","report title");
-  fossil_print(aRptOutFrmt,zFullTicketRptRn,zFullTicketRptTitle);
+  fossil_print("%s\t%s\n","report number","report title");
+  fossil_print("%s\t%s\n",zFullTicketRptRn,zFullTicketRptTitle);
   db_prepare(&q,"SELECT rn,title FROM reportfmt ORDER BY rn");
   while( db_step(&q)==SQLITE_ROW ){
     const char *zRn = db_column_text(&q, 0);
     const char *zTitle = db_column_text(&q, 1);
 
-    fossil_print(aRptOutFrmt,zRn,zTitle);
+    fossil_print("%s\t%s\n",zRn,zTitle);
   }
   db_finalize(&q);
 }
@@ -1062,7 +1309,7 @@ static void output_no_tabs_file(const char *z){
         }
         z += j;
       }
-      break; 
+      break;
   }
 }
 
@@ -1072,8 +1319,8 @@ static void output_no_tabs_file(const char *z){
 int output_separated_file(
   void *pUser,     /* Pointer to row-count integer */
   int nArg,        /* Number of columns in this result row */
-  char **azArg,    /* Text of data in all columns */
-  char **azName    /* Names of the columns */
+  const char **azArg, /* Text of data in all columns */
+  const char **azName /* Names of the columns */
 ){
   int *pCount = (int*)pUser;
   int i;
@@ -1094,10 +1341,10 @@ int output_separated_file(
 
 /*
 ** Generate a report.  The rn query parameter is the report number.
-** The output is written to stdout as flat file. The zFilter paramater
+** The output is written to stdout as flat file. The zFilter parameter
 ** is a full WHERE-condition.
 */
-void rptshow( 
+void rptshow(
     const char *zRep,
     const char *zSepIn,
     const char *zFilter,
@@ -1110,7 +1357,7 @@ void rptshow(
   int count = 0;
   int rn;
 
-  if (!zRep || !strcmp(zRep,zFullTicketRptRn) || !strcmp(zRep,zFullTicketRptTitle) ){
+  if( !zRep || !strcmp(zRep,zFullTicketRptRn) || !strcmp(zRep,zFullTicketRptTitle) ){
     zSql = "SELECT * FROM ticket";
   }else{
     rn = atoi(zRep);
@@ -1136,7 +1383,7 @@ void rptshow(
   tktEncode = enc;
   zSep = zSepIn;
   report_restrict_sql(&zErr1);
-  sqlite3_exec_readonly(g.db, zSql, output_separated_file, &count, &zErr2);
+  db_exec_readonly(g.db, zSql, output_separated_file, &count, &zErr2);
   report_unrestrict_sql();
   if( zFilter ){
     free(zSql);
