@@ -22,7 +22,12 @@
 */
 #include "config.h"
 #include "sqlcmd.h"
-#include <zlib.h>
+#if defined(FOSSIL_ENABLE_MINIZ)
+#  define MINIZ_HEADER_FILE_ONLY
+#  include "miniz.c"
+#else
+#  include <zlib.h>
+#endif
 
 /*
 ** Implementation of the "content(X)" SQL function.  Return the complete
@@ -44,7 +49,7 @@ static void sqlcmd_content(
   rid = name_to_rid(zName);
   if( rid==0 ) return;
   if( content_get(rid, &cx) ){
-    sqlite3_result_blob(context, blob_buffer(&cx), blob_size(&cx), 
+    sqlite3_result_blob(context, blob_buffer(&cx), blob_size(&cx),
                                  SQLITE_TRANSIENT);
     blob_reset(&cx);
   }
@@ -63,6 +68,7 @@ static void sqlcmd_compress(
   unsigned char *pOut;
   unsigned int nIn;
   unsigned long int nOut;
+  int rc;
 
   pIn = sqlite3_value_blob(argv[0]);
   nIn = sqlite3_value_bytes(argv[0]);
@@ -72,8 +78,13 @@ static void sqlcmd_compress(
   pOut[1] = nIn>>16 & 0xff;
   pOut[2] = nIn>>8 & 0xff;
   pOut[3] = nIn & 0xff;
-  compress(&pOut[4], &nOut, pIn, nIn);
-  sqlite3_result_blob(context, pOut, nOut+4, sqlite3_free);
+  rc = compress(&pOut[4], &nOut, pIn, nIn);
+  if( rc==Z_OK ){
+    sqlite3_result_blob(context, pOut, nOut+4, sqlite3_free);
+  }else{
+    sqlite3_free(pOut);
+    sqlite3_result_error(context, "input cannot be zlib compressed", -1);
+  }
 }
 
 /*
@@ -100,12 +111,27 @@ static void sqlcmd_decompress(
   if( rc==Z_OK ){
     sqlite3_result_blob(context, pOut, nOut, sqlite3_free);
   }else{
+    sqlite3_free(pOut);
     sqlite3_result_error(context, "input is not zlib compressed", -1);
   }
 }
 
 /*
-** This is the "automatic extensionn" initializer that runs right after
+** Add the content(), compress(), and decompress() SQL functions to
+** database connection db.
+*/
+int add_content_sql_commands(sqlite3 *db){
+  sqlite3_create_function(db, "content", 1, SQLITE_UTF8, 0,
+                          sqlcmd_content, 0, 0);
+  sqlite3_create_function(db, "compress", 1, SQLITE_UTF8, 0,
+                          sqlcmd_compress, 0, 0);
+  sqlite3_create_function(db, "decompress", 1, SQLITE_UTF8, 0,
+                          sqlcmd_decompress, 0, 0);
+  return SQLITE_OK;
+}
+
+/*
+** This is the "automatic extension" initializer that runs right after
 ** the connection to the repository database is opened.  Set up the
 ** database connection to be more useful to the human operator.
 */
@@ -114,38 +140,71 @@ static int sqlcmd_autoinit(
   const char **pzErrMsg,
   const void *notUsed
 ){
-  sqlite3_create_function(db, "content", 1, SQLITE_ANY, 0,
-                          sqlcmd_content, 0, 0);
-  sqlite3_create_function(db, "compress", 1, SQLITE_ANY, 0,
-                          sqlcmd_compress, 0, 0);
-  sqlite3_create_function(db, "decompress", 1, SQLITE_ANY, 0,
-                          sqlcmd_decompress, 0, 0);
+  add_content_sql_commands(db);
+  db_add_aux_functions(db);
+  re_add_sql_func(db);
+  search_sql_setup(db);
+  foci_register(db);
   g.repositoryOpen = 1;
   g.db = db;
   return SQLITE_OK;
 }
 
-
 /*
 ** COMMAND: sqlite3
 **
-** Usage: %fossil sqlite3 ?DATABASE? ?OPTIONS?
+** Usage: %fossil sqlite3 ?FOSSIL_OPTS? ?DATABASE? ?SHELL_OPTS?
 **
-** Run the standalone sqlite3 command-line shell on DATABASE with OPTIONS.
+** Run the standalone sqlite3 command-line shell on DATABASE with SHELL_OPTS.
 ** If DATABASE is omitted, then the repository that serves the working
-** directory is opened.
+** directory is opened.  See https://www.sqlite.org/cli.html for additional
+** information.
+**
+** Fossil Options:
+**
+**    --no-repository           Skip opening the repository database.
 **
 ** WARNING:  Careless use of this command can corrupt a Fossil repository
 ** in ways that are unrecoverable.  Be sure you know what you are doing before
-** running any SQL commands that modifies the repository database.
+** running any SQL commands that modify the repository database.
+**
+** The following extensions to the usual SQLite commands are provided:
+**
+**    content(X)                Return the content of artifact X.  X can be a
+**                              SHA1 hash or prefix or a tag.
+**
+**    compress(X)               Compress text X.
+**
+**    decompress(X)             Decompress text X.  Undoes the work of
+**                              compress(X).
+**
+**    checkin_mtime(X,Y)        Return the mtime for the file Y (a BLOB.RID)
+**                              found in check-in X (another BLOB.RID value).
+**
+**    symbolic_name_to_rid(X)   Return the BLOB.RID corresponding to symbolic
+**                              name X.
+**
+**    now()                     Return the number of seconds since 1970.
+**
+**    REGEXP                    The REGEXP operator works, unlike in
+**                              standard SQLite.
+**
+**    files_of_checkin(X)       A table-valued function that returns info on
+**                              all files contained in check-in X.  Example:
+**                                SELECT * FROM files_of_checkin('trunk');
 */
-void sqlite3_cmd(void){
+void cmd_sqlite3(void){
+  int noRepository;
   extern int sqlite3_shell(int, char**);
-  db_find_and_open_repository(OPEN_ANY_SCHEMA, 0);
-  db_close(1);
+  noRepository = find_option("no-repository", 0, 0)!=0;
+  if( !noRepository ){
+    db_find_and_open_repository(OPEN_ANY_SCHEMA, 0);
+  }
+  fossil_close(1, noRepository);
   sqlite3_shutdown();
   sqlite3_shell(g.argc-1, g.argv+1);
-  g.db = 0;
+  sqlite3_cancel_auto_extension((void(*)(void))sqlcmd_autoinit);
+  fossil_close(0, noRepository);
 }
 
 /*
@@ -156,3 +215,15 @@ void fossil_open(const char **pzRepoName){
   sqlite3_auto_extension((void(*)(void))sqlcmd_autoinit);
   *pzRepoName = g.zRepositoryName;
 }
+
+/*
+** This routine closes the Fossil databases and/or invalidates the global
+** state variables that keep track of them.
+*/
+void fossil_close(int bDb, int noRepository){
+  if( bDb ) db_close(1);
+  if( noRepository ) g.zRepositoryName = 0;
+  g.db = 0;
+  g.repositoryOpen = 0;
+  g.localOpen = 0;
+}
