@@ -24,14 +24,14 @@
 /*
 ** SQL code to implement the tables needed by the stash.
 */
-static const char zStashInit[] = 
-@ CREATE TABLE IF NOT EXISTS %s.stash(
+static const char zStashInit[] =
+@ CREATE TABLE IF NOT EXISTS localdb.stash(
 @   stashid INTEGER PRIMARY KEY,     -- Unique stash identifier
-@   vid INTEGER,                     -- The baseline check-out for this stash
+@   vid INTEGER,                     -- The baseline checkout for this stash
 @   comment TEXT,                    -- Comment for this stash.  Or NULL
 @   ctime TIMESTAMP                  -- When the stash was created
 @ );
-@ CREATE TABLE IF NOT EXISTS %s.stashfile(
+@ CREATE TABLE IF NOT EXISTS localdb.stashfile(
 @   stashid INTEGER REFERENCES stash,  -- Stash that contains this file
 @   rid INTEGER,                       -- Baseline content in BLOB table or 0.
 @   isAdded BOOLEAN,                   -- True if this is an added file
@@ -41,7 +41,7 @@ static const char zStashInit[] =
 @   origname TEXT,                     -- Original filename
 @   newname TEXT,                      -- New name for file at next check-in
 @   delta BLOB,                        -- Delta from baseline. Content if rid=0
-@   PRIMARY KEY(origname, stashid)
+@   PRIMARY KEY(newname, stashid)
 @ );
 @ INSERT OR IGNORE INTO vvar(name, value) VALUES('stash-next', 1);
 ;
@@ -60,23 +60,23 @@ static void stash_add_file_or_dir(int stashid, int vid, const char *zFName){
   Stmt ins;             /* Insert statement */
 
   zFile = mprintf("%/", zFName);
-  file_tree_name(zFile, &fname, 1);
+  file_tree_name(zFile, &fname, 0, 1);
   zTreename = blob_str(&fname);
   blob_zero(&sql);
-  blob_appendf(&sql,
+  blob_append_sql(&sql,
     "SELECT deleted, isexe, islink, mrid, pathname, coalesce(origname,pathname)"
     "  FROM vfile"
     " WHERE vid=%d AND (chnged OR deleted OR origname NOT NULL OR mrid==0)",
     vid
   );
   if( fossil_strcmp(zTreename,".")!=0 ){
-    blob_appendf(&sql,
+    blob_append_sql(&sql,
       "   AND (pathname GLOB '%q/*' OR origname GLOB '%q/*'"
             "  OR pathname=%Q OR origname=%Q)",
       zTreename, zTreename, zTreename, zTreename
     );
   }
-  db_prepare(&q, blob_str(&sql));
+  db_prepare(&q, "%s", blob_sql_text(&sql));
   blob_reset(&sql);
   db_prepare(&ins,
      "INSERT INTO stashfile(stashid, rid, isAdded, isRemoved, isExec, isLink,"
@@ -116,7 +116,7 @@ static void stash_add_file_or_dir(int stashid, int vid, const char *zFName){
       /* A modified file */
       Blob orig;
       Blob disk;
-      
+
       if( isNewLink ){
         blob_read_link(&disk, zPath);
       }else{
@@ -160,11 +160,17 @@ static int stash_create(void){
   if( zComment==0 ){
     Blob prompt;                       /* Prompt for stash comment */
     Blob comment;                      /* User comment reply */
+#if defined(_WIN32) || defined(__CYGWIN__)
+    int bomSize;
+    const unsigned char *bom = get_utf8_bom(&bomSize);
+    blob_init(&prompt, (const char *) bom, bomSize);
+#else
     blob_zero(&prompt);
+#endif
     blob_append(&prompt,
        "\n"
        "# Enter a description of what is being stashed.  Lines beginning\n"
-       "# with \"#\" are ignored.  Stash comments are plain text except.\n"
+       "# with \"#\" are ignored.  Stash comments are plain text except\n"
        "# newlines are not preserved.\n",
        -1);
     prompt_for_user_comment(&comment, &prompt);
@@ -192,15 +198,19 @@ static int stash_create(void){
 }
 
 /*
-** Apply a stash to the current check-out.
+** Apply a stash to the current checkout.
 */
 static void stash_apply(int stashid, int nConflict){
+  int vid;
   Stmt q;
   db_prepare(&q,
      "SELECT rid, isRemoved, isExec, isLink, origname, newname, delta"
      "  FROM stashfile WHERE stashid=%d",
      stashid
   );
+  vid = db_lget_int("checkout",0);
+  db_multi_exec("CREATE TEMP TABLE sfile(x TEXT PRIMARY KEY %s)",
+                filename_collation());
   while( db_step(&q)==SQLITE_ROW ){
     int rid = db_column_int(&q, 0);
     int isRemoved = db_column_int(&q, 1);
@@ -214,10 +224,10 @@ static void stash_apply(int stashid, int nConflict){
     undo_save(zNew);
     blob_zero(&delta);
     if( rid==0 ){
+      db_multi_exec("INSERT OR IGNORE INTO sfile(x) VALUES(%Q)", zNew);
       db_ephemeral_blob(&q, 6, &delta);
       blob_write_to_file(&delta, zNPath);
       file_wd_setexe(zNPath, isExec);
-      fossil_print("ADD %s\n", zNew);
     }else if( isRemoved ){
       fossil_print("DELETE %s\n", zOrig);
       file_delete(zOPath);
@@ -239,7 +249,7 @@ static void stash_apply(int stashid, int nConflict){
         if( isLink ){
           symlink_create(blob_str(&b), zNPath);
         }else{
-          blob_write_to_file(&b, zNPath);          
+          blob_write_to_file(&b, zNPath);
         }
         file_wd_setexe(zNPath, isExec);
         fossil_print("UPDATE %s\n", zNew);
@@ -250,8 +260,8 @@ static void stash_apply(int stashid, int nConflict){
           blob_zero(&b); /* because we reset it later */
           fossil_print("***** Cannot merge symlink %s\n", zNew);
         }else{
-          rc = merge_3way(&a, zOPath, &b, &out);
-          blob_write_to_file(&out, zNPath);          
+          rc = merge_3way(&a, zOPath, &b, &out, 0);
+          blob_write_to_file(&out, zNPath);
           blob_reset(&out);
           file_wd_setexe(zNPath, isExec);
         }
@@ -270,8 +280,14 @@ static void stash_apply(int stashid, int nConflict){
     if( fossil_strcmp(zOrig,zNew)!=0 ){
       undo_save(zOrig);
       file_delete(zOPath);
+      db_multi_exec(
+        "UPDATE vfile SET pathname='%q', origname='%q'"
+        " WHERE pathname='%q' %s AND vid=%d",
+        zNew, zOrig, zOrig, filename_collation(), vid
+      );
     }
   }
+  stash_add_files_in_sfile(vid);
   db_finalize(&q);
   if( nConflict ){
     fossil_print(
@@ -307,7 +323,7 @@ static void stash_diff(
     const char *zOrig = db_column_text(&q, 4);
     const char *zNew = db_column_text(&q, 5);
     char *zOPath = mprintf("%s%s", g.zLocalRoot, zOrig);
-    Blob delta, a, b, disk;
+    Blob a, b;
     if( rid==0 ){
       db_ephemeral_blob(&q, 6, &a);
       fossil_print("ADDED %s\n", zNew);
@@ -333,13 +349,14 @@ static void stash_diff(
       diff_file_mem(&a, &empty, isBin1, isBin2, zOrig, zDiffCmd,
                     zBinGlob, fIncludeBinary, diffFlags);
     }else{
+      Blob delta, disk;
       int isOrigLink = file_wd_islink(zOPath);
       db_ephemeral_blob(&q, 6, &delta);
       if( fBaseline==0 ){
         if( isOrigLink ){
           blob_read_link(&disk, zOPath);
         }else{
-          blob_read_from_file(&disk, zOPath);        
+          blob_read_from_file(&disk, zOPath);
         }
       }
       fossil_print("CHANGED %s\n", zNew);
@@ -359,8 +376,8 @@ static void stash_diff(
         blob_reset(&b);
       }
       if( !fBaseline ) blob_reset(&disk);
+      blob_reset(&delta);
     }
-    blob_reset(&delta);
  }
   db_finalize(&q);
 }
@@ -383,14 +400,14 @@ static void stash_drop(int stashid){
 ** throw an error if the stash is empty.
 */
 static int stash_get_id(const char *zStashId){
-  int stashid = 0;
+  int stashid;
   if( zStashId==0 ){
     stashid = db_int(0, "SELECT max(stashid) FROM stash");
     if( stashid==0 ) fossil_fatal("empty stash");
   }else{
     stashid = atoi(zStashId);
     if( !db_exists("SELECT 1 FROM stash WHERE stashid=%d", stashid) ){
-      fossil_fatal("no such stash: %d\n", stashid);
+      fossil_fatal("no such stash: %s", zStashId);
     }
   }
   return stashid;
@@ -402,31 +419,30 @@ static int stash_get_id(const char *zStashId){
 ** Usage: %fossil stash SUBCOMMAND ARGS...
 **
 **  fossil stash
-**  fossil stash save ?-m COMMENT? ?FILES...?
-**  fossil stash snapshot ?-m COMMENT? ?FILES...?
+**  fossil stash save ?-m|--comment COMMENT? ?FILES...?
+**  fossil stash snapshot ?-m|--comment COMMENT? ?FILES...?
 **
 **     Save the current changes in the working tree as a new stash.
 **     Then revert the changes back to the last check-in.  If FILES
 **     are listed, then only stash and revert the named files.  The
 **     "save" verb can be omitted if and only if there are no other
 **     arguments.  The "snapshot" verb works the same as "save" but
-**     omits the revert, keeping the check-out unchanged.
+**     omits the revert, keeping the checkout unchanged.
 **
-**  fossil stash list ?--detail?
-**  fossil stash ls ?-l?
+**  fossil stash list|ls ?-v|--verbose? ?-W|--width <num>?
 **
 **     List all changes sets currently stashed.  Show information about
-**     individual files in each changeset if --detail or -l is used.
+**     individual files in each changeset if -v or --verbose is used.
 **
-**  fossil stash show ?STASHID? ?DIFF-FLAGS?
+**  fossil stash show|cat ?STASHID? ?DIFF-OPTIONS?
 **
-**     Show the content of a stash
+**     Show the contents of a stash.
 **
 **  fossil stash pop
 **  fossil stash apply ?STASHID?
 **
 **     Apply STASHID or the most recently create stash to the current
-**     working check-out.  The "pop" command deletes that changeset from
+**     working checkout.  The "pop" command deletes that changeset from
 **     the stash after applying it but the "apply" command retains the
 **     changeset.
 **
@@ -436,42 +452,54 @@ static int stash_get_id(const char *zStashId){
 **     changes of STASHID.  Keep STASHID so that it can be reused
 **     This command is undoable.
 **
-**  fossil stash drop ?STASHID? ?--all?
-**  fossil stash rm   ?STASHID? ?--all?
+**  fossil stash drop|rm ?STASHID? ?-a|--all?
 **
 **     Forget everything about STASHID.  Forget the whole stash if the
-**     --all flag is used.  Individual drops are undoable but --all is not.
+**     -a|--all flag is used.  Individual drops are undoable but -a|--all
+**     is not.
 **
-**  fossil stash diff ?STASHID?
-**  fossil stash gdiff ?STASHID?
+**  fossil stash diff ?STASHID? ?DIFF-OPTIONS?
+**  fossil stash gdiff ?STASHID? ?DIFF-OPTIONS?
 **
 **     Show diffs of the current working directory and what that
-**     directory would be if STASHID were applied.  
+**     directory would be if STASHID were applied.
 **
 ** SUMMARY:
 **  fossil stash
-**  fossil stash save ?-m COMMENT? ?FILES...?
-**  fossil stash snapshot ?-m COMMENT? ?FILES...?
-**  fossil stash list|ls  ?-l? ?--detail?
-**  fossil stash show ?STASHID? ?DIFF-OPTIONS?
+**  fossil stash save ?-m|--comment COMMENT? ?FILES...?
+**  fossil stash snapshot ?-m|--comment COMMENT? ?FILES...?
+**  fossil stash list|ls ?-v|--verbose? ?-W|--width <num>?
+**  fossil stash show|cat ?STASHID? ?DIFF-OPTIONS?
 **  fossil stash pop
-**  fossil stash apply ?STASHID?
-**  fossil stash goto ?STASHID?
-**  fossil stash rm|drop ?STASHID? ?--all?
-**  fossil stash [g]diff ?STASHID? ?DIFF-OPTIONS?
+**  fossil stash apply|goto ?STASHID?
+**  fossil stash drop|rm ?STASHID? ?-a|--all?
+**  fossil stash diff ?STASHID? ?DIFF-OPTIONS?
+**  fossil stash gdiff ?STASHID? ?DIFF-OPTIONS?
 */
 void stash_cmd(void){
-  const char *zDb;
   const char *zCmd;
   int nCmd;
-  int stashid;
-
+  int stashid = 0;
+  int rc;
   undo_capture_command_line();
   db_must_be_within_tree();
-  db_open_config(0);
+  db_open_config(0, 0);
   db_begin_transaction();
-  zDb = db_name("localdb");
-  db_multi_exec(zStashInit, zDb, zDb);
+  db_multi_exec(zStashInit /*works-like:""*/);
+  rc = db_exists("SELECT 1 FROM sqlite_master"
+                 " WHERE name='stashfile'"
+                 "   AND sql GLOB '* PRIMARY KEY(origname, stashid)*'");
+  if( rc!=0 ){
+    db_multi_exec(
+      "CREATE TABLE localdb.stashfile_tmp AS SELECT * FROM stashfile;"
+      "DROP TABLE stashfile;"
+    );
+    db_multi_exec(zStashInit /*works-like:""*/);
+    db_multi_exec(
+      "INSERT INTO stashfile SELECT * FROM stashfile_tmp;"
+      "DROP TABLE stashfile_tmp;"
+    );
+  }
   if( g.argc<=2 ){
     zCmd = "save";
   }else{
@@ -493,6 +521,7 @@ void stash_cmd(void){
       }
       db_finalize(&q);
       newArgv[0] = g.argv[0];
+      newArgv[1] = 0;
       g.argv = newArgv;
       g.argc = nFile+2;
       if( nFile==0 ) return;
@@ -505,15 +534,28 @@ void stash_cmd(void){
   }else
   if( memcmp(zCmd, "list", nCmd)==0 || memcmp(zCmd, "ls", nCmd)==0 ){
     Stmt q, q2;
-    int n = 0;
-    int fDetail = find_option("detail","l",0)!=0;
+    int n = 0, width;
+    int verboseFlag = find_option("verbose","v",0)!=0;
+    const char *zWidth = find_option("width","W",1);
+
+    if( zWidth ){
+      width = atoi(zWidth);
+      if( (width!=0) && (width<=46) ){
+        fossil_fatal("-W|--width value must be >46 or 0");
+      }
+    }else{
+      width = -1;
+    }
+    if( !verboseFlag ){
+      verboseFlag = find_option("detail","l",0)!=0; /* deprecated */
+    }
     verify_all_options();
     db_prepare(&q,
        "SELECT stashid, (SELECT uuid FROM blob WHERE rid=vid),"
        "       comment, datetime(ctime) FROM stash"
-       " ORDER BY ctime DESC"
+       " ORDER BY ctime"
     );
-    if( fDetail ){
+    if( verboseFlag ){
       db_prepare(&q2, "SELECT isAdded, isRemoved, origname, newname"
                       "  FROM stashfile WHERE stashid=$id");
     }
@@ -529,9 +571,9 @@ void stash_cmd(void){
       zCom = db_column_text(&q, 2);
       if( zCom && zCom[0] ){
         fossil_print("       ");
-        comment_print(zCom, 7, 79);
+        comment_print(zCom, 0, 7, width, g.comFmtFlags);
       }
-      if( fDetail ){
+      if( verboseFlag ){
         db_bind_int(&q2, "$id", stashid);
         while( db_step(&q2)==SQLITE_ROW ){
           int isAdded = db_column_int(&q2, 0);
@@ -552,23 +594,31 @@ void stash_cmd(void){
       }
     }
     db_finalize(&q);
-    if( fDetail ) db_finalize(&q2);
+    if( verboseFlag ) db_finalize(&q2);
     if( n==0 ) fossil_print("empty stash\n");
   }else
   if( memcmp(zCmd, "drop", nCmd)==0 || memcmp(zCmd, "rm", nCmd)==0 ){
-    int allFlag = find_option("all", 0, 0)!=0;
-    if( g.argc>4 ) usage("drop STASHID");
+    int allFlag = find_option("all", "a", 0)!=0;
     if( allFlag ){
       Blob ans;
-      blob_zero(&ans);
+      char cReply;
       prompt_user("This action is not undoable.  Continue (y/N)? ", &ans);
-      if( blob_str(&ans)[0]=='y' ){
+      cReply = blob_str(&ans)[0];
+      if( cReply=='y' || cReply=='Y' ){
         db_multi_exec("DELETE FROM stash; DELETE FROM stashfile;");
       }
-    }else{
-      stashid = stash_get_id(g.argc==4 ? g.argv[3] : 0);
+    }else if( g.argc>=4 ){
+      int i;
       undo_begin();
-      undo_save_stash(stashid);
+      for(i=3; i<g.argc; i++){
+        stashid = stash_get_id(g.argv[i]);
+        undo_save_stash(stashid);
+        stash_drop(stashid);
+      }
+      undo_finish();
+    }else{
+      undo_begin();
+      undo_save_stash(0);
       stash_drop(stashid);
       undo_finish();
     }
@@ -592,7 +642,6 @@ void stash_cmd(void){
   if( memcmp(zCmd, "goto", nCmd)==0 ){
     int nConflict;
     int vid;
-
     if( g.argc>4 ) usage("apply STASHID");
     stashid = stash_get_id(g.argc==4 ? g.argv[3] : 0);
     undo_begin();
@@ -607,28 +656,31 @@ void stash_cmd(void){
   if( memcmp(zCmd, "diff", nCmd)==0
    || memcmp(zCmd, "gdiff", nCmd)==0
    || memcmp(zCmd, "show", nCmd)==0
+   || memcmp(zCmd, "cat", nCmd)==0
   ){
     const char *zDiffCmd = 0;
     const char *zBinGlob = 0;
     int fIncludeBinary = 0;
+    int fBaseline = zCmd[0]=='s' || zCmd[0]=='c';
     u64 diffFlags;
 
     if( find_option("tk",0,0)!=0 ){
       db_close(0);
-      diff_tk((zCmd[0]=='s' ? "stash show" : "stash diff"), 3);
+      diff_tk(fBaseline ? "stash show" : "stash diff", 3);
       return;
     }
     if( find_option("internal","i",0)==0 ){
-      zDiffCmd = diff_command_external(0);
+      zDiffCmd = diff_command_external(memcmp(zCmd, "gdiff", nCmd)==0);
     }
     diffFlags = diff_options();
-    if( g.argc>4 ) usage(mprintf("%s STASHID", zCmd));
+    if( find_option("verbose","v",0)!=0 ) diffFlags |= DIFF_VERBOSE;
+    if( g.argc>4 ) usage(mprintf("%s ?STASHID? ?DIFF-OPTIONS?", zCmd));
     if( zDiffCmd ){
       zBinGlob = diff_get_binary_glob();
       fIncludeBinary = diff_include_binary_files();
     }
     stashid = stash_get_id(g.argc==4 ? g.argv[3] : 0);
-    stash_diff(stashid, zDiffCmd, zBinGlob, zCmd[0]=='s', fIncludeBinary,
+    stash_diff(stashid, zDiffCmd, zBinGlob, fBaseline, fIncludeBinary,
                diffFlags);
   }else
   if( memcmp(zCmd, "help", nCmd)==0 ){
