@@ -52,7 +52,7 @@ static void tag_propagate(
   ** Three returns:  (1) rid of the child.  (2) timestamp of child.
   ** (3) True to propagate or false to block.
   */
-  db_prepare(&s, 
+  db_prepare(&s,
      "SELECT cid, plink.mtime,"
      "       coalesce(srcid=0 AND tagxref.mtime<:mtime, %d) AS doit"
      "  FROM plink LEFT JOIN tagxref ON cid=rid AND tagid=%d"
@@ -62,7 +62,7 @@ static void tag_propagate(
   db_bind_double(&s, ":mtime", mtime);
 
   if( tagType==2 ){
-    /* Set the propagated tag marker on checkin :rid */
+    /* Set the propagated tag marker on check-in :rid */
     db_prepare(&ins,
        "REPLACE INTO tagxref(tagid, tagtype, srcid, origid, value, mtime, rid)"
        "VALUES(%d,2,0,%d,%Q,:mtime,:rid)",
@@ -70,7 +70,7 @@ static void tag_propagate(
     );
     db_bind_double(&ins, ":mtime", mtime);
   }else{
-    /* Remove all references to the tag from checkin :rid */
+    /* Remove all references to the tag from check-in :rid */
     zValue = 0;
     db_prepare(&ins,
        "DELETE FROM tagxref WHERE tagid=%d AND rid=:rid", tagid
@@ -150,6 +150,9 @@ int tag_findid(const char *zTag, int createFlag){
 
 /*
 ** Insert a tag into the database.
+**
+** Also translate zTag into a tagid and return the tagid.  (In other words
+** if zTag is "bgcolor" then return TAG_BGCOLOR.)
 */
 int tag_insert(
   const char *zTag,        /* Name of the tag (w/o the "+" or "-" prefix */
@@ -181,7 +184,7 @@ int tag_insert(
     /* Another entry that is more recent already exists.  Do nothing */
     return tagid;
   }
-  db_prepare(&s, 
+  db_prepare(&s,
     "REPLACE INTO tagxref(tagid,tagtype,srcId,origid,value,mtime,rid)"
     " VALUES(%d,%d,%d,%d,%Q,:mtime,%d)",
     tagid, tagtype, srcId, rid, zValue, rid
@@ -215,7 +218,8 @@ int tag_insert(
     }
   }
   if( zCol ){
-    db_multi_exec("UPDATE event SET %s=%Q WHERE objid=%d", zCol, zValue, rid);
+    db_multi_exec("UPDATE event SET \"%w\"=%Q WHERE objid=%d",
+                  zCol, zValue, rid);
     if( tagid==TAG_COMMENT ){
       char *zCopy = mprintf("%s", zValue);
       wiki_extract_links(zCopy, rid, 0, mtime, 1, WIKI_INLINE);
@@ -229,6 +233,9 @@ int tag_insert(
                   " WHERE objid=%d",
                   zValue, rid);
   }
+  if( tagid==TAG_PARENT && tagtype==1 ){
+    manifest_reparent_checkin(rid, zValue);
+  }
   if( tagtype==1 ) tagtype = 0;
   tag_propagate(rid, tagid, tagtype, rid, zValue, mtime);
   return tagid;
@@ -237,7 +244,8 @@ int tag_insert(
 
 /*
 ** COMMAND: test-tag
-** %fossil test-tag (+|*|-)TAGNAME ARTIFACT-ID ?VALUE?
+**
+** Usage: %fossil test-tag (+|*|-)TAGNAME ARTIFACT-ID ?VALUE?
 **
 ** Add a tag or anti-tag to the rebuildable tables of the local repository.
 ** No tag artifact is created so the new tag is erased the next
@@ -258,7 +266,7 @@ void testtag_cmd(void){
     case '+':  tagtype = 1;  break;
     case '*':  tagtype = 2;  break;
     case '-':  tagtype = 0;  break;
-    default:   
+    default:
       fossil_fatal("tag should begin with '+', '*', or '-'");
       return;
   }
@@ -270,12 +278,25 @@ void testtag_cmd(void){
   zValue = g.argc==5 ? g.argv[4] : 0;
   db_begin_transaction();
   tag_insert(zTag, tagtype, zValue, -1, 0.0, rid);
-  db_end_transaction(0); 
+  db_end_transaction(0);
 }
 
+/*
+** OR this value into the tagtype argument to tag_add_artifact to
+** cause the tag to be displayed on standard output rather than be
+** inserted.  Used for --dryrun options and debugging.
+*/
+#if INTERFACE
+#define TAG_ADD_DRYRUN  0x04
+#endif
+
 /*
 ** Add a control record to the repository that either creates
 ** or cancels a tag.
+**
+** tagtype should normally be 0, 1, or 2.  But if the TAG_ADD_DRYRUN bit
+** is also set, then simply print the text of the tag on standard output
+** (for testing purposes) rather than create the tag.
 */
 void tag_add_artifact(
   const char *zPrefix,        /* Prefix to prepend to tag name */
@@ -293,7 +314,12 @@ void tag_add_artifact(
   Blob ctrl;
   Blob cksum;
   static const char zTagtype[] = { '-', '+', '*' };
+  int dryRun = 0;
 
+  if( tagtype & TAG_ADD_DRYRUN ){
+    tagtype &= ~TAG_ADD_DRYRUN;
+    dryRun = 1;
+  }
   assert( tagtype>=0 && tagtype<=2 );
   user_select();
   blob_zero(&uuid);
@@ -324,38 +350,64 @@ void tag_add_artifact(
   }else{
     blob_appendf(&ctrl, "\n");
   }
-  blob_appendf(&ctrl, "U %F\n", zUserOvrd ? zUserOvrd : g.zLogin);
+  blob_appendf(&ctrl, "U %F\n", zUserOvrd ? zUserOvrd : login_name());
   md5sum_blob(&ctrl, &cksum);
   blob_appendf(&ctrl, "Z %b\n", &cksum);
-  nrid = content_put(&ctrl);
-  manifest_crosslink(nrid, &ctrl);
+  if( dryRun ){
+    fossil_print("%s", blob_str(&ctrl));
+    blob_reset(&ctrl);
+  }else{
+    nrid = content_put(&ctrl);
+    manifest_crosslink(nrid, &ctrl, MC_PERMIT_HOOKS);
+  }
   assert( blob_is_reset(&ctrl) );
+  manifest_to_disk(rid);
 }
 
 /*
 ** COMMAND: tag
+**
 ** Usage: %fossil tag SUBCOMMAND ...
 **
-** Run various subcommands to control tags and properties
+** Run various subcommands to control tags and properties.
 **
-**     %fossil tag add ?--raw? ?--propagate? TAGNAME CHECK-IN ?VALUE?
+**     %fossil tag add ?OPTIONS? TAGNAME CHECK-IN ?VALUE?
 **
 **         Add a new tag or property to CHECK-IN. The tag will
 **         be usable instead of a CHECK-IN in commands such as
 **         update and merge.  If the --propagate flag is present,
-**         the tag value propages to all descendants of CHECK-IN
+**         the tag value propagates to all descendants of CHECK-IN
+**
+**         Options:
+**           --raw                       Raw tag name.
+**           --propagate                 Propagating tag.
+**           --date-override DATETIME    Set date and time added.
+**           --user-override USER        Name USER when adding the tag.
+**           --dryrun|-n                 Display the tag text, but to not
+**                                       actually insert it into the database.
+**
+**         The --date-override and --user-override options support
+**         importing history from other SCM systems. DATETIME has
+**         the form 'YYYY-MMM-DD HH:MM:SS'.
 **
 **     %fossil tag cancel ?--raw? TAGNAME CHECK-IN
 **
 **         Remove the tag TAGNAME from CHECK-IN, and also remove
-**         the propagation of the tag to any descendants.
+**         the propagation of the tag to any descendants.  Use the
+**         the --dryrun or -n options to see what would have happened.
 **
-**     %fossil tag find ?--raw? ?--type TYPE? TAGNAME
+**     %fossil tag find ?OPTIONS? TAGNAME
 **
 **         List all objects that use TAGNAME.  TYPE can be "ci" for
-**         checkins or "e" for events.
+**         check-ins or "e" for events. The limit option limits the number
+**         of results to the given value.
 **
-**     %fossil tag list ?--raw? ?CHECK-IN?
+**         Options:
+**           --raw           Raw tag name.
+**           -t|--type TYPE  One of "ci", or "e".
+**           -n|--limit N    Limit to N results.
+**
+**     %fossil tag list|ls ?--raw? ?CHECK-IN?
 **
 **         List all tags, or if CHECK-IN is supplied, list
 **         all tags and their values for CHECK-IN.
@@ -379,16 +431,14 @@ void tag_add_artifact(
 **
 ** will assume that "decaf" is a tag/branch name.
 **
-** only allow --date-override and --user-override in 
-**   %fossil tag add --date-override 'YYYY-MMM-DD HH:MM:SS' \\
-**                   --user-override user 
-** in order to import history from other scm systems
 */
 void tag_cmd(void){
   int n;
   int fRaw = find_option("raw","",0)!=0;
   int fPropagate = find_option("propagate","",0)!=0;
   const char *zPrefix = fRaw ? "" : "sym-";
+  const char *zFindLimit = find_option("limit","n",1);
+  const int nFindLimit = zFindLimit ? atoi(zFindLimit) : -2000;
 
   db_find_and_open_repository(0, 0);
   if( g.argc<3 ){
@@ -401,15 +451,17 @@ void tag_cmd(void){
 
   if( strncmp(g.argv[2],"add",n)==0 ){
     char *zValue;
+    int dryRun = 0;
     const char *zDateOvrd = find_option("date-override",0,1);
     const char *zUserOvrd = find_option("user-override",0,1);
+    if( find_option("dryrun","n",0)!=0 ) dryRun = TAG_ADD_DRYRUN;
     if( g.argc!=5 && g.argc!=6 ){
-      usage("add ?--raw? ?--propagate? TAGNAME CHECK-IN ?VALUE?");
+      usage("add ?options? TAGNAME CHECK-IN ?VALUE?");
     }
     zValue = g.argc==6 ? g.argv[5] : 0;
     db_begin_transaction();
     tag_add_artifact(zPrefix, g.argv[3], g.argv[4], zValue,
-                     1+fPropagate,zDateOvrd,zUserOvrd);
+                     1+fPropagate+dryRun,zDateOvrd,zUserOvrd);
     db_end_transaction(0);
   }else
 
@@ -419,29 +471,37 @@ void tag_cmd(void){
   }else
 
   if( strncmp(g.argv[2],"cancel",n)==0 ){
+    int dryRun = 0;
+    if( find_option("dryrun","n",0)!=0 ) dryRun = TAG_ADD_DRYRUN;
     if( g.argc!=5 ){
-      usage("cancel ?--raw? TAGNAME CHECK-IN");
+      usage("cancel ?options? TAGNAME CHECK-IN");
     }
     db_begin_transaction();
-    tag_add_artifact(zPrefix, g.argv[3], g.argv[4], 0, 0, 0, 0);
+    tag_add_artifact(zPrefix, g.argv[3], g.argv[4], 0, dryRun, 0, 0);
     db_end_transaction(0);
   }else
 
   if( strncmp(g.argv[2],"find",n)==0 ){
     Stmt q;
     const char *zType = find_option("type","t",1);
+    Blob sql = empty_blob;
     if( zType==0 || zType[0]==0 ) zType = "*";
     if( g.argc!=4 ){
-      usage("find ?--raw? TAGNAME");
+      usage("find ?--raw? ?-t|--type TYPE? ?-n|--limit #? TAGNAME");
     }
     if( fRaw ){
-      db_prepare(&q,
+      blob_append_sql(&sql,
         "SELECT blob.uuid FROM tagxref, blob"
         " WHERE tagid=(SELECT tagid FROM tag WHERE tagname=%Q)"
         "   AND tagxref.tagtype>0"
         "   AND blob.rid=tagxref.rid",
         g.argv[3]
       );
+      if( nFindLimit>0 ){
+        blob_append_sql(&sql, " LIMIT %d", nFindLimit);
+      }
+      db_prepare(&q, "%s", blob_sql_text(&sql));
+      blob_reset(&sql);
       while( db_step(&q)==SQLITE_ROW ){
         fossil_print("%s\n", db_column_text(&q, 0));
       }
@@ -450,7 +510,7 @@ void tag_cmd(void){
       int tagid = db_int(0, "SELECT tagid FROM tag WHERE tagname='sym-%q'",
                          g.argv[3]);
       if( tagid>0 ){
-        db_prepare(&q,
+        blob_append_sql(&sql,
           "%s"
           "  AND event.type GLOB '%q'"
           "  AND blob.rid IN ("
@@ -460,16 +520,18 @@ void tag_cmd(void){
           " ORDER BY event.mtime DESC",
           timeline_query_for_tty(), zType, tagid
         );
-        print_timeline(&q, 2000, 0);
+        db_prepare(&q, "%s", blob_sql_text(&sql));
+        blob_reset(&sql);
+        print_timeline(&q, nFindLimit, 79, 0);
         db_finalize(&q);
       }
     }
   }else
 
-  if( strncmp(g.argv[2],"list",n)==0 ){
+  if(( strncmp(g.argv[2],"list",n)==0 )||( strncmp(g.argv[2],"ls",n)==0 )){
     Stmt q;
     if( g.argc==3 ){
-      db_prepare(&q, 
+      db_prepare(&q,
         "SELECT tagname FROM tag"
         " WHERE EXISTS(SELECT 1 FROM tagxref"
         "               WHERE tagid=tag.tagid"
@@ -510,7 +572,7 @@ void tag_cmd(void){
       }
       db_finalize(&q);
     }else{
-      usage("tag list ?CHECK-IN?");
+      usage("list ?CHECK-IN?");
     }
   }else
   {
@@ -525,17 +587,75 @@ tag_cmd_usage:
 }
 
 /*
-** WEBPAGE: /taglist
+** COMMAND: reparent*
+**
+** Usage: %fossil reparent [OPTIONS] CHECK-IN PARENT ....
+**
+** Create a "parent" tag that causes CHECK-IN to be interpreted as a
+** child of PARENT.  If multiple PARENTs are listed, then the first is
+** the primary parent and others are merge ancestors.
+**
+** This is an experts-only command.  It is used to patch up a repository
+** that has been damaged by a shun or that has been pieced together from
+** two or more separate repositories.  You should never need to reparent
+** during normal operations.
+**
+** Reparenting is accomplished by adding a parent tag.  So to undo the
+** reparenting operation, simply delete the tag.
+**
+**    --test           Make database entries but do not add the tag artifact.
+**                     So the reparent operation will be undone by the next
+**                     "fossil rebuild" command.
+**    --dryrun | -n    Print the tag that would have been created but do not
+**                     actually change the database in any way.
+*/
+void reparent_cmd(void){
+  int bTest = find_option("test","",0)!=0;
+  int rid;
+  int i;
+  Blob value;
+  char *zUuid;
+  int dryRun = 0;
+
+  if( find_option("dryrun","n",0)!=0 ) dryRun = TAG_ADD_DRYRUN;
+  db_find_and_open_repository(0, 0);
+  verify_all_options();
+  if( g.argc<4 ){
+    usage("reparent [OPTIONS] PARENT ...");
+  }
+  rid = name_to_typed_rid(g.argv[2], "ci");
+  blob_init(&value, 0, 0);
+  for(i=3; i<g.argc; i++){
+    int pid = name_to_typed_rid(g.argv[i], "ci");
+    if( i>3 ) blob_append(&value, " ", 1);
+    zUuid = rid_to_uuid(pid);
+    blob_append(&value, zUuid, UUID_SIZE);
+    fossil_free(zUuid);
+  }
+  if( bTest && !dryRun ){
+    tag_insert("parent", 1, blob_str(&value), -1, 0.0, rid);
+  }else{
+    zUuid = rid_to_uuid(rid);
+    tag_add_artifact("","parent",zUuid,blob_str(&value),1|dryRun,0,0);
+  }
+}
+
+
+/*
+** WEBPAGE: taglist
+**
+** List all non-propagating symbolic tags.
 */
 void taglist_page(void){
   Stmt q;
 
   login_check_credentials();
   if( !g.perm.Read ){
-    login_needed();
+    login_needed(g.anon.Read);
   }
   login_anonymous_available();
   style_header("Tags");
+  style_adunit_config(ADUNIT_RIGHT_OK);
   style_submenu_element("Timeline", "Timeline", "tagtimeline");
   @ <h2>Non-propagating tags:</h2>
   db_prepare(&q,
@@ -551,7 +671,7 @@ void taglist_page(void){
   while( db_step(&q)==SQLITE_ROW ){
     const char *zName = db_column_text(&q, 0);
     if( g.perm.Hyperlink ){
-      @ <li>%z(xhref("class='taglink'","%R/timeline?t=%T",zName))
+      @ <li>%z(xhref("class='taglink'","%R/timeline?t=%T&n=200",zName))
       @ %h(zName)</a></li>
     }else{
       @ <li><span class="tagDsp">%h(zName)</span></li>
@@ -564,12 +684,15 @@ void taglist_page(void){
 
 /*
 ** WEBPAGE: /tagtimeline
+**
+** Render a timeline with all check-ins that contain non-propagating
+** symbolic tags.
 */
 void tagtimeline_page(void){
   Stmt q;
 
   login_check_credentials();
-  if( !g.perm.Read ){ login_needed(); return; }
+  if( !g.perm.Read ){ login_needed(g.anon.Read); return; }
 
   style_header("Tagged Check-ins");
   style_submenu_element("List", "List", "taglist");
@@ -583,14 +706,8 @@ void tagtimeline_page(void){
     " ORDER BY event.mtime DESC",
     timeline_query_for_www()
   );
-  www_print_timeline(&q, 0, 0, 0, 0);
+  www_print_timeline(&q, 0, 0, 0, 0, 0);
   db_finalize(&q);
   @ <br />
-  @ <script  type="text/JavaScript">
-  @ function xin(id){
-  @ }
-  @ function xout(id){
-  @ }
-  @ </script>
   style_footer();
 }
