@@ -21,6 +21,9 @@
 #include "config.h"
 #include "user.h"
 
+#if defined(_WIN32)
+#include <conio.h>
+#endif
 
 /*
 ** Strip leading and trailing space from a string and add the string
@@ -36,17 +39,17 @@ static void strip_string(Blob *pBlob, char *z){
        z[i] = 0;
        break;
     }
-    if( z[i]<' ' ) z[i] = ' ';
+    if( z[i]>0 && z[i]<' ' ) z[i] = ' ';
   }
   blob_append(pBlob, z, -1);
 }
 
-#if defined(_WIN32)
+#if defined(_WIN32) || defined(__BIONIC__)
 #ifdef __MINGW32__
 #include <conio.h>
 #endif
 /*
-** getpass for Windows
+** getpass for Windows and Android
 */
 static char *getpass(const char *prompt){
   static char pwd[64];
@@ -55,7 +58,11 @@ static char *getpass(const char *prompt){
   fputs(prompt,stderr);
   fflush(stderr);
   for(i=0; i<sizeof(pwd)-1; ++i){
+#if defined(_WIN32)
     pwd[i] = _getch();
+#else
+    pwd[i] = getc(stdin);
+#endif
     if(pwd[i]=='\r' || pwd[i]=='\n'){
       break;
     }
@@ -84,11 +91,129 @@ static char *getpass(const char *prompt){
 }
 #endif
 
+#if defined(_WIN32) || defined(WIN32)
+# include <io.h>
+# include <fcntl.h>
+# undef popen
+# define popen _popen
+# undef pclose
+# define pclose _pclose
+#endif
+
+/*
+** Scramble substitution matrix:
+*/
+static char aSubst[256];
+
+/*
+** Descramble the password
+*/
+static void userDescramble(char *z){
+  int i;
+  for(i=0; z[i]; i++) z[i] = aSubst[(unsigned char)z[i]];
+}
+
+/* Print a string in 5-letter groups */
+static void printFive(const unsigned char *z){
+  int i;
+  for(i=0; z[i]; i++){
+    if( i>0 && (i%5)==0 ) putchar(' ');
+    putchar(z[i]);
+  }
+  putchar('\n');
+}
+
+/* Return a pseudo-random integer between 0 and N-1 */
+static int randint(int N){
+  unsigned char x;
+  assert( N<256 );
+  sqlite3_randomness(1, &x);
+  return x % N;
+}
+
+/*
+** Generate and print a random scrambling of letters a through z (omitting x)
+** and set up the aSubst[] matrix to descramble.
+*/
+static void userGenerateScrambleCode(void){
+  unsigned char zOrig[30];
+  unsigned char zA[30];
+  unsigned char zB[30];
+  int nA = 25;
+  int nB = 0;
+  int i;
+  memcpy(zOrig, "abcdefghijklmnopqrstuvwyz", nA+1);
+  memcpy(zA, zOrig, nA+1);
+  assert( nA==(int)strlen((char*)zA) );
+  for(i=0; i<sizeof(aSubst); i++) aSubst[i] = i;
+  printFive(zA);
+  while( nA>0 ){
+    int x = randint(nA);
+    zB[nB++] = zA[x];
+    zA[x] = zA[--nA];
+  }
+  assert( nB==25 );
+  zB[nB] = 0;
+  printFive(zB);
+  for(i=0; i<nB; i++) aSubst[zB[i]] = zOrig[i];
+}
+
+/*
+** Return the value of the FOSSIL_SECURITY_LEVEL environment variable.
+** Or return 0 if that variable does not exist.
+*/
+int fossil_security_level(void){
+  const char *zLevel = fossil_getenv("FOSSIL_SECURITY_LEVEL");
+  if( zLevel==0 ) return 0;
+  return atoi(zLevel);
+}
+
+
 /*
 ** Do a single prompt for a passphrase.  Store the results in the blob.
+**
+**
+** The return value is a pointer to a static buffer that is overwritten
+** on subsequent calls to this same routine.
 */
 static void prompt_for_passphrase(const char *zPrompt, Blob *pPassphrase){
-  char *z = getpass(zPrompt);
+  char *z;
+#if 0
+  */
+  ** If the FOSSIL_PWREADER environment variable is set, then it will
+  ** be the name of a program that prompts the user for their password/
+  ** passphrase in a secure manner.  The program should take one or more
+  ** arguments which are the prompts and should output the acquired
+  ** passphrase as a single line on stdout.  This function will read the
+  ** output using popen().
+  **
+  ** If FOSSIL_PWREADER is not set, or if it is not the name of an
+  ** executable, then use the C-library getpass() routine.
+  */
+  const char *zProg = fossil_getenv("FOSSIL_PWREADER");
+  if( zProg && zProg[0] ){
+    static char zPass[100];
+    Blob cmd;
+    FILE *in;
+    blob_zero(&cmd);
+    blob_appendf(&cmd, "%s \"Fossil Passphrase\" \"%s\"", zProg, zPrompt);
+    zPass[0] = 0;
+    in = popen(blob_str(&cmd), "r");
+    fgets(zPass, sizeof(zPass), in);
+    pclose(in);
+    blob_reset(&cmd);
+    z = zPass;
+  }else
+#endif
+  if( fossil_security_level()>=2 ){
+    userGenerateScrambleCode();
+    z = getpass(zPrompt);
+    if( z ) userDescramble(z);
+    printf("\033[3A\033[J");  /* Erase previous three lines */
+    fflush(stdout);
+  }else{
+    z = getpass(zPrompt);
+  }
   strip_string(pPassphrase, z);
 }
 
@@ -127,6 +252,38 @@ void prompt_for_password(
   blob_reset(&secondTry);
 }
 
+/*
+** Prompt to save Fossil user password
+*/
+int save_password_prompt(const char *passwd){
+  Blob x;
+  char c;
+  const char *old = db_get("last-sync-pw", 0);
+  if( (old!=0) && fossil_strcmp(unobscure(old), passwd)==0 ){
+     return 0;
+  }
+  if( fossil_security_level()>=1 ) return 0;
+  prompt_user("remember password (Y/n)? ", &x);
+  c = blob_str(&x)[0];
+  blob_reset(&x);
+  return ( c!='n' && c!='N' );
+}
+
+/*
+** Prompt for Fossil user password
+*/
+char *prompt_for_user_password(const char *zUser){
+  char *zPrompt = mprintf("\rpassword for %s: ", zUser);
+  char *zPw;
+  Blob x;
+  fossil_force_newline();
+  prompt_for_password(zPrompt, &x, 0);
+  free(zPrompt);
+  zPw = mprintf("%b", &x);
+  blob_reset(&x);
+  return zPw;
+}
+
 /*
 ** Prompt the user to enter a single line of text.
 */
@@ -134,15 +291,17 @@ void prompt_user(const char *zPrompt, Blob *pIn){
   char *z;
   char zLine[1000];
   blob_zero(pIn);
+  fossil_force_newline();
   fossil_print("%s", zPrompt);
   fflush(stdout);
   z = fgets(zLine, sizeof(zLine), stdin);
   if( z ){
+    int n = (int)strlen(z);
+    if( n>0 && z[n-1]=='\n' ) fossil_new_line_started();
     strip_string(pIn, z);
   }
 }
 
-
 /*
 ** COMMAND: user*
 **
@@ -161,6 +320,7 @@ void prompt_user(const char *zPrompt, Blob *pIn){
 **        user for command-line interaction.
 **
 **    %fossil user list
+**    %fossil user ls
 **
 **        List all users known to the repository
 **
@@ -212,8 +372,8 @@ void user_cmd(void){
     );
     free(zPw);
   }else if( n>=2 && strncmp(g.argv[2],"default",n)==0 ){
-    user_select();
     if( g.argc==3 ){
+      user_select();
       fossil_print("%s\n", g.zLogin);
     }else{
       if( !db_exists("SELECT 1 FROM user WHERE login=%Q", g.argv[3]) ){
@@ -225,7 +385,7 @@ void user_cmd(void){
         db_set("default-user", g.argv[3], 0);
       }
     }
-  }else if( n>=2 && strncmp(g.argv[2],"list",n)==0 ){
+  }else if(( n>=2 && strncmp(g.argv[2],"list",n)==0 ) || ( n>=2 && strncmp(g.argv[2],"ls",n)==0 )){
     Stmt q;
     db_prepare(&q, "SELECT login, info FROM user ORDER BY login");
     while( db_step(&q)==SQLITE_ROW ){
@@ -258,7 +418,7 @@ void user_cmd(void){
   }else if( n>=2 && strncmp(g.argv[2],"capabilities",2)==0 ){
     int uid;
     if( g.argc!=4 && g.argc!=5 ){
-      usage("user capabilities USERNAME ?PERMISSIONS?");
+      usage("capabilities USERNAME ?PERMISSIONS?");
     }
     uid = db_int(0, "SELECT uid FROM user WHERE login=%Q", g.argv[3]);
     if( uid==0 ){
@@ -272,7 +432,7 @@ void user_cmd(void){
     }
     fossil_print("%s\n", db_text(0, "SELECT cap FROM user WHERE uid=%d", uid));
   }else{
-    fossil_panic("user subcommand should be one of: "
+    fossil_fatal("user subcommand should be one of: "
                  "capabilities default list new password");
   }
 }
@@ -304,17 +464,19 @@ static int attempt_user(const char *zLogin){
 **
 **   (3)  Check the default user in the repository
 **
-**   (4)  Try the USER environment variable.
+**   (4)  Try the FOSSIL_USER environment variable.
 **
-**   (5)  Try the USERNAME environment variable.
+**   (5)  Try the USER environment variable.
 **
-**   (6)  Check if the user can be extracted from the remote URL.
+**   (6)  Try the LOGNAME environment variable.
+**
+**   (7)  Try the USERNAME environment variable.
+**
+**   (8)  Check if the user can be extracted from the remote URL.
 **
 ** The user name is stored in g.zLogin.  The uid is in g.userUid.
 */
 void user_select(void){
-  char *zUrl;
-
   if( g.userUid ) return;
   if( g.zLogin ){
     if( attempt_user(g.zLogin)==0 ){
@@ -328,15 +490,16 @@ void user_select(void){
 
   if( attempt_user(db_get("default-user", 0)) ) return;
 
+  if( attempt_user(fossil_getenv("FOSSIL_USER")) ) return;
+
   if( attempt_user(fossil_getenv("USER")) ) return;
 
+  if( attempt_user(fossil_getenv("LOGNAME")) ) return;
+
   if( attempt_user(fossil_getenv("USERNAME")) ) return;
 
-  zUrl = db_get("last-sync-url", 0);
-  if( zUrl ){
-    url_parse(zUrl);
-    if( attempt_user(g.urlUser) ) return;
-  }
+  url_parse(0, 0);
+  if( g.url.user && attempt_user(g.url.user) ) return;
 
   fossil_print(
     "Cannot figure out who you are!  Consider using the --user\n"
@@ -346,6 +509,32 @@ void user_select(void){
   fossil_fatal("cannot determine user");
 }
 
+/*
+** COMMAND: test-usernames
+**
+** Usage: %fossil test-usernames
+**
+** Print details about sources of fossil usernames.
+*/
+void test_usernames_cmd(void){
+  db_find_and_open_repository(0, 0);
+
+  fossil_print("Initial g.zLogin: %s\n", g.zLogin);
+  fossil_print("Initial g.userUid: %d\n", g.userUid);
+  fossil_print("checkout default-user: %s\n", g.localOpen ?
+               db_lget("default-user","") : "<<no open checkout>>");
+  fossil_print("default-user: %s\n", db_get("default-user",""));
+  fossil_print("FOSSIL_USER: %s\n", fossil_getenv("FOSSIL_USER"));
+  fossil_print("USER: %s\n", fossil_getenv("USER"));
+  fossil_print("LOGNAME: %s\n", fossil_getenv("LOGNAME"));
+  fossil_print("USERNAME: %s\n", fossil_getenv("USERNAME"));
+  url_parse(0, 0);
+  fossil_print("URL user: %s\n", g.url.user);
+  user_select();
+  fossil_print("Final g.zLogin: %s\n", g.zLogin);
+  fossil_print("Final g.userUid: %d\n", g.userUid);
+}
+
 
 /*
 ** COMMAND: test-hash-passwords
@@ -367,26 +556,48 @@ void user_hash_passwords_cmd(void){
   );
 }
 
+/*
+** COMMAND: test-prompt-user
+**
+** Usage: %fossil test-prompt-user PROMPT
+**
+** Prompts the user for input and then prints it verbatim (i.e. without
+** a trailing line terminator).
+*/
+void test_prompt_user_cmd(void){
+  Blob answer;
+  if( g.argc!=3 ) usage("PROMPT");
+  prompt_user(g.argv[2], &answer);
+  fossil_print("%s", blob_str(&answer));
+}
+
 /*
 ** WEBPAGE: access_log
 **
-**    y=N      1: success only.  2: failure only.  3: both
-**    n=N      Number of entries to show
-**    o=N      Skip this many entries
+** Show login attempts, including timestamp and IP address.
+** Requires Admin privileges.
+**
+** Query parameters:
+**
+**    y=N      1: success only.  2: failure only.  3: both (default: 3)
+**    n=N      Number of entries to show (default: 200)
+**    o=N      Skip this many entries (default: 0)
 */
 void access_log_page(void){
   int y = atoi(PD("y","3"));
-  int n = atoi(PD("n","50"));
+  int n = atoi(PD("n","200"));
   int skip = atoi(PD("o","0"));
   Blob sql;
   Stmt q;
   int cnt = 0;
   int rc;
+  int fLogEnabled;
 
   login_check_credentials();
-  if( !g.perm.Admin ){ login_needed(); return; }
+  if( !g.perm.Admin ){ login_needed(0); return; }
   create_accesslog_table();
 
+
   if( P("delall") && P("delallbtn") ){
     db_multi_exec("DELETE FROM accesslog");
     cgi_redirectf("%s/access_log?y=%d&n=%d&o=%o", g.zTop, y, n, skip);
@@ -411,25 +622,28 @@ void access_log_page(void){
   }
   style_header("Access Log");
   blob_zero(&sql);
-  blob_append(&sql, 
-    "SELECT uname, ipaddr, datetime(mtime, 'localtime'), success"
-    "  FROM accesslog", -1
+  blob_append_sql(&sql,
+    "SELECT uname, ipaddr, datetime(mtime,toLocal()), success"
+    "  FROM accesslog"
   );
   if( y==1 ){
     blob_append(&sql, "  WHERE success", -1);
   }else if( y==2 ){
     blob_append(&sql, "  WHERE NOT success", -1);
   }
-  blob_appendf(&sql,"  ORDER BY rowid DESC LIMIT %d OFFSET %d", n+1, skip);
+  blob_append_sql(&sql,"  ORDER BY rowid DESC LIMIT %d OFFSET %d", n+1, skip);
   if( skip ){
     style_submenu_element("Newer", "Newer entries",
               "%s/access_log?o=%d&n=%d&y=%d", g.zTop, skip>=n ? skip-n : 0,
               n, y);
   }
-  rc = db_prepare_ignore_error(&q, blob_str(&sql));
-  @ <center><table border="1" cellpadding="5">
-  @ <tr><th width="33%%">Date</th><th width="34%%">User</th>
-  @ <th width="33%%">IP Address</th></tr>
+  rc = db_prepare_ignore_error(&q, "%s", blob_sql_text(&sql));
+  fLogEnabled = db_get_boolean("access-log", 0);
+  @ <div align="center">Access logging is %s(fLogEnabled?"on":"off").
+  @ (Change this on the <a href="setup_settings">settings</a> page.)</div>
+  @ <table border="1" cellpadding="5" id="logtable" align="center">
+  @ <thead><tr><th width="33%%">Date</th><th width="34%%">User</th>
+  @ <th width="33%%">IP Address</th></tr></thead><tbody>
   while( rc==SQLITE_OK && db_step(&q)==SQLITE_ROW ){
     const char *zName = db_column_text(&q, 0);
     const char *zIP = db_column_text(&q, 1);
@@ -452,9 +666,9 @@ void access_log_page(void){
     style_submenu_element("All", "All entries",
           "%s/access_log?n=10000000", g.zTop);
   }
-  @ </table></center>
+  @ </tbody></table>
   db_finalize(&q);
-  @ <hr>
+  @ <hr />
   @ <form method="post" action="%s(g.zTop)/access_log">
   @ <label><input type="checkbox" name="delold">
   @ Delete all but the most recent 200 entries</input></label>
@@ -475,5 +689,6 @@ void access_log_page(void){
   @ Delete all entries</input></label>
   @ <input type="submit" name="delallbtn" value="Delete"></input>
   @ </form>
+  output_table_sorting_javascript("logtable", "Ttt", 1);
   style_footer();
 }
