@@ -47,14 +47,14 @@ int fast_uuid_to_rid(const char *zUuid){
 ** For this routine, the UUID must be exact.  For a match against
 ** user input with mixed case, use resolve_uuid().
 **
-** If the UUID is not found and phantomize is 1 or 2, then attempt to 
+** If the UUID is not found and phantomize is 1 or 2, then attempt to
 ** create a phantom record.  A private phantom is created for 2 and
 ** a public phantom is created for 1.
 */
 int uuid_to_rid(const char *zUuid, int phantomize){
   int rid, sz;
   char z[UUID_SIZE+1];
-  
+
   sz = strlen(zUuid);
   if( sz!=UUID_SIZE || !validate16(zUuid, sz) ){
     return 0;
@@ -70,34 +70,38 @@ int uuid_to_rid(const char *zUuid, int phantomize){
 
 
 /*
-** Load a vfile from a record ID.
+** Load a vfile from a record ID.  Return the number of files with
+** missing content.
 */
-void load_vfile_from_rid(int vid){
-  int rid, size;
+int load_vfile_from_rid(int vid){
+  int rid, size, nMissing;
   Stmt ins, ridq;
   Manifest *p;
   ManifestFile *pFile;
 
   if( db_exists("SELECT 1 FROM vfile WHERE vid=%d", vid) ){
-    return;
+    return 0;
   }
 
   db_begin_transaction();
-  p = manifest_get(vid, CFTYPE_MANIFEST);
-  if( p==0 ) return;
-  db_multi_exec("DELETE FROM vfile WHERE vid=%d", vid);
+  p = manifest_get(vid, CFTYPE_MANIFEST, 0);
+  if( p==0 ) {
+    db_end_transaction(1);
+    return 0;
+  }
   db_prepare(&ins,
     "INSERT INTO vfile(vid,isexe,islink,rid,mrid,pathname) "
     " VALUES(:vid,:isexe,:islink,:id,:id,:name)");
   db_prepare(&ridq, "SELECT rid,size FROM blob WHERE uuid=:uuid");
   db_bind_int(&ins, ":vid", vid);
   manifest_file_rewind(p);
+  nMissing = 0;
   while( (pFile = manifest_file_next(p,0))!=0 ){
     if( pFile->zUuid==0 || uuid_is_shunned(pFile->zUuid) ) continue;
     db_bind_text(&ridq, ":uuid", pFile->zUuid);
     if( db_step(&ridq)==SQLITE_ROW ){
       rid = db_column_int(&ridq, 0);
-      size = db_column_int(&ridq, 0);
+      size = db_column_int(&ridq, 1);
     }else{
       rid = 0;
       size = 0;
@@ -105,6 +109,7 @@ void load_vfile_from_rid(int vid){
     db_reset(&ridq);
     if( rid==0 || size<0 ){
       fossil_warning("content missing for %s", pFile->zName);
+      nMissing++;
       continue;
     }
     db_bind_int(&ins, ":isexe", ( manifest_file_mperm(pFile)==PERM_EXE ));
@@ -118,6 +123,7 @@ void load_vfile_from_rid(int vid){
   db_finalize(&ins);
   manifest_destroy(p);
   db_end_transaction(0);
+  return nMissing;
 }
 
 #if INTERFACE
@@ -132,15 +138,22 @@ void load_vfile_from_rid(int vid){
 #endif /* INTERFACE */
 
 /*
-** Look at every VFILE entry with the given vid and update
-** VFILE.CHNGED field according to whether or not
-** the file has changed.  0 means no change.  1 means edited.  2 means
-** the file has changed due to a merge.  3 means the file was added
-** by a merge.
+** Look at every VFILE entry with the given vid and update VFILE.CHNGED field
+** according to whether or not the file has changed.
+** - 0 means no change.
+** - 1 means edited.
+** - 2 means changed due to a merge.
+** - 3 means added by a merge.
+** - 4 means changed due to an integrate merge.
+** - 5 means added by an integrate merge.
+** - 6 means became executable but has unmodified contents.
+** - 7 means became a symlink whose target equals its old contents.
+** - 8 means lost executable status but has unmodified contents.
+** - 9 means lost symlink status and has contents equal to its old target.
 **
 ** If VFILE.DELETED is true or if VFILE.RID is zero, then the file was either
 ** removed from configuration management via "fossil rm" or added via
-** "fossil add", respectively, and in both cases we always know that 
+** "fossil add", respectively, and in both cases we always know that
 ** the file has changed without having the check the size, mtime,
 ** or on-disk content.
 **
@@ -166,14 +179,20 @@ void vfile_check_signature(int vid, unsigned int cksigFlags){
 
   db_begin_transaction();
   db_prepare(&q, "SELECT id, %Q || pathname,"
-                 "       vfile.mrid, deleted, chnged, uuid, size, mtime"
+                 "       vfile.mrid, deleted, chnged, uuid, size, mtime,"
+                 "      CASE WHEN isexe THEN %d WHEN islink THEN %d ELSE %d END"
                  "  FROM vfile LEFT JOIN blob ON vfile.mrid=blob.rid"
-                 " WHERE vid=%d ", g.zLocalRoot, vid);
+                 " WHERE vid=%d ", g.zLocalRoot, PERM_EXE, PERM_LNK, PERM_REG,
+                 vid);
   while( db_step(&q)==SQLITE_ROW ){
     int id, rid, isDeleted;
     const char *zName;
     int chnged = 0;
     int oldChnged;
+#ifndef _WIN32
+    int origPerm;
+    int currentPerm;
+#endif
     i64 oldMtime;
     i64 currentMtime;
     i64 origSize;
@@ -185,9 +204,13 @@ void vfile_check_signature(int vid, unsigned int cksigFlags){
     isDeleted = db_column_int(&q, 3);
     oldChnged = chnged = db_column_int(&q, 4);
     oldMtime = db_column_int64(&q, 7);
-    currentSize = file_wd_size(zName);
     origSize = db_column_int64(&q, 6);
+    currentSize = file_wd_size(zName);
     currentMtime = file_wd_mtime(0);
+#ifndef _WIN32
+    origPerm = db_column_int(&q, 8);
+    currentPerm = file_wd_perm(zName);
+#endif
     if( chnged==0 && (isDeleted || rid==0) ){
       /* "fossil rm" or "fossil add" always change the file */
       chnged = 1;
@@ -215,7 +238,7 @@ void vfile_check_signature(int vid, unsigned int cksigFlags){
       if( blob_compare(&fileCksum, &origCksum)==0 ) chnged = 0;
       blob_reset(&origCksum);
       blob_reset(&fileCksum);
-    }else if( (chnged==0 || chnged==2)
+    }else if( (chnged==0 || chnged==2 || chnged==4)
            && (useMtime==0 || currentMtime!=oldMtime) ){
       /* For files that were formerly believed to be unchanged or that were
       ** changed by merging, if their mtime changes, or unconditionally
@@ -232,7 +255,7 @@ void vfile_check_signature(int vid, unsigned int cksigFlags){
       blob_reset(&origCksum);
       blob_reset(&fileCksum);
     }
-    if( (cksigFlags & CKSIG_SETMTIME) && (chnged==0 || chnged==2) ){
+    if( (cksigFlags & CKSIG_SETMTIME) && (chnged==0 || chnged==2 || chnged==4) ){
       i64 desiredMtime;
       if( mtime_of_manifest_file(vid,rid,&desiredMtime)==0 ){
         if( currentMtime!=desiredMtime ){
@@ -241,6 +264,21 @@ void vfile_check_signature(int vid, unsigned int cksigFlags){
         }
       }
     }
+#ifndef _WIN32
+    if( chnged==0 || chnged==6 || chnged==7 || chnged==8 || chnged==9 ){
+      if( origPerm==currentPerm ){
+        chnged = 0;
+      }else if( currentPerm==PERM_EXE ){
+        chnged = 6;
+      }else if( currentPerm==PERM_LNK ){
+        chnged = 7;
+      }else if( origPerm==PERM_EXE ){
+        chnged = 8;
+      }else if( origPerm==PERM_LNK ){
+        chnged = 9;
+      }
+    }
+#endif
     if( currentMtime!=oldMtime || chnged!=oldChnged ){
       db_multi_exec("UPDATE vfile SET mtime=%lld, chnged=%d WHERE id=%d",
                     currentMtime, chnged, id);
@@ -306,19 +344,17 @@ void vfile_to_disk(
       blob_reset(&ans);
       if( cReply=='a' || cReply=='A' ){
         promptFlag = 0;
-        cReply = 'y';
-      }
-      if( cReply=='n' || cReply=='N' ){
+      } else if( cReply!='y' && cReply!='Y' ){
         blob_reset(&content);
         continue;
       }
     }
     if( verbose ) fossil_print("%s\n", &zName[nRepos]);
-    if( file_wd_isdir(zName) == 1 ){
+    if( file_wd_isdir(zName)==1 ){
       /*TODO(dchest): remove directories? */
-      fossil_fatal("%s is directory, cannot overwrite\n", zName);
-    }    
-    if( file_wd_size(zName)>=0 && (isLink || file_wd_islink(zName)) ){
+      fossil_fatal("%s is directory, cannot overwrite", zName);
+    }
+    if( file_wd_size(zName)>=0 && (isLink || file_wd_islink(0)) ){
       file_delete(zName);
     }
     if( isLink ){
@@ -394,10 +430,10 @@ static int is_temporary_file(const char *zName){
      "output",
   };
   int i, j, n;
-  
-  if( strglob("ci-comment-????????????.txt", zName) ) return 1;
+
+  if( sqlite3_strglob("ci-comment-????????????.txt", zName)==0 ) return 1;
   for(; zName[0]!=0; zName++){
-    if( zName[0]=='/' && strglob("/ci-comment-????????????.txt", zName) ){
+    if( zName[0]=='/' && sqlite3_strglob("/ci-comment-????????????.txt", zName)==0 ){
       return 1;
     }
     if( zName[0]!='-' ) continue;
@@ -409,7 +445,7 @@ static int is_temporary_file(const char *zName){
         for(j=n+2; zName[j] && fossil_isdigit(zName[j]); j++){}
         if( zName[j]==0 ) return 1;
       }
-    }      
+    }
   }
   return 0;
 }
@@ -420,6 +456,7 @@ static int is_temporary_file(const char *zName){
 */
 #define SCAN_ALL    0x001    /* Includes files that begin with "." */
 #define SCAN_TEMP   0x002    /* Only Fossil-generated files like *-baseline */
+#define SCAN_NESTED 0x004    /* Scan for empty dirs in nested checkouts */
 #endif /* INTERFACE */
 
 /*
@@ -430,26 +467,33 @@ static int is_temporary_file(const char *zName){
 ** Subdirectories are scanned recursively.
 ** Omit files named in VFILE.
 **
-** Files whose names begin with "." are omitted unless allFlag is true.
+** Files whose names begin with "." are omitted unless the SCAN_ALL
+** flag is set.
 **
-** Any files or directories that match the glob pattern pIgnore are 
-** excluded from the scan.  Name matching occurs after the first
-** nPrefix characters are elided from the filename.
+** Any files or directories that match the glob patterns pIgnore*
+** are excluded from the scan.  Name matching occurs after the
+** first nPrefix characters are elided from the filename.
 */
-void vfile_scan(Blob *pPath, int nPrefix, unsigned scanFlags, Glob *pIgnore){
+void vfile_scan(
+  Blob *pPath,           /* Directory to be scanned */
+  int nPrefix,           /* Number of bytes in directory name */
+  unsigned scanFlags,    /* Zero or more SCAN_xxx flags */
+  Glob *pIgnore1,        /* Do not add files that match this GLOB */
+  Glob *pIgnore2         /* Omit files matching this GLOB too */
+){
   DIR *d;
   int origSize;
-  const char *zDir;
   struct dirent *pEntry;
   int skipAll = 0;
   static Stmt ins;
   static int depth = 0;
-  void *zMbcs;
+  void *zNative;
 
   origSize = blob_size(pPath);
-  if( pIgnore ){
+  if( pIgnore1 || pIgnore2 ){
     blob_appendf(pPath, "/");
-    if( glob_match(pIgnore, &blob_str(pPath)[nPrefix+1]) ) skipAll = 1;
+    if( glob_match(pIgnore1, &blob_str(pPath)[nPrefix+1]) ) skipAll = 1;
+    if( glob_match(pIgnore2, &blob_str(pPath)[nPrefix+1]) ) skipAll = 1;
     blob_resize(pPath, origSize);
   }
   if( skipAll ) return;
@@ -457,14 +501,14 @@ void vfile_scan(Blob *pPath, int nPrefix, unsigned scanFlags, Glob *pIgnore){
   if( depth==0 ){
     db_prepare(&ins,
        "INSERT OR IGNORE INTO sfile(x) SELECT :file"
-       "  WHERE NOT EXISTS(SELECT 1 FROM vfile WHERE pathname=:file)"
+       "  WHERE NOT EXISTS(SELECT 1 FROM vfile WHERE"
+       " pathname=:file %s)", filename_collation()
     );
   }
   depth++;
 
-  zDir = blob_str(pPath);
-  zMbcs = fossil_utf8_to_unicode(zDir);
-  d = opendir(zMbcs);
+  zNative = fossil_utf8_to_path(blob_str(pPath), 1);
+  d = opendir(zNative);
   if( d ){
     while( (pEntry=readdir(d))!=0 ){
       char *zPath;
@@ -474,28 +518,39 @@ void vfile_scan(Blob *pPath, int nPrefix, unsigned scanFlags, Glob *pIgnore){
         if( pEntry->d_name[1]==0 ) continue;
         if( pEntry->d_name[1]=='.' && pEntry->d_name[2]==0 ) continue;
       }
-      zUtf8 = fossil_unicode_to_utf8(pEntry->d_name);
+      zUtf8 = fossil_path_to_utf8(pEntry->d_name);
       blob_appendf(pPath, "/%s", zUtf8);
       zPath = blob_str(pPath);
-      if( glob_match(pIgnore, &zPath[nPrefix+1]) ){
+      if( glob_match(pIgnore1, &zPath[nPrefix+1]) ||
+          glob_match(pIgnore2, &zPath[nPrefix+1]) ){
         /* do nothing */
+#ifdef _DIRENT_HAVE_D_TYPE
+      }else if( (pEntry->d_type==DT_UNKNOWN || pEntry->d_type==DT_LNK)
+          ? (file_wd_isdir(zPath)==1) : (pEntry->d_type==DT_DIR) ){
+#else
       }else if( file_wd_isdir(zPath)==1 ){
+#endif
         if( !vfile_top_of_checkout(zPath) ){
-          vfile_scan(pPath, nPrefix, scanFlags, pIgnore);
+          vfile_scan(pPath, nPrefix, scanFlags, pIgnore1, pIgnore2);
         }
+#ifdef _DIRENT_HAVE_D_TYPE
+      }else if( (pEntry->d_type==DT_UNKNOWN || pEntry->d_type==DT_LNK)
+          ? (file_wd_isfile_or_link(zPath)) : (pEntry->d_type==DT_REG) ){
+#else
       }else if( file_wd_isfile_or_link(zPath) ){
+#endif
         if( (scanFlags & SCAN_TEMP)==0 || is_temporary_file(zUtf8) ){
           db_bind_text(&ins, ":file", &zPath[nPrefix+1]);
           db_step(&ins);
           db_reset(&ins);
         }
       }
-      fossil_mbcs_free(zUtf8);
+      fossil_path_free(zUtf8);
       blob_resize(pPath, origSize);
     }
     closedir(d);
   }
-  fossil_mbcs_free(zMbcs);
+  fossil_path_free(zNative);
 
   depth--;
   if( depth==0 ){
@@ -503,6 +558,130 @@ void vfile_scan(Blob *pPath, int nPrefix, unsigned scanFlags, Glob *pIgnore){
   }
 }
 
+/*
+** Scans the specified base directory for any directories within it, while
+** keeping a count of how many files they each contains, either directly or
+** indirectly.
+**
+** Subdirectories are scanned recursively.
+** Omit files named in VFILE.
+**
+** Directories whose names begin with "." are omitted unless the SCAN_ALL
+** flag is set.
+**
+** Any directories that match the glob patterns pIgnore* are excluded from
+** the scan.  Name matching occurs after the first nPrefix characters are
+** elided from the filename.
+**
+** Returns the total number of files found.
+*/
+int vfile_dir_scan(
+  Blob *pPath,           /* Base directory to be scanned */
+  int nPrefix,           /* Number of bytes in base directory name */
+  unsigned scanFlags,    /* Zero or more SCAN_xxx flags */
+  Glob *pIgnore1,        /* Do not add directories that match this GLOB */
+  Glob *pIgnore2         /* Omit directories matching this GLOB too */
+){
+  int result = 0;
+  DIR *d;
+  int origSize;
+  struct dirent *pEntry;
+  int skipAll = 0;
+  static Stmt ins;
+  static Stmt upd;
+  static int depth = 0;
+  void *zNative;
+
+  origSize = blob_size(pPath);
+  if( pIgnore1 || pIgnore2 ){
+    blob_appendf(pPath, "/");
+    if( glob_match(pIgnore1, &blob_str(pPath)[nPrefix+1]) ) skipAll = 1;
+    if( glob_match(pIgnore2, &blob_str(pPath)[nPrefix+1]) ) skipAll = 1;
+    blob_resize(pPath, origSize);
+  }
+  if( skipAll ) return result;
+
+  if( depth==0 ){
+    db_multi_exec("DROP TABLE IF EXISTS dscan_temp;"
+                  "CREATE TEMP TABLE dscan_temp("
+                  "  x TEXT PRIMARY KEY %s, y INTEGER)",
+                  filename_collation());
+    db_prepare(&ins,
+       "INSERT OR IGNORE INTO dscan_temp(x, y) SELECT :file, :count"
+       "  WHERE NOT EXISTS(SELECT 1 FROM vfile WHERE"
+       " pathname GLOB :file || '/*' %s)", filename_collation()
+    );
+    db_prepare(&upd,
+       "UPDATE OR IGNORE dscan_temp SET y = coalesce(y, 0) + 1"
+       "  WHERE x=:file %s",
+       filename_collation()
+    );
+  }
+  depth++;
+
+  zNative = fossil_utf8_to_path(blob_str(pPath), 1);
+  d = opendir(zNative);
+  if( d ){
+    while( (pEntry=readdir(d))!=0 ){
+      char *zOrigPath;
+      char *zPath;
+      char *zUtf8;
+      if( pEntry->d_name[0]=='.' ){
+        if( (scanFlags & SCAN_ALL)==0 ) continue;
+        if( pEntry->d_name[1]==0 ) continue;
+        if( pEntry->d_name[1]=='.' && pEntry->d_name[2]==0 ) continue;
+      }
+      zOrigPath = mprintf("%s", blob_str(pPath));
+      zUtf8 = fossil_path_to_utf8(pEntry->d_name);
+      blob_appendf(pPath, "/%s", zUtf8);
+      zPath = blob_str(pPath);
+      if( glob_match(pIgnore1, &zPath[nPrefix+1]) ||
+          glob_match(pIgnore2, &zPath[nPrefix+1]) ){
+        /* do nothing */
+#ifdef _DIRENT_HAVE_D_TYPE
+      }else if( (pEntry->d_type==DT_UNKNOWN || pEntry->d_type==DT_LNK)
+          ? (file_wd_isdir(zPath)==1) : (pEntry->d_type==DT_DIR) ){
+#else
+      }else if( file_wd_isdir(zPath)==1 ){
+#endif
+        if( (scanFlags & SCAN_NESTED) || !vfile_top_of_checkout(zPath) ){
+          char *zSavePath = mprintf("%s", zPath);
+          int count = vfile_dir_scan(pPath, nPrefix, scanFlags, pIgnore1,
+                                     pIgnore2);
+          db_bind_text(&ins, ":file", &zSavePath[nPrefix+1]);
+          db_bind_int(&ins, ":count", count);
+          db_step(&ins);
+          db_reset(&ins);
+          fossil_free(zSavePath);
+          result += count; /* found X normal files? */
+        }
+#ifdef _DIRENT_HAVE_D_TYPE
+      }else if( (pEntry->d_type==DT_UNKNOWN || pEntry->d_type==DT_LNK)
+          ? (file_wd_isfile_or_link(zPath)) : (pEntry->d_type==DT_REG) ){
+#else
+      }else if( file_wd_isfile_or_link(zPath) ){
+#endif
+        db_bind_text(&upd, ":file", zOrigPath);
+        db_step(&upd);
+        db_reset(&upd);
+        result++; /* found 1 normal file */
+      }
+      fossil_path_free(zUtf8);
+      blob_resize(pPath, origSize);
+      fossil_free(zOrigPath);
+    }
+    closedir(d);
+  }
+  fossil_path_free(zNative);
+
+  depth--;
+  if( depth==0 ){
+    db_finalize(&upd);
+    db_finalize(&ins);
+  }
+  return result;
+}
+
 /*
 ** Compute an aggregate MD5 checksum over the disk image of every
 ** file in vid.  The file names are part of the checksum.  The resulting
@@ -530,7 +709,7 @@ void vfile_aggregate_checksum_disk(int vid, Blob *pOut){
   char zBuf[4096];
 
   db_must_be_within_tree();
-  db_prepare(&q, 
+  db_prepare(&q,
       "SELECT %Q || pathname, pathname, origname, is_selected(id), rid"
       "  FROM vfile"
       " WHERE (NOT deleted OR NOT is_selected(id)) AND vid=%d"
@@ -549,7 +728,7 @@ void vfile_aggregate_checksum_disk(int vid, Blob *pOut){
         /* Instead of file content, use link destination path */
         Blob pathBuf;
 
-        sqlite3_snprintf(sizeof(zBuf), zBuf, " %ld\n", 
+        sqlite3_snprintf(sizeof(zBuf), zBuf, " %ld\n",
                          blob_read_link(&pathBuf, zFullpath));
         md5sum_step_text(zBuf, -1);
         md5sum_step_text(blob_str(&pathBuf), -1);
@@ -619,9 +798,9 @@ void vfile_compare_repository_to_disk(int vid){
   Stmt q;
   Blob disk, repo;
   char *zOut;
-  
+
   db_must_be_within_tree();
-  db_prepare(&q, 
+  db_prepare(&q,
       "SELECT %Q || pathname, pathname, rid FROM vfile"
       " WHERE NOT deleted AND vid=%d AND is_selected(id)"
       " ORDER BY if_selected(id, pathname, origname) /*scan*/",
@@ -685,7 +864,7 @@ void vfile_aggregate_checksum_repository(int vid, Blob *pOut){
   char zBuf[100];
 
   db_must_be_within_tree();
- 
+
   db_prepare(&q, "SELECT pathname, origname, rid, is_selected(id)"
                  " FROM vfile"
                  " WHERE (NOT deleted OR NOT is_selected(id))"
@@ -723,23 +902,26 @@ void vfile_aggregate_checksum_repository(int vid, Blob *pOut){
 ** "R" card near the end of the manifest.
 **
 ** In a well-formed manifest, the two checksums computed here, pOut and
-** pManOut, should be identical.  
+** pManOut, should be identical.
 */
 void vfile_aggregate_checksum_manifest(int vid, Blob *pOut, Blob *pManOut){
   int fid;
   Blob file;
+  Blob err;
   Manifest *pManifest;
   ManifestFile *pFile;
   char zBuf[100];
 
   blob_zero(pOut);
+  blob_zero(&err);
   if( pManOut ){
     blob_zero(pManOut);
   }
   db_must_be_within_tree();
-  pManifest = manifest_get(vid, CFTYPE_MANIFEST);
+  pManifest = manifest_get(vid, CFTYPE_MANIFEST, &err);
   if( pManifest==0 ){
-    fossil_panic("manifest file (%d) is malformed", vid);
+    fossil_fatal("manifest file (%d) is malformed:\n%s",
+                 vid, blob_str(&err));
   }
   manifest_file_rewind(pManifest);
   while( (pFile = manifest_file_next(pManifest,0))!=0 ){
@@ -765,6 +947,11 @@ void vfile_aggregate_checksum_manifest(int vid, Blob *pOut, Blob *pManOut){
 
 /*
 ** COMMAND: test-agg-cksum
+**
+** Display the aggregate checksum for content computed in several
+** different ways.  The aggregate checksum is used during "fossil commit"
+** to double-check that the information about to be committed to the
+** repository exactly matches the information currently in the check-out.
 */
 void test_agg_cksum_cmd(void){
   int vid;
