@@ -23,8 +23,23 @@
 #ifdef _WIN32
 /* This code is for win32 only */
 #include <windows.h>
+#include <process.h>
 #include "winhttp.h"
 
+/*
+** The HttpServer structure holds information about an instance of
+** the HTTP server itself.
+*/
+typedef struct HttpServer HttpServer;
+struct HttpServer {
+  HANDLE hStoppedEvent; /* Event to signal when server is stopped,
+                        ** must be closed by callee. */
+  char *zStopper;       /* The stopper file name, must be freed by
+                        ** callee. */
+  SOCKET listener;      /* Socket on which the server is listening,
+                        ** may be closed by callee. */
+};
+
 /*
 ** The HttpRequest structure holds information about each incoming
 ** HTTP request.
@@ -34,7 +49,8 @@ struct HttpRequest {
   int id;                /* ID counter */
   SOCKET s;              /* Socket on which to receive data */
   SOCKADDR_IN addr;      /* Address from which data is coming */
-  const char *zOptions;  /* --notfound and/or --localauth options */
+  int flags;             /* Flags passed to win32_http_server() */
+  const char *zOptions;  /* --baseurl, --notfound, --localauth, --th-trace */
 };
 
 /*
@@ -60,20 +76,75 @@ static int find_content_length(const char *zHdr){
   return 0;
 }
 
+/*
+** Issue a fatal error.
+*/
+static NORETURN void winhttp_fatal(
+  const char *zOp,
+  const char *zService,
+  const char *zErr
+){
+  fossil_fatal("unable to %s service '%s': %s", zOp, zService, zErr);
+}
+
+/*
+** Make sure the server stops as soon as possible after the stopper file
+** is found.  If there is no stopper file name, do nothing.
+*/
+static void win32_server_stopper(void *pAppData){
+  HttpServer *p = (HttpServer*)pAppData;
+  if( p!=0 ){
+    HANDLE hStoppedEvent = p->hStoppedEvent;
+    const char *zStopper = p->zStopper;
+    SOCKET listener = p->listener;
+    if( hStoppedEvent!=NULL && zStopper!=0 && listener!=INVALID_SOCKET ){
+      while( 1 ){
+        DWORD dwResult = WaitForMultipleObjectsEx(1, &hStoppedEvent, FALSE,
+                                                  1000, TRUE);
+        if( dwResult!=WAIT_IO_COMPLETION && dwResult!=WAIT_TIMEOUT ){
+          /* The event is either invalid, signaled, or abandoned.  Bail
+          ** out now because those conditions should indicate the parent
+          ** thread is dead or dying. */
+          break;
+        }
+        if( file_size(zStopper)>=0 ){
+          /* The stopper file has been found.  Attempt to close the server
+          ** listener socket now and then exit. */
+          closesocket(listener);
+          p->listener = INVALID_SOCKET;
+          break;
+        }
+      }
+    }
+    if( hStoppedEvent!=NULL ){
+      CloseHandle(hStoppedEvent);
+      p->hStoppedEvent = NULL;
+    }
+    if( zStopper!=0 ){
+      fossil_free(p->zStopper);
+      p->zStopper = 0;
+    }
+    fossil_free(p);
+  }
+}
+
 /*
 ** Process a single incoming HTTP request.
 */
-static void win32_process_one_http_request(void *pAppData){
+static void win32_http_request(void *pAppData){
   HttpRequest *p = (HttpRequest*)pAppData;
   FILE *in = 0, *out = 0;
   int amt, got;
   int wanted = 0;
   char *z;
+  char zCmdFName[MAX_PATH];
   char zRequestFName[MAX_PATH];
   char zReplyFName[MAX_PATH];
   char zCmd[2000];          /* Command-line to process the request */
   char zHdr[2000];          /* The HTTP request header */
 
+  sqlite3_snprintf(MAX_PATH, zCmdFName,
+                   "%s_cmd%d.txt", zTempPrefix, p->id);
   sqlite3_snprintf(MAX_PATH, zRequestFName,
                    "%s_in%d.txt", zTempPrefix, p->id);
   sqlite3_snprintf(MAX_PATH, zReplyFName,
@@ -110,9 +181,29 @@ static void win32_process_one_http_request(void *pAppData){
   }
   fclose(out);
   out = 0;
-  sqlite3_snprintf(sizeof(zCmd), zCmd, "\"%s\" http \"%s\" %s %s %s --nossl%s",
-    g.nameOfExe, g.zRepositoryName, zRequestFName, zReplyFName,
-    inet_ntoa(p->addr.sin_addr), p->zOptions
+  /*
+  ** The repository name is only needed if there was no open checkout.  This
+  ** is designed to allow the open checkout for the interactive user to work
+  ** with the local Fossil server started via the "ui" command.
+  */
+  if( (p->flags & HTTP_SERVER_HAD_CHECKOUT)==0 ){
+    assert( g.zRepositoryName && g.zRepositoryName[0] );
+    sqlite3_snprintf(sizeof(zCmd), zCmd, "%s%s\n%s\n%s\n%s",
+      get_utf8_bom(0), zRequestFName, zReplyFName, inet_ntoa(p->addr.sin_addr),
+      g.zRepositoryName
+    );
+  }else{
+    sqlite3_snprintf(sizeof(zCmd), zCmd, "%s%s\n%s\n%s",
+      get_utf8_bom(0), zRequestFName, zReplyFName, inet_ntoa(p->addr.sin_addr)
+    );
+  }
+  out = fossil_fopen(zCmdFName, "wb");
+  if( out==0 ) goto end_request;
+  fwrite(zCmd, 1, strlen(zCmd), out);
+  fclose(out);
+
+  sqlite3_snprintf(sizeof(zCmd), zCmd, "\"%s\" http -args \"%s\" --nossl%s",
+    g.nameOfExe, zCmdFName, p->zOptions
   );
   fossil_system(zCmd);
   in = fossil_fopen(zReplyFName, "rb");
@@ -128,9 +219,73 @@ end_request:
   closesocket(p->s);
   file_delete(zRequestFName);
   file_delete(zReplyFName);
-  free(p);
+  file_delete(zCmdFName);
+  fossil_free(p);
 }
 
+/*
+** Process a single incoming SCGI request.
+*/
+static void win32_scgi_request(void *pAppData){
+  HttpRequest *p = (HttpRequest*)pAppData;
+  FILE *in = 0, *out = 0;
+  int amt, got, nHdr, i;
+  int wanted = 0;
+  char zRequestFName[MAX_PATH];
+  char zReplyFName[MAX_PATH];
+  char zCmd[2000];          /* Command-line to process the request */
+  char zHdr[2000];          /* The SCGI request header */
+
+  sqlite3_snprintf(MAX_PATH, zRequestFName,
+                   "%s_in%d.txt", zTempPrefix, p->id);
+  sqlite3_snprintf(MAX_PATH, zReplyFName,
+                   "%s_out%d.txt", zTempPrefix, p->id);
+  out = fossil_fopen(zRequestFName, "wb");
+  if( out==0 ) goto end_request;
+  amt = 0;
+  got = recv(p->s, zHdr, sizeof(zHdr), 0);
+  if( got==SOCKET_ERROR ) goto end_request;
+  amt = fwrite(zHdr, 1, got, out);
+  nHdr = 0;
+  for(i=0; zHdr[i]>='0' && zHdr[i]<='9'; i++){
+    nHdr = 10*nHdr + zHdr[i] - '0';
+  }
+  wanted = nHdr + i + 1;
+  if( strcmp(zHdr+i+1, "CONTENT_LENGTH")==0 ){
+    wanted += atoi(zHdr+i+15);
+  }
+  while( wanted>amt ){
+    got = recv(p->s, zHdr, wanted<sizeof(zHdr) ? wanted : sizeof(zHdr), 0);
+    if( got<=0 ) break;
+    fwrite(zHdr, 1, got, out);
+    wanted += got;
+  }
+  fclose(out);
+  out = 0;
+  assert( g.zRepositoryName && g.zRepositoryName[0] );
+  sqlite3_snprintf(sizeof(zCmd), zCmd,
+    "\"%s\" http \"%s\" \"%s\" %s \"%s\" --scgi --nossl%s",
+    g.nameOfExe, zRequestFName, zReplyFName, inet_ntoa(p->addr.sin_addr),
+    g.zRepositoryName, p->zOptions
+  );
+  fossil_system(zCmd);
+  in = fossil_fopen(zReplyFName, "rb");
+  if( in ){
+    while( (got = fread(zHdr, 1, sizeof(zHdr), in))>0 ){
+      send(p->s, zHdr, got, 0);
+    }
+  }
+
+end_request:
+  if( out ) fclose(out);
+  if( in ) fclose(in);
+  closesocket(p->s);
+  file_delete(zRequestFName);
+  file_delete(zReplyFName);
+  fossil_free(p);
+}
+
+
 /*
 ** Start a listening socket and process incoming HTTP requests on
 ** that socket.
@@ -139,25 +294,40 @@ void win32_http_server(
   int mnPort, int mxPort,   /* Range of allowed TCP port numbers */
   const char *zBrowser,     /* Command to launch browser.  (Or NULL) */
   const char *zStopper,     /* Stop server when this file is exists (Or NULL) */
+  const char *zBaseUrl,     /* The --baseurl option, or NULL */
   const char *zNotFound,    /* The --notfound option, or NULL */
+  const char *zFileGlob,    /* The --fileglob option, or NULL */
+  const char *zIpAddr,      /* Bind to this IP address, if not NULL */
   int flags                 /* One or more HTTP_SERVER_ flags */
 ){
+  HANDLE hStoppedEvent;
   WSADATA wd;
   SOCKET s = INVALID_SOCKET;
   SOCKADDR_IN addr;
   int idCnt = 0;
   int iPort = mnPort;
   Blob options;
-  WCHAR zTmpPath[MAX_PATH];
+  wchar_t zTmpPath[MAX_PATH];
 
-  if( zStopper ) file_delete(zStopper);
   blob_zero(&options);
+  if( zBaseUrl ){
+    blob_appendf(&options, " --baseurl %s", zBaseUrl);
+  }
   if( zNotFound ){
     blob_appendf(&options, " --notfound %s", zNotFound);
   }
+  if( zFileGlob ){
+    blob_appendf(&options, " --files-urlenc %T", zFileGlob);
+  }
   if( g.useLocalauth ){
     blob_appendf(&options, " --localauth");
   }
+  if( g.thTrace ){
+    blob_appendf(&options, " --th-trace");
+  }
+  if( flags & HTTP_SERVER_REPOLIST ){
+    blob_appendf(&options, " --repolist");
+  }
   if( WSAStartup(MAKEWORD(1,1), &wd) ){
     fossil_fatal("unable to initialize winsock");
   }
@@ -168,7 +338,12 @@ void win32_http_server(
     }
     addr.sin_family = AF_INET;
     addr.sin_port = htons(iPort);
-    if( flags & HTTP_SERVER_LOCALHOST ){
+    if( zIpAddr ){
+      addr.sin_addr.s_addr = inet_addr(zIpAddr);
+      if( addr.sin_addr.s_addr == (-1) ){
+        fossil_fatal("not a valid IP address: %s", zIpAddr);
+      }
+    }else if( flags & HTTP_SERVER_LOCALHOST ){
       addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
     }else{
       addr.sin_addr.s_addr = htonl(INADDR_ANY);
@@ -196,21 +371,41 @@ void win32_http_server(
   if( !GetTempPathW(MAX_PATH, zTmpPath) ){
     fossil_fatal("unable to get path to the temporary directory.");
   }
-  zTempPrefix = mprintf("%sfossil_server_P%d_", fossil_unicode_to_utf8(zTmpPath), iPort);
-  fossil_print("Listening for HTTP requests on TCP port %d\n", iPort);
+  zTempPrefix = mprintf("%sfossil_server_P%d_",
+                        fossil_unicode_to_utf8(zTmpPath), iPort);
+  fossil_print("Listening for %s requests on TCP port %d\n",
+               (flags&HTTP_SERVER_SCGI)!=0?"SCGI":"HTTP", iPort);
   if( zBrowser ){
-    zBrowser = mprintf(zBrowser, iPort);
+    zBrowser = mprintf(zBrowser /*works-like:"%d"*/, iPort);
     fossil_print("Launch webbrowser: %s\n", zBrowser);
     fossil_system(zBrowser);
   }
   fossil_print("Type Ctrl-C to stop the HTTP server\n");
+  /* Create an event used to signal when this server is exiting. */
+  hStoppedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
+  assert( hStoppedEvent!=NULL );
+  /* If there is a stopper file name, start the dedicated thread now.
+  ** It will attempt to close the listener socket within 1 second of
+  ** the stopper file being created. */
+  if( zStopper ){
+    HttpServer *pServer = fossil_malloc(sizeof(HttpServer));
+    memset(pServer, 0, sizeof(HttpServer));
+    DuplicateHandle(GetCurrentProcess(), hStoppedEvent,
+                    GetCurrentProcess(), &pServer->hStoppedEvent,
+                    0, FALSE, DUPLICATE_SAME_ACCESS);
+    assert( pServer->hStoppedEvent!=NULL );
+    pServer->zStopper = fossil_strdup(zStopper);
+    pServer->listener = s;
+    file_delete(zStopper);
+    _beginthread(win32_server_stopper, 0, (void*)pServer);
+  }
   /* Set the service status to running and pass the listener socket to the
   ** service handling procedures. */
   win32_http_service_running(s);
   for(;;){
     SOCKET client;
     SOCKADDR_IN client_addr;
-    HttpRequest *p;
+    HttpRequest *pRequest;
     int len = sizeof(client_addr);
     int wsaError;
 
@@ -227,18 +422,23 @@ void win32_http_server(
         WSACleanup();
         fossil_fatal("error from accept()");
       }
-    }else if( zStopper && file_size(zStopper)>=0 ){
-      break;
     }
-    p = fossil_malloc( sizeof(*p) );
-    p->id = ++idCnt;
-    p->s = client;
-    p->addr = client_addr;
-    p->zOptions = blob_str(&options);
-    _beginthread(win32_process_one_http_request, 0, (void*)p);
+    pRequest = fossil_malloc(sizeof(HttpRequest));
+    pRequest->id = ++idCnt;
+    pRequest->s = client;
+    pRequest->addr = client_addr;
+    pRequest->flags = flags;
+    pRequest->zOptions = blob_str(&options);
+    if( flags & HTTP_SERVER_SCGI ){
+      _beginthread(win32_scgi_request, 0, (void*)pRequest);
+    }else{
+      _beginthread(win32_http_request, 0, (void*)pRequest);
+    }
   }
   closesocket(s);
   WSACleanup();
+  SetEvent(hStoppedEvent);
+  CloseHandle(hStoppedEvent);
 }
 
 /*
@@ -248,23 +448,25 @@ void win32_http_server(
 typedef struct HttpService HttpService;
 struct HttpService {
   int port;                 /* Port on which the http server should run */
+  const char *zBaseUrl;     /* The --baseurl option, or NULL */
   const char *zNotFound;    /* The --notfound option, or NULL */
+  const char *zFileGlob;    /* The --files option, or NULL */
   int flags;                /* One or more HTTP_SERVER_ flags */
   int isRunningAsService;   /* Are we running as a service ? */
-  const WCHAR *zServiceName;/* Name of the service */
+  const wchar_t *zServiceName;/* Name of the service */
   SOCKET s;                 /* Socket on which the http server listens */
 };
 
 /*
 ** Variables used for running as windows service.
 */
-static HttpService hsData = {8080, NULL, 0, 0, NULL, INVALID_SOCKET};
+static HttpService hsData = {8080, NULL, NULL, NULL, 0, 0, NULL, INVALID_SOCKET};
 static SERVICE_STATUS ssStatus;
 static SERVICE_STATUS_HANDLE sshStatusHandle;
 
 /*
 ** Get message string of the last system error. Return a pointer to the
-** message string. Call fossil_mbcs_free() to deallocate any memory used
+** message string. Call fossil_unicode_free() to deallocate any memory used
 ** to store the message string when done.
 */
 static char *win32_get_last_errmsg(void){
@@ -273,7 +475,7 @@ static char *win32_get_last_errmsg(void){
   LPWSTR tmp = NULL;
   char *zMsg = NULL;
 
-  /* Try first to get the error text in english. */
+  /* Try first to get the error text in English. */
   nMsg = FormatMessageW(
            FORMAT_MESSAGE_ALLOCATE_BUFFER |
            FORMAT_MESSAGE_FROM_SYSTEM     |
@@ -399,7 +601,8 @@ static void WINAPI win32_http_service_main(
 
    /* Execute the http server */
   win32_http_server(hsData.port, hsData.port,
-                    NULL, NULL, hsData.zNotFound, hsData.flags);
+                    NULL, NULL, hsData.zBaseUrl, hsData.zNotFound,
+                    hsData.zFileGlob, 0, hsData.flags);
 
   /* Service has stopped now. */
   win32_report_service_status(SERVICE_STOPPED, NO_ERROR, 0);
@@ -426,7 +629,9 @@ LOCAL void win32_http_service_running(SOCKET s){
 */
 int win32_http_service(
   int nPort,                /* TCP port number */
+  const char *zBaseUrl,     /* The --baseurl option, or NULL */
   const char *zNotFound,    /* The --notfound option, or NULL */
+  const char *zFileGlob,    /* The --files option, or NULL */
   int flags                 /* One or more HTTP_SERVER_ flags */
 ){
   /* Define the service table. */
@@ -435,7 +640,9 @@ int win32_http_service(
 
   /* Initialize the HttpService structure. */
   hsData.port = nPort;
+  hsData.zBaseUrl = zBaseUrl;
   hsData.zNotFound = zNotFound;
+  hsData.zFileGlob = zFileGlob;
   hsData.flags = flags;
 
   /* Try to start the control dispatcher thread for the service. */
@@ -451,7 +658,8 @@ int win32_http_service(
 
 /* dupe ifdef needed for mkindex
 ** COMMAND: winsrv*
-** Usage: fossil winsrv METHOD ?SERVICE-NAME? ?OPTIONS?
+**
+** Usage: %fossil winsrv METHOD ?SERVICE-NAME? ?OPTIONS?
 **
 ** Where METHOD is one of: create delete show start stop.
 **
@@ -462,7 +670,7 @@ int win32_http_service(
 ** In the following description of the methods, "Fossil-DSCM" will be
 ** used as the default SERVICE-NAME:
 **
-**    fossil winsrv create ?SERVICE-NAME? ?OPTIONS?
+**    %fossil winsrv create ?SERVICE-NAME? ?OPTIONS?
 **
 **         Creates a service. Available options include:
 **
@@ -495,6 +703,10 @@ int win32_http_service(
 **         The following options are more or less the same as for the "server"
 **         command and influence the behaviour of the http server:
 **
+**         --baseurl URL
+**
+**              Use URL as the base (useful for reverse proxies)
+**
 **         -P|--port TCPPORT
 **
 **              Specifies the TCP port (default port is 8080) on which the
@@ -526,24 +738,32 @@ int win32_http_service(
 **              and the "localauth" setting is off and the connection is from
 **              localhost.
 **
+**         --repolist
 **
-**    fossil winsrv delete ?SERVICE-NAME?
+**              If REPOSITORY is directory, URL "/" lists all repositories.
+**
+**         --scgi
+**
+**              Create an SCGI server instead of an HTTP server
+**
+**
+**    %fossil winsrv delete ?SERVICE-NAME?
 **
 **         Deletes a service. If the service is currently running, it will be
 **         stopped first and then deleted.
 **
 **
-**    fossil winsrv show ?SERVICE-NAME?
+**    %fossil winsrv show ?SERVICE-NAME?
 **
 **         Shows how the service is configured and its current state.
 **
 **
-**    fossil winsrv start ?SERVICE-NAME?
+**    %fossil winsrv start ?SERVICE-NAME?
 **
 **         Start the service.
 **
 **
-**    fossil winsrv stop ?SERVICE-NAME?
+**    %fossil winsrv stop ?SERVICE-NAME?
 **
 **         Stop the service.
 **
@@ -568,41 +788,49 @@ void cmd_win32_service(void){
     SC_HANDLE hSvc;
     SERVICE_DESCRIPTIONW
       svcDescr = {L"Fossil - Distributed Software Configuration Management"};
-    char *zErrFmt = "unable to create service '%s': %s";
     DWORD dwStartType = SERVICE_DEMAND_START;
+    const char *zAltBase    = find_option("baseurl", 0, 1);
     const char *zDisplay    = find_option("display", "D", 1);
     const char *zStart      = find_option("start", "S", 1);
     const char *zUsername   = find_option("username", "U", 1);
     const char *zPassword   = find_option("password", "W", 1);
     const char *zPort       = find_option("port", "P", 1);
     const char *zNotFound   = find_option("notfound", 0, 1);
+    const char *zFileGlob   = find_option("files", 0, 1);
     const char *zLocalAuth  = find_option("localauth", 0, 0);
-    const char *zRepository = find_option("repository", "R", 1);
+    const char *zRepository = find_repository_option();
+    int useSCGI             = find_option("scgi", 0, 0)!=0;
+    int allowRepoList       = find_option("repolist",0,0)!=0;
     Blob binPath;
 
     verify_all_options();
     if( g.argc==4 ){
       zSvcName = g.argv[3];
     }else if( g.argc>4 ){
-      fossil_fatal("to much arguments for create method.");
+      fossil_fatal("too many arguments for create method.");
     }
     /* Process service creation specific options. */
     if( !zDisplay ){
       zDisplay = zSvcName;
     }
+    /* Per MSDN, the password parameter cannot be NULL.  Must use empty
+    ** string instead (i.e. in the call to CreateServiceW). */
+    if( !zPassword ){
+      zPassword = "";
+    }
     if( zStart ){
       if( strncmp(zStart, "auto", strlen(zStart))==0 ){
         dwStartType = SERVICE_AUTO_START;
       }else if( strncmp(zStart, "manual", strlen(zStart))==0 ){
         dwStartType = SERVICE_DEMAND_START;
       }else{
-        fossil_fatal(zErrFmt, zSvcName,
+        winhttp_fatal("create", zSvcName,
                      "specify 'auto' or 'manual' for the '-S|--start' option");
       }
     }
     /* Process options for Fossil running as server. */
     if( zPort && (atoi(zPort)<=0) ){
-      fossil_fatal(zErrFmt, zSvcName,
+      winhttp_fatal("create", zSvcName,
                    "port number must be in the range 1 - 65535.");
     }
     if( !zRepository ){
@@ -617,17 +845,21 @@ void cmd_win32_service(void){
     /* Build the fully-qualified path to the service binary file. */
     blob_zero(&binPath);
     blob_appendf(&binPath, "\"%s\" server", g.nameOfExe);
+    if( zAltBase ) blob_appendf(&binPath, " --baseurl %s", zAltBase);
     if( zPort ) blob_appendf(&binPath, " --port %s", zPort);
+    if( useSCGI ) blob_appendf(&binPath, " --scgi");
+    if( allowRepoList ) blob_appendf(&binPath, " --repolist");
     if( zNotFound ) blob_appendf(&binPath, " --notfound \"%s\"", zNotFound);
+    if( zFileGlob ) blob_appendf(&binPath, " --files-urlenc %T", zFileGlob);
     if( zLocalAuth ) blob_append(&binPath, " --localauth", -1);
     blob_appendf(&binPath, " \"%s\"", g.zRepositoryName);
     /* Create the service. */
     hScm = OpenSCManagerW(NULL, NULL, SC_MANAGER_ALL_ACCESS);
-    if( !hScm ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hScm ) winhttp_fatal("create", zSvcName, win32_get_last_errmsg());
     hSvc = CreateServiceW(
              hScm,                                    /* Handle to the SCM */
-             fossil_utf8_to_unicode(zSvcName),           /* Name of the service */
-             fossil_utf8_to_unicode(zDisplay),           /* Display name */
+             fossil_utf8_to_unicode(zSvcName),        /* Name of the service */
+             fossil_utf8_to_unicode(zDisplay),        /* Display name */
              SERVICE_ALL_ACCESS,                      /* Desired access */
              SERVICE_WIN32_OWN_PROCESS,               /* Service type */
              dwStartType,                             /* Start type */
@@ -636,10 +868,10 @@ void cmd_win32_service(void){
              NULL,                                    /* Load ordering group */
              NULL,                                    /* Tag value */
              NULL,                                    /* Service dependencies */
-             fossil_utf8_to_unicode(zUsername),          /* Service account */
-             fossil_utf8_to_unicode(zPassword)           /* Account password */
+             zUsername ? fossil_utf8_to_unicode(zUsername) : 0, /* Account */
+             fossil_utf8_to_unicode(zPassword)        /* Account password */
            );
-    if( !hSvc ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hSvc ) winhttp_fatal("create", zSvcName, win32_get_last_errmsg());
     /* Set the service description. */
     ChangeServiceConfig2W(hSvc, SERVICE_CONFIG_DESCRIPTION, &svcDescr);
     fossil_print("Service '%s' successfully created.\n", zSvcName);
@@ -650,24 +882,24 @@ void cmd_win32_service(void){
     SC_HANDLE hScm;
     SC_HANDLE hSvc;
     SERVICE_STATUS sstat;
-    char *zErrFmt = "unable to delete service '%s': %s";
 
     verify_all_options();
     if( g.argc==4 ){
       zSvcName = g.argv[3];
     }else if( g.argc>4 ){
-      fossil_fatal("to much arguments for delete method.");
+      fossil_fatal("too many arguments for delete method.");
     }
     hScm = OpenSCManagerW(NULL, NULL, SC_MANAGER_ALL_ACCESS);
-    if( !hScm ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
-    hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName), SERVICE_ALL_ACCESS);
-    if( !hSvc ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hScm ) winhttp_fatal("delete", zSvcName, win32_get_last_errmsg());
+    hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName),
+                        SERVICE_ALL_ACCESS);
+    if( !hSvc ) winhttp_fatal("delete", zSvcName, win32_get_last_errmsg());
     QueryServiceStatus(hSvc, &sstat);
     if( sstat.dwCurrentState!=SERVICE_STOPPED ){
       fossil_print("Stopping service '%s'", zSvcName);
       if( sstat.dwCurrentState!=SERVICE_STOP_PENDING ){
         if( !ControlService(hSvc, SERVICE_CONTROL_STOP, &sstat) ){
-          fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+          winhttp_fatal("delete", zSvcName, win32_get_last_errmsg());
         }
       }
       while( sstat.dwCurrentState!=SERVICE_STOPPED ){
@@ -681,7 +913,7 @@ void cmd_win32_service(void){
       if( GetLastError()==ERROR_SERVICE_MARKED_FOR_DELETE ){
         fossil_warning("Service '%s' already marked for delete.\n", zSvcName);
       }else{
-        fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+        winhttp_fatal("delete", zSvcName, win32_get_last_errmsg());
       }
     }else{
       fossil_print("Service '%s' successfully deleted.\n", zSvcName);
@@ -697,7 +929,6 @@ void cmd_win32_service(void){
     LPSERVICE_DESCRIPTIONW pSvcDescr;
     BOOL bStatus;
     DWORD nRequired;
-    const char *zErrFmt = "unable to show service '%s': %s";
     static const char *const zSvcTypes[] = {
       "Driver service",
       "File system driver service",
@@ -724,20 +955,20 @@ void cmd_win32_service(void){
     if( g.argc==4 ){
       zSvcName = g.argv[3];
     }else if( g.argc>4 ){
-      fossil_fatal("to much arguments for show method.");
+      fossil_fatal("too many arguments for show method.");
     }
     hScm = OpenSCManagerW(NULL, NULL, GENERIC_READ);
-    if( !hScm ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hScm ) winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName), GENERIC_READ);
-    if( !hSvc ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hSvc ) winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     /* Get the service configuration */
     bStatus = QueryServiceConfigW(hSvc, NULL, 0, &nRequired);
     if( !bStatus && GetLastError()!=ERROR_INSUFFICIENT_BUFFER ){
-      fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+      winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     }
     pSvcConfig = fossil_malloc(nRequired);
     bStatus = QueryServiceConfigW(hSvc, pSvcConfig, nRequired, &nRequired);
-    if( !bStatus ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !bStatus ) winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     /* Translate the service type */
     switch( pSvcConfig->dwServiceType ){
       case SERVICE_KERNEL_DRIVER:       zSvcType = zSvcTypes[0]; break;
@@ -758,15 +989,15 @@ void cmd_win32_service(void){
     bStatus = QueryServiceConfig2W(hSvc, SERVICE_CONFIG_DESCRIPTION,
                                   NULL, 0, &nRequired);
     if( !bStatus && GetLastError()!=ERROR_INSUFFICIENT_BUFFER ){
-      fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+      winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     }
     pSvcDescr = fossil_malloc(nRequired);
     bStatus = QueryServiceConfig2W(hSvc, SERVICE_CONFIG_DESCRIPTION,
                                   (LPBYTE)pSvcDescr, nRequired, &nRequired);
-    if( !bStatus ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !bStatus ) winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     /* Retrieves the current status of the specified service. */
     bStatus = QueryServiceStatus(hSvc, &sstat);
-    if( !bStatus ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !bStatus ) winhttp_fatal("show", zSvcName, win32_get_last_errmsg());
     /* Translate the current state. */
     switch( sstat.dwCurrentState ){
       case SERVICE_STOPPED:          zSvcState = zSvcStates[0]; break;
@@ -800,24 +1031,24 @@ void cmd_win32_service(void){
     SC_HANDLE hScm;
     SC_HANDLE hSvc;
     SERVICE_STATUS sstat;
-    char *zErrFmt = "unable to start service '%s': %s";
 
     verify_all_options();
     if( g.argc==4 ){
       zSvcName = g.argv[3];
     }else if( g.argc>4 ){
-      fossil_fatal("to much arguments for start method.");
+      fossil_fatal("too many arguments for start method.");
     }
     hScm = OpenSCManagerW(NULL, NULL, SC_MANAGER_ALL_ACCESS);
-    if( !hScm ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
-    hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName), SERVICE_ALL_ACCESS);
-    if( !hSvc ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hScm ) winhttp_fatal("start", zSvcName, win32_get_last_errmsg());
+    hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName),
+                        SERVICE_ALL_ACCESS);
+    if( !hSvc ) winhttp_fatal("start", zSvcName, win32_get_last_errmsg());
     QueryServiceStatus(hSvc, &sstat);
     if( sstat.dwCurrentState!=SERVICE_RUNNING ){
       fossil_print("Starting service '%s'", zSvcName);
       if( sstat.dwCurrentState!=SERVICE_START_PENDING ){
         if( !StartServiceW(hSvc, 0, NULL) ){
-          fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+          winhttp_fatal("start", zSvcName, win32_get_last_errmsg());
         }
       }
       while( sstat.dwCurrentState!=SERVICE_RUNNING ){
@@ -836,24 +1067,24 @@ void cmd_win32_service(void){
     SC_HANDLE hScm;
     SC_HANDLE hSvc;
     SERVICE_STATUS sstat;
-    char *zErrFmt = "unable to stop service '%s': %s";
 
     verify_all_options();
     if( g.argc==4 ){
       zSvcName = g.argv[3];
     }else if( g.argc>4 ){
-      fossil_fatal("to much arguments for stop method.");
+      fossil_fatal("too many arguments for stop method.");
     }
     hScm = OpenSCManagerW(NULL, NULL, SC_MANAGER_ALL_ACCESS);
-    if( !hScm ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
-    hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName), SERVICE_ALL_ACCESS);
-    if( !hSvc ) fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+    if( !hScm ) winhttp_fatal("stop", zSvcName, win32_get_last_errmsg());
+    hSvc = OpenServiceW(hScm, fossil_utf8_to_unicode(zSvcName),
+                        SERVICE_ALL_ACCESS);
+    if( !hSvc ) winhttp_fatal("stop", zSvcName, win32_get_last_errmsg());
     QueryServiceStatus(hSvc, &sstat);
     if( sstat.dwCurrentState!=SERVICE_STOPPED ){
       fossil_print("Stopping service '%s'", zSvcName);
       if( sstat.dwCurrentState!=SERVICE_STOP_PENDING ){
         if( !ControlService(hSvc, SERVICE_CONTROL_STOP, &sstat) ){
-          fossil_fatal(zErrFmt, zSvcName, win32_get_last_errmsg());
+          winhttp_fatal("stop", zSvcName, win32_get_last_errmsg());
         }
       }
       while( sstat.dwCurrentState!=SERVICE_STOPPED ){
