@@ -1,33 +1,49 @@
 #!/bin/sh
 #
-# Run this TCL script to generate a WIKI page that contains a 
+# Run this TCL script to generate a WIKI page that contains a
 # permuted index of the various documentation files.
 #
-#    tclsh mkindex.tcl >permutedindex.wiki
+#    tclsh mkindex.tcl
 #
 
 set doclist {
-  bugtheory.wiki {Bug Tracking In Fossil}
+  aboutcgi.wiki {How CGI Works In Fossil}
+  adding_code.wiki {Adding New Features To Fossil}
+  adding_code.wiki {Hacking Fossil}
+  antibot.wiki {Defense against Spiders and Bots}
+  blame.wiki {The Annotate/Blame Algorithm Of Fossil}
   branching.wiki {Branching, Forking, Merging, and Tagging}
-  build.wiki {Building and Installing Fossil}
-  checkin_names.wiki {Checkin And Version Names}
-  checkin.wiki {Check-in Checklist}
+  bugtheory.wiki {Bug Tracking In Fossil}
+  build.wiki {Compiling and Installing Fossil}
   changes.wiki {Fossil Changelog}
+  checkin_names.wiki {Check-in And Version Names}
+  checkin.wiki {Check-in Checklist}
+  childprojects.wiki {Child Projects}
   copyright-release.html {Contributor License Agreement}
   concepts.wiki {Fossil Core Concepts}
   contribute.wiki {Contributing Code or Documentation To The Fossil Project}
+  customgraph.md {Theming: Customizing the Timeline Graph}
+  customskin.md {Theming: Customizing The Appearance of Web Pages}
   custom_ticket.wiki {Customizing The Ticket System}
   delta_encoder_algorithm.wiki {Fossil Delta Encoding Algorithm}
   delta_format.wiki {Fossil Delta Format}
   embeddeddoc.wiki {Embedded Project Documentation}
+  encryptedrepos.wiki {How To Use Encrypted Repositories}
+  env-opts.md {Environment Variables and Global Options}
   event.wiki {Events}
   faq.wiki {Frequently Asked Questions}
   fileformat.wiki {Fossil File Format}
+  fiveminutes.wiki {Update and Running in 5 Minutes as a Single User}
   foss-cklist.wiki {Checklist For Successful Open-Source Projects}
+  fossil-from-msvc.wiki {Integrating Fossil in the Microsoft Express 2010 IDE}
   fossil-v-git.wiki {Fossil Versus Git}
+  hacker-howto.wiki {Hacker How-To}
+  /help {Lists of Commands and Webpages}
+  hints.wiki {Fossil Tips And Usage Hints}
   index.wiki {Home Page}
   inout.wiki {Import And Export To And From Git}
   makefile.wiki {The Fossil Build Process}
+  /md_rules {Markdown Formatting Rules}
   newrepo.wiki {How To Create A New Fossil Repository}
   password.wiki {Password Management And Authentication}
   pop.wiki {Principles Of Operations}
@@ -42,6 +58,7 @@ set doclist {
   selfhost.wiki {Fossil Self Hosting Repositories}
   server.wiki {How To Configure A Fossil Server}
   settings.wiki {Fossil Settings}
+  /sitemap {Site Map}
   shunning.wiki {Shunning: Deleting Content From Fossil}
   stats.wiki {Performance Statistics}
   style.wiki {Source Code Style Guidelines}
@@ -50,9 +67,16 @@ set doclist {
   tech_overview.wiki {A Technical Overview Of The Design And Implementation
                       Of Fossil}
   tech_overview.wiki {SQLite Databases Used By Fossil}
+  th1.md {The TH1 Scripting Language}
+  tickets.wiki {The Fossil Ticket System}
   theory1.wiki {Thoughts On The Design Of The Fossil DVCS}
+  unvers.wiki {Unversioned Files}
+  webpage-ex.md {Webpage Examples}
   webui.wiki {The Fossil Web Interface}
+  whyusefossil.wiki {Why You Should Use Fossil}
+  whyusefossil.wiki {Benefits Of Version Control}
   wikitheory.wiki {Wiki In Fossil}
+  /wiki_rules {Wiki Formatting Rules}
 }
 
 set permindex {}
@@ -71,24 +95,33 @@ foreach {file title} $doclist {
   }
 }
 set permindex [lsort -dict -index 0 $permindex]
-set out [open permutedindex.wiki w]
+set out [open permutedindex.html w]
 fconfigure $out -encoding utf-8 -translation lf
-puts $out "<title>Index Of Fossil Documentation</title>"
+puts $out \
+"<div class='fossil-doc' data-title='Index Of Fossil Documentation'>"
 puts $out {
+<center>
+<form action='$ROOT/docsrch' method='GET'>
+<input type="text" name="s" size="40" autofocus>
+<input type="submit" value="Search Docs">
+</form>
+</center>
 <h2>Primary Documents:</h2>
 <ul>
-<li> [./quickstart.wiki | Quick-start Guide]
-<li> [./faq.wiki | FAQ]
-<li> [./build.wiki | Building and installing Fossil]
-<li> [../COPYRIGHT-BSD2.txt | License]
-<li> [http://www.fossil-scm.org/schimpf-book/home | Jim Schimpf's book]
-<li> [/help | Command-line help]
+<li> <a href='quickstart.wiki'>Quick-start Guide</a>
+<li> <a href='faq.wiki'>FAQ</a>
+<li> <a href='build.wiki'>Compiling and installing Fossil</a>
+<li> <a href='../COPYRIGHT-BSD2.txt'>License</a>
+<li> <a href='http://www.fossil-scm.org/schimpf-book/home'>Jim Schimpf's
+book</a>
+<li> <a href='$ROOT/help'>Command-line help</a>
 </ul>
 <a name="pindex"></a>
 <h2>Permuted Index:</h2>
 <ul>}
 foreach entry $permindex {
   foreach {title file} $entry break
+  if {[string match /* $file]} {set file ../../..$file}
   puts $out "<li><a href=\"$file\">$title</a></li>"
 }
-puts $out "</ul>"
+puts $out "</ul></div>"
